#include "EvalCorrs.h"
#include "TemplateMappings.h"

/**
 * Evaluate correspondences (with ground truth)
 * Measure template quality (without ground truth)
 */
int main(int argc, char** argv )
{
	if (argc<2)
	{
		std::cout<<"Usage (note it appends to outFile):"<<std::endl;
		std::cout<<"   "<<argv[0]<<" [params]"<<std::endl;;
		std::cout<<"Args: "<<std::endl;
		std::cout<<"    -writeFile outFile - write file"<<std::endl;
		std::cout<<"    -evalFuzzyCorrs fuzzy.bin - fuzzy corrs file"<<std::endl;
		std::cout<<"    -readTemplate file.tmplt - read template"<<std::endl;
		std::cout<<"    -metric [Sharpness Corrs] - evaluation metric"<<std::endl;
		std::cout<<"    -coAlignedPointsFiles filePrefix "<<std::endl;
		std::cout<<"    -sharpnessResolution Nx Ny Nz "<<std::endl;
		std::cout<<"    -dbDir dir "<<std::endl;
		std::cout<<"    -matchDir dir "<<std::endl;
		std::cout<<"    -vis "<<std::endl;		
		exit(0);
	}
	
	bool visualize = false;
	VKString writeFile = "none";
	VKString readTemplate = "none";
	VKString metric = "none";
	VKString coAlignedPointsFiles = "none";
	VKString dbDir = "none";
	VKString matchDir = "none";
	VKString evalFuzzyCorrs = "none";
	int Ndefault = 128;
	int Nx=Ndefault, Ny=Ndefault, Nz=Ndefault;

	for (int i=1; i<argc; i++)
	{
		if (strcmp(argv[i], "-writeFile")==0)
			writeFile = argv[++i];
		else if (strcmp(argv[i], "-readTemplate")==0)
			readTemplate = argv[++i];
		else if (strcmp(argv[i], "-metric")==0)
			metric = argv[++i];
		else if (strcmp(argv[i], "-coAlignedPointsFiles")==0)
			coAlignedPointsFiles = argv[++i];
		else if (strcmp(argv[i], "-dbDir")==0)
			dbDir = argv[++i];
		else if (strcmp(argv[i], "-matchDir")==0)
			matchDir = argv[++i];
		else if (strcmp(argv[i], "-evalFuzzyCorrs")==0)
			evalFuzzyCorrs = argv[++i];
		else if (strcmp(argv[i], "-sharpnessResolution")==0)
		{
			Nx = atoi(argv[++i]);
			Ny = atoi(argv[++i]);
			Nz = atoi(argv[++i]);
		}
		else if (strcmp(argv[i], "-vis")==0)
			visualize = true;
	}
	
	STemplate * st = NULL;
	TMatch * rootMatch = NULL;
	VKString writeOut="";
	if (readTemplate!="none")
	{
		st = STemplate::ReadTemplate(readTemplate, NULL);
		rootMatch = st->CreateSeedConfig(0, NULL, std::vector<int>(), R3identity_affine);
	}
	
	std::cout<<"Metric: "<<metric.c_str()<<std::endl;
	if (metric == "Sharpness")
	{
		VKStringList prefix;
		prefix.push_back("CoAligned");
		prefix.push_back("RigidAligned");
		
		for (int i=0; i<prefix.count(); i++)
		{
			std::vector<double> energies;
			int currPrefix = 0;
			while(true)
			{	
				VKString coAlignedFile = "";
				if (currPrefix==0)
					coAlignedFile = coAlignedPointsFiles+prefix[i]+"_D0.bin";
				else
					coAlignedFile = coAlignedPointsFiles+prefix[i]+"_S"+VKString::number(currPrefix)+".bin";
				currPrefix++;
				if (!FileManagement::FileExists(coAlignedFile.c_str()))
					break;
				std::vector<std::vector<R3Point> > coAlignedPoints;
				ReadCoAlignedPoints(coAlignedFile, coAlignedPoints);
				// find boundaries
				double xmin=10e10, ymin=10e10, zmin=10e10, xmax=-10e10, ymax=-10e10, zmax=-10e10;
				int npoints=0;
				for (int m=0; m<(int)coAlignedPoints.size(); m++)
				for (int p=0; p<(int)coAlignedPoints[m].size(); p++)
					npoints++;
				for (int p=0; p<rootMatch->NTemplatePoints(); p++)
				{
					R3Point gp = rootMatch->DeformedTemplateSample(p);
					xmin = vkMin(xmin, gp.X()-.1);
					ymin = vkMin(ymin, gp.Y()-.1);
					zmin = vkMin(zmin, gp.Z()-.1);
					xmax = vkMax(xmax, gp.X()+.1);
					ymax = vkMax(ymax, gp.Y()+.1);
					zmax = vkMax(zmax, gp.Z()+.1);
				}
				// create grid and rasterize
				R3Grid gridTmplt(Nx, Ny, Nz, R3Box(xmin, ymin, zmin, xmax, ymax, zmax));
				R3Grid gridCoAligned(Nx, Ny, Nz, R3Box(xmin, ymin, zmin, xmax, ymax, zmax));
				for (int p=0; p<rootMatch->NTemplatePoints(); p++)
				{
					R3Point gp = rootMatch->DeformedTemplateSample(p);
					gridTmplt.RasterizeWorldPoint(gp, 1);
				}
				gridTmplt.SquaredDistanceTransform();
				gridTmplt.Gauss();
				
				int nentries=0;
				for (int m=0; m<(int)coAlignedPoints.size(); m++)
				for (int p=0; p<(int)coAlignedPoints[m].size(); p++)
				{
					R3Point gp = gridCoAligned.GridPosition(coAlignedPoints[m][p]);
					bool validGridPos = ((int)gp.X() < Nx && gp.X() >= 0);
					validGridPos = validGridPos && ((int)gp.Y() < Ny && gp.Y() >= 0);
					validGridPos = validGridPos && ((int)gp.Z() < Nz && gp.Z() >= 0);
					if (validGridPos)
					{
						nentries++;
						double w = gridTmplt.GridValue((int)gp.X(), (int)gp.Y(), (int)gp.Z());
						gridCoAligned.AddGridValue((int)gp.X(), (int)gp.Y(), (int)gp.Z(), w);
					}
					else
						std::cout<<"[WARNING] Invalid grid pos: ["<<gp.X()<<" "<<gp.Y()<<" "<<gp.Z()<<"] "<<std::endl;
				}

				gridCoAligned.Divide((double)nentries);
				energies.push_back(gridCoAligned.StandardDeviation()*Nx*Ny*Nz * 1000.);
			}
			if (energies.size()==0)
				std::cout<<"[ERROR] Could not find energies. Dir = "<<coAlignedPointsFiles.c_str()<<std::endl;
			assert(energies.size()>0);
			writeOut += metric+"_"+prefix[i]+" ";
			for (int e=0; e<(int)energies.size(); e++)
				writeOut += VKString::number(energies[e])+(e==((int)energies.size()-1)?"\n":" ");
		}
	}
	else if (metric == "Corrs")
	{
		if (evalFuzzyCorrs!="none")	// fuzzy correspondences matrix (not a template!)	// NOTE: Matchdir = model list
			EvalFuzzyCorrs(evalFuzzyCorrs, matchDir, dbDir, writeFile, visualize);
		else
		{
			assert(st!=NULL);
			EvalCorrs(st, matchDir+"/LabeledModels.txt", matchDir, dbDir, writeFile, visualize);
		}
	}
	else
		std::cout<<"[ERROR] Unknown metric "<<metric.c_str()<<" in EvalTemplate.cpp"<<std::endl;
	
	if (metric!="Corrs")
	{
		if (writeFile!="none")
		{
			std::cout<<"Writing: "<<writeFile.c_str()<<std::endl;
			std::ofstream textStream(writeFile.c_str());
			assert(textStream.is_open());
			textStream<<writeOut.c_str();
			textStream.close();
		}
	}
	return 0;
}


