#include "TemplateMappings.h"

///// UnqiueTemplatePoint /////
bool UnqiueTemplatePoint::m_bKeepOffset = true;

R3Point MapPoint(int gtid, std::vector<std::vector<Tmplt2TmpltMapping*> > * tmplt2tmpltMapping,
				 std::vector<Model2TmpltMapping*> * mdl2tmpltMapping, int fromModelID, int toModelID,
				 bool verb, bool & valid, int *opt_t1, int * opt_t2)
{
	
	valid = gtid < (int)(*mdl2tmpltMapping)[fromModelID]->m_bValidFeatPnt.size();
	valid = valid && (*mdl2tmpltMapping)[fromModelID]->m_bValidFeatPnt[gtid];
	if (!valid)
		return R3Point(0,0,0);
	
	int c1 = (*mdl2tmpltMapping)[fromModelID]->m_iConfigID;
	int c2 = (*mdl2tmpltMapping)[toModelID]->m_iConfigID;

	// feature point -> template ID
	UnqiueTemplatePoint t1 = (*mdl2tmpltMapping)[fromModelID]->FromFeature2Tmplt(gtid, valid);
	
	if (!valid)
		return R3Point(0,0,0);
	assert(t1.m_iUnqiuePointID>=0);
	
	if (opt_t1!=NULL)
		*opt_t1 = t1.m_iUnqiuePointID;
	
	// template ID -> template ID (in different configurations)
//	if (!(c1 < (int)(*tmplt2tmpltMapping).size() && c2 < (int)(*tmplt2tmpltMapping).size() && c1>0 && c2>0))
//	{
		//std::cout<<"[ERROR] Invalid configurations: "<<c1<<" "<<c2<<" / "<<(*tmplt2tmpltMapping).size()<<std::endl;
//	}
	assert(c1 < (int)(*tmplt2tmpltMapping).size() && c2 < (int)(*tmplt2tmpltMapping).size() && c1>=0 && c2>=0);
	assert((*tmplt2tmpltMapping)[c1][c2]!=NULL);
	UnqiueTemplatePoint t2 = (*tmplt2tmpltMapping)[c1][c2]->MapConfig1ToConfig2(t1);
	
	// template ID -> model point
	R3Point p2 = (*mdl2tmpltMapping)[toModelID]->FromTemplate2Model(t2, &(t2.m_iUnqiuePointID));
	assert(t2.m_iUnqiuePointID>=0);
	if (opt_t2!=NULL)
		*opt_t2 = t2.m_iUnqiuePointID;
	
	if (verb)
	{
		R3Point p1gt = (*mdl2tmpltMapping)[fromModelID]->m_GroundTruthPoint[gtid];
		R3Point p2gt = (*mdl2tmpltMapping)[toModelID]->m_GroundTruthPoint[gtid];
		std::cout<<"Mapping: "<<fromModelID<<" -> "<<toModelID<<". P = "<<gtid<<std::endl;
		std::cout<<"    Configs: "<<c1<<" -> "<<c2<<std::endl;
		std::cout<<"    p1gt: ["<<p1gt.X()<<" "<<p1gt.Y()<<" "<<p1gt.Z()<<"] "<<std::endl;
		std::cout<<"    t1id: "<<t1.m_iUnqiuePointID<<std::endl;
		std::cout<<"    t1: ["<<t1.m_PntPos.X()<<" "<<t1.m_PntPos.Y()<<" "<<t1.m_PntPos.Z()<<"] "<<std::endl;
		std::cout<<"    t1.offset: ["<<t1.m_Offset.X()<<" "<<t1.m_Offset.Y()<<" "<<t1.m_Offset.Z()<<"] "<<std::endl;
		std::cout<<"    t2id: "<<t2.m_iUnqiuePointID<<std::endl;
		std::cout<<"    t2: ["<<t2.m_PntPos.X()<<" "<<t2.m_PntPos.Y()<<" "<<t2.m_PntPos.Z()<<"] "<<std::endl;
		std::cout<<"    t2.offset: ["<<t2.m_Offset.X()<<" "<<t2.m_Offset.Y()<<" "<<t2.m_Offset.Z()<<"] "<<std::endl;
		std::cout<<"    p2: ["<<p2.X()<<" "<<p2.Y()<<" "<<p2.Z()<<"] "<<std::endl;
		std::cout<<"    p2gt: ["<<p2gt.X()<<" "<<p2gt.Y()<<" "<<p2gt.Z()<<"] "<<std::endl;
	}
	
	return p2;
}

UnqiueTemplatePoint::UnqiueTemplatePoint(const R3Point &pnt, const R3Vector & offset, int uniqueID)
{
	m_iUnqiuePointID = uniqueID;
	m_PntPos = pnt;
	m_Offset = offset;
}

UnqiueTemplatePoint::UnqiueTemplatePoint(const UnqiueTemplatePoint & pnt)
{
	m_iUnqiuePointID = pnt.m_iUnqiuePointID;
	m_PntPos = pnt.m_PntPos;
	m_Offset = pnt.m_Offset;
}

///// Tmplt2TmpltMapping /////
Tmplt2TmpltMapping::Tmplt2TmpltMapping(STemplate * st, TMatch * config1, TMatch * config2)
{
	// use alignment for now...
	m_iConfig1 = config1->SeedID();
	m_iConfig2 = config2->SeedID();
	
	bool sameIndexing = false;	// indexing does not change (in practice changes due to re-sampling)
		
	if (sameIndexing)	// BELOW IS NEVER EXECUTED! - can be if we have meshes w/o resampling
	{
		assert(false);
		if (m_iConfig1!=m_iConfig2)
		{
			ShapePointCloud * pnts2 = new ShapePointCloud(NULL, "tmppnts");
			for (int i=0; i<config2->NTemplatePoints(); i++)
				pnts2->AddPoint(ShapePoint(st->TemplatePointInCanonicalSpace(config2, i)));
			const R3Kdtree<const PointInSet *> * kdTree2 = pnts2->GetKdTree();
			
			for (int i=0; i<config1->NTemplatePoints(); i++)
			{
				R3Point t1 = st->TemplatePointInCanonicalSpace(config1, i);
				const PointInSet * closest = kdTree2->FindClosest(t1);
				assert(closest!=NULL);
				m_Config1ToConfig2.push_back(UnqiueTemplatePoint(closest->GetPosition(), R3Vector(0,0,0), closest->id));
			}
//			delete pnts2;
		}
		else 
		{
			for (int i=0; i<config1->NTemplatePoints(); i++)
			{
				R3Point p = st->TemplatePointInCanonicalSpace(config2, i);
				m_Config1ToConfig2.push_back(UnqiueTemplatePoint(p, R3Vector(0,0,0), i));
			}
		}
	}
	
	m_pMatch1 = config1;
	m_pMatch2 = config2;
}

bool Tmplt2TmpltMapping::Verify(int config1, int config2) const
{
	assert(m_iConfig1==config1 && m_iConfig2==config2);
}

UnqiueTemplatePoint Tmplt2TmpltMapping::MapConfig1ToConfig2(const UnqiueTemplatePoint & pnt) const
{
	if (m_Config1ToConfig2.size()>0 && pnt.m_iUnqiuePointID>=0)
	{
		assert(false);	// NOTE: a potential problem - indexing might change due to included / excluded parts
		assert(pnt.m_iUnqiuePointID>=0 && pnt.m_iUnqiuePointID < (int)m_Config1ToConfig2.size());
		return m_Config1ToConfig2[pnt.m_iUnqiuePointID];
	}
	
	return UnqiueTemplatePoint(pnt.m_PntPos, pnt.m_Offset, -1);
}


///// Model2TmpltMapping /////
Model2TmpltMapping::Model2TmpltMapping(TMatch * match, std::vector<ShapePoint *> & featPnts)
{
	m_iConfigID = match->SeedID();
	const ShapePointCloud * matchedPnts = match->GetMatchedGeomAsPnts();
	assert(matchedPnts!=NULL);
	m_MeshNormCenter = matchedPnts->Centroid();
	m_fMeshNormRadius = matchedPnts->AverageRadius(m_MeshNormCenter);

	// if no points are included
	for (int i=0; i<match->NTemplatePoints(); i++)
	{
		int deformID, locPntID;
		match->LocalPntID(i, deformID, locPntID);
		bool included = match->Deformation(deformID).IsIncluded();
			
		m_TemplateAllID2IncludedMap.push_back(-1);
		if (included)
		{
			m_TemplateAllID2IncludedMap[i] = (int)m_TemplateInlcuded2AllIDMap.size();
			m_TemplateInlcuded2AllIDMap.push_back(i);
		}
	}
	
	// pnts == deformed template samples
	ShapePointCloud * pnts = new ShapePointCloud(NULL, "tmppnts");
	for (int i=0; i<(int)m_TemplateInlcuded2AllIDMap.size(); i++)
		pnts->AddPoint(ShapePoint(match->DeformedTemplateSample(m_TemplateInlcuded2AllIDMap[i])));
	const R3Kdtree<const PointInSet *> * kdTree = pnts->GetKdTree();
	
	m_pUndeformedTemplatePnts = new ShapePointCloud(NULL, "tmppnts");
	for (int i=0; i<(int)m_TemplateInlcuded2AllIDMap.size(); i++)
	{
		int gid = m_TemplateInlcuded2AllIDMap[i];
		R3Point p = match->GetSTemplate()->TemplatePointInCanonicalSpace(match, gid);
		
		//R3Point defP = match->DeformedTemplateSample(gid);
		//R3Point p = defP;
//		if (p.X()>5)
//			std::cout<<"Mapping: "<<i<<": ["<<defP.X()<<" -> "<<p.X()<<"] "<<std::endl;
		
		m_pUndeformedTemplatePnts->AddPoint(ShapePoint(p));
	}
	m_pKdTreeUndeformedTemplate = m_pUndeformedTemplatePnts->GetKdTree();

	// feat pnts -> template
	for (int i=0; i<(int)featPnts.size(); i++)
	{
		if (featPnts[i]==NULL)
		{
			m_FeatPnt2Template.push_back(UnqiueTemplatePoint(R3Point(0,0,0), R3Vector(0,0,0), -1));
			m_bValidFeatPnt.push_back(false);
			m_GroundTruthPoint.push_back(R3Point(0,0,0));
		}
		else
		{
			R3Point fp = (featPnts[i]->GetPosition()+(-m_MeshNormCenter)) / m_fMeshNormRadius;
			match->GlobalXForm().Apply(fp);
			const PointInSet * closest = kdTree->FindClosest(fp);
			// deformed template point
			assert(closest!=NULL);
			int gid = m_TemplateInlcuded2AllIDMap[closest->id];
			// same point undeformed
			R3Point pos = m_pUndeformedTemplatePnts->PointPosition(closest->id);
			//std::cout<<"TMPLT PNT: "<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<" "<<std::endl;
			// vector from template to aligned model in deformed space
			R3Vector offset = fp - closest->GetPosition();
			//std::cout<<"offset: ["<<fp.X()<<" "<<fp.Y()<<" "<<fp.Z()<<"] "<<std::endl;
			//R3Vector offset(0,0,0);	// vector from template to aligned model
			m_FeatPnt2Template.push_back(UnqiueTemplatePoint(pos, offset, gid));
			
			m_bValidFeatPnt.push_back(true);
			m_GroundTruthPoint.push_back(featPnts[i]->GetPosition());
		}
	}
	
	// template -> model	// vk: NOT VALID IF THERE IS TRANSFORMATION
	const R3Kdtree<const PointInSet *> * matchedKdTree = matchedPnts->GetKdTree();
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		R3Point tp = pnts->PointPosition(i);
		match->GlobalXForm().ApplyInverse(tp);
		tp = (tp * m_fMeshNormRadius) + m_MeshNormCenter;
		const PointInSet * closest = matchedKdTree->FindClosest(tp);
		assert(closest!=NULL);
		m_Tmplt2Model.push_back(closest->GetPosition());
		//m_Tmplt2Model.push_back(tp);
	}
	
	m_fRadius = GetMeshRadius(matchedPnts->GetMesh());

//	m_fRadius = 0;
//	for (int p1=0; p1<matchedPnts->NumPoints(); p1++)
//	for (int p2=p1+1; p2<matchedPnts->NumPoints(); p2++)
//		m_fRadius = vkMax(m_fRadius, R3Distance(matchedPnts->PointPosition(p1), matchedPnts->PointPosition(p2)));
//	delete pnts;
	m_pMatch = match;
}

R3Point Model2TmpltMapping::GetGroundTruthPoint(int featPntID, bool & valid) const
{
	if (featPntID>=0 && featPntID < (int)m_GroundTruthPoint.size())
	{
		valid = m_bValidFeatPnt[featPntID];
		return m_GroundTruthPoint[featPntID];
	}
	else
	{
		valid = false;
		return R3Point(0,0,0);
	}
}

UnqiueTemplatePoint Model2TmpltMapping::FromFeature2Tmplt(int featPntID, bool & valid) const
{
	if (featPntID>=0 && featPntID < (int)m_FeatPnt2Template.size())
	{
		valid = m_bValidFeatPnt[featPntID];
		return m_FeatPnt2Template[featPntID];
	}
	else
	{
		valid = false;
		return UnqiueTemplatePoint(R3Point(0,0,0), R3Vector(0,0,0), -1);
	}
}
 
R3Point Model2TmpltMapping::FromTemplate2Model(const UnqiueTemplatePoint & uniquePnt, int * optTmpltID) const
{
	int tmpltID =  uniquePnt.m_iUnqiuePointID;
	if (tmpltID<0)
	{
		assert(m_pKdTreeUndeformedTemplate!=NULL);
		const PointInSet * closest = m_pKdTreeUndeformedTemplate->FindClosest(uniquePnt.m_PntPos);
		assert(closest!=NULL);
		tmpltID = closest->id;
	}
	else
	{
		assert(false);
		tmpltID = m_TemplateAllID2IncludedMap[tmpltID];
	}
	if (optTmpltID!=NULL)
		*optTmpltID = tmpltID;
	
	if (!UnqiueTemplatePoint::m_bKeepOffset)
	{
		assert(tmpltID >= 0 && tmpltID < (int)m_Tmplt2Model.size());
		return m_Tmplt2Model[tmpltID];
	}
	else
	{
		// get deformed point tmpltID
		R3Point deformedTmpltPnt = m_pMatch->DeformedTemplateSample(tmpltID);
		
		// transform by offset
		R3Point modelSpacePnt = deformedTmpltPnt + uniquePnt.m_Offset;
		
		// project on mesh
		m_pMatch->GlobalXForm().ApplyInverse(modelSpacePnt);
		modelSpacePnt = modelSpacePnt * m_fMeshNormRadius + m_MeshNormCenter;
		const PointInSet * closest = m_pMatch->GetMatchedGeomAsPnts()->GetKdTree()->FindClosest(modelSpacePnt);
		assert(closest!=NULL);
		return closest->GetPosition();
	}
}

double Model2TmpltMapping::Radius() const
{
	return m_fRadius;
}

int Model2TmpltMapping::ConfigID() const
{
	return m_iConfigID;
}


//////////////// TMapDebugApp
TMapDebugApp::TMapDebugApp(ModelDatabase * db, std::vector<std::vector<Tmplt2TmpltMapping*> > & t2tmapping,
						   std::vector<Model2TmpltMapping*> & perModelAlign)
{
	m_pT2TMapping = &t2tmapping;
	m_pMdl2TmpltMapping = &perModelAlign;
	m_pGtPnts = NULL;
	m_pFuzzyPnts = NULL;
	m_pV2VMap = NULL;
	m_pMaxPairwiseErr = NULL;
	m_iCurrPair = -1;
}

TMapDebugApp::TMapDebugApp(ModelDatabase * db, std::vector<std::vector<ShapePoint*> > & gtpnts,
						   std::vector<ShapePointCloud *> & fuzzyPnts, std::map<int, R3Point> * v2vmaps)
{
	m_pT2TMapping = NULL;
	m_pMdl2TmpltMapping = NULL;
	m_pGtPnts = &gtpnts;
	m_pFuzzyPnts = &fuzzyPnts;
	m_pV2VMap = v2vmaps;
	m_pMaxPairwiseErr = NULL;
	m_iCurrPair = -1;
}


bool TMapDebugApp::KeyPressed(DatabaseVisWindow * window, const char key, int mouseX, int mouseY)
{
	if (key=='`')
	{
		std::cout<<"Hot keys:"<<std::endl;
		std::cout<<"     p 		- print all errors"<<std::endl;
		std::cout<<"     0-9 	- print mapping details for a point"<<std::endl;
		std::cout<<"     ,/. 	- iterate over least/most-error pairs"<<std::endl;
		std::cout<<"     </> 	- iterate by model over least/most-error pairs"<<std::endl;
	}
	else if (key=='p')
	{
		int mid1 = window->WindowToModel()[0];
		int mid2 = window->WindowToModel()[1];
		std::cout<<"Errors: "<<std::endl;

		for (int p=0; p<(int)(*m_pMdl2TmpltMapping)[mid1]->m_GroundTruthPoint.size(); p++)
		{
			bool valid = (*m_pMdl2TmpltMapping)[mid1]->m_bValidFeatPnt[p];
			R3Point p1 = (*m_pMdl2TmpltMapping)[mid1]->m_GroundTruthPoint[p];
			R3Point p2, p2gt;
			if (valid)
			{
				p2 = MapPoint(p, m_pT2TMapping, m_pMdl2TmpltMapping, mid1, mid2, false, valid);
				p2gt = (*m_pMdl2TmpltMapping)[mid2]->m_GroundTruthPoint[p];
			}
			if (valid)
			{
				double r = (*m_pMdl2TmpltMapping)[mid2]->m_fRadius;
				std::cout<<"    Err["<<p<<"]: "<<(R3Distance(p2, p2gt)/r)<<std::endl;
			}
			else
			{
				std::cout<<"    Err["<<p<<"]: Invalid "<<std::endl;
			}
		}
		return true;
	}
	else if (key-'0'>=0 && key-'0'<=9)
	{
		bool valid;
		int mid1 = window->WindowToModel()[0];
		int mid2 = window->WindowToModel()[1];
		MapPoint((int)(key-'0'), m_pT2TMapping, m_pMdl2TmpltMapping, mid1, mid2, true, valid);
		return true;
	}
	else if (key=='.' || key==',' || key=='<'|| key=='>')
	{
		if (m_pMaxPairwiseErr==NULL)
		{
			std::cout<<"[WARNING] Pairwise errors not set"<<std::endl;
			return true;
		}
		int nmodels = (int)sqrt((double)m_pMaxPairwiseErr->size());
		if (m_iCurrPair < 0)
			m_iCurrPair = 0;
		if (key=='.')
			m_iCurrPair++;
		else if (key==',')
			m_iCurrPair--;
		else
		{
			int mid1Orig = (*m_pMaxPairwiseErr)[m_iCurrPair].id / nmodels;
			int mid2Orig = (*m_pMaxPairwiseErr)[m_iCurrPair].id % nmodels;
			int mid1 = mid1Orig;
			int mid2 = mid2Orig;
			while (mid1 == mid1Orig || mid2 == mid2Orig || mid1==mid2Orig || mid2==mid1Orig)
			{
				if (key=='<')
					m_iCurrPair--;
				else if (key=='>')
					m_iCurrPair++;
				if (m_iCurrPair < 0)
				{
					std::cout<<"[WARNING] No models with higher error"<<std::endl;
					m_iCurrPair = 0;
					break;
				}
				else if (m_iCurrPair>=(int)m_pMaxPairwiseErr->size())
				{
					std::cout<<"[WARNING] No models with smaller error"<<std::endl;
					m_iCurrPair = (int)m_pMaxPairwiseErr->size();
					break;
				}
				mid1 = (*m_pMaxPairwiseErr)[m_iCurrPair].id / nmodels;
				mid2 = (*m_pMaxPairwiseErr)[m_iCurrPair].id % nmodels;
			}
		}
		if (m_iCurrPair < 0)
			m_iCurrPair = (int)m_pMaxPairwiseErr->size()-1;
		else if (m_iCurrPair>=(int)m_pMaxPairwiseErr->size())
			m_iCurrPair = 0;
		
		int mid1 = (*m_pMaxPairwiseErr)[m_iCurrPair].id / nmodels;
		int mid2 = (*m_pMaxPairwiseErr)[m_iCurrPair].id % nmodels;
		window->SelectModelWindow(0);
		window->SelectModelInWindow(mid1);
		window->SelectModelWindow(1);
		window->SelectModelInWindow(mid2);

		
		std::cout<<"Pair: "<<m_iCurrPair<<" / "<<m_pMaxPairwiseErr->size();
		std::cout<<". "<<mid1<<" -> "<<mid2<<". "<<(*m_pMaxPairwiseErr)[m_iCurrPair].value<<std::endl;
		
		return true;
	}
	return false;
}

bool TMapDebugApp::Draw(DatabaseVisWindow * window)
{
	double br = 0, bg = 0, bb = 0;
	//window->GetBackgroundColor(&br, &bg, &bb);
	glClearColor(br, bg, bb, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	if (m_pT2TMapping!=NULL)
		DrawTemplate(window);
	else
		DrawFuzzy(window);
		
	if (window->IsDesignFlagOn(DatabaseVisWindow::RENDER_DESIGN_BOXES))
	{
		for (int i=0; i<2; i++)
			window->DrawSubwindow(i);
	}
	
	if (window->IsDesignFlagOn(DatabaseVisWindow::RENDER_MODEL_LIST))
		window->DrawModelListSubwindow();
	return true;
}

void TMapDebugApp::SetMaxPairwiseErr(std::vector<Sortable> * maxpairerr)
{
	m_pMaxPairwiseErr = maxpairerr ; 
}

void TMapDebugApp::DrawTemplate(DatabaseVisWindow * window)
{
	assert(m_pMdl2TmpltMapping!=NULL && m_pT2TMapping!=NULL);
	assert(window->GetDrawableModels().size()>=2);
	std::vector<R3Point> mappedPnts;
	std::vector<bool> mappedPntValid;
	std::map<int, int> mappedPntsTmplt1;
	std::map<int, int> mappedPntsTmplt2;
	
	for (int i=0; i<2; i++)
	{
		int mid = window->WindowToModel()[i];
		VisModel3D * mvis = window->GetDrawableModels()[mid];
		mvis->Draw(window, mvis->SubwindowCameraID(i, window));
		window->LoadStandardSceneProjectionCamera();
		R3Affine camera = mvis->GetCameraTransform(mvis->SubwindowCameraID(i, window));
		camera.Load();
		static GLfloat material[4];
		assert(mid >= 0 && mid < (int)(*m_pMdl2TmpltMapping).size());
		double rad = 0.05 * mvis->GetMeshRadius();
		bool valid;
		for (int p=0; p<(int)(*m_pMdl2TmpltMapping)[mid]->m_GroundTruthPoint.size(); p++)
		{
			R3Point pos = (*m_pMdl2TmpltMapping)[mid]->GetGroundTruthPoint(p, valid);
			if (valid)
			{
				window->MapIntToRadicalColor(p, material);
				glEnable(GL_LIGHTING);
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
				R3Sphere(pos, rad).Draw();
				window->GetRenderer()->RenderText(pos, VKString("true ")+VKString::number(p),
												  1., 1., 1., VisWindowRenderer::TEXT_SIZE_SMALL);
			}
		}
		
		if (i==0)	// find mapped points
		{
			int c1 = (*m_pMdl2TmpltMapping)[mid]->m_iConfigID;
			//int mid2 = window->GetDrawableModels()[i+1]->GetModel()->ModelID();
			int mid2 = window->WindowToModel()[1];
			int c2 = (*m_pMdl2TmpltMapping)[mid2]->m_iConfigID;
			for (int p=0; p<(int)(*m_pMdl2TmpltMapping)[mid]->m_GroundTruthPoint.size(); p++)
			{
				int opt_t1, opt_t2;
				
				R3Point p2 = MapPoint(p, m_pT2TMapping, m_pMdl2TmpltMapping, mid, mid2,
									  false, valid, &opt_t1, &opt_t2);
				if (valid)
				{
					mappedPntsTmplt1[opt_t1] = p;
					mappedPntsTmplt2[opt_t2] = p;
				}
				mappedPnts.push_back(p2);
				mappedPntValid.push_back(valid);
			}
		}
		
		// draw template
		TMatch * match = (*m_pMdl2TmpltMapping)[mid]->m_pMatch;
		R3Point c = match->GetMatchedGeomAsPnts()->Centroid();
		double aveRad = match->GetMatchedGeomAsPnts()->AverageRadius(c);

		// Draw undeformed template: for debugging
//		for (int p=0; p<(*m_pMdl2TmpltMapping)[mid]->m_pUndeformedTemplatePnts->NumPoints(); p++)
//		{
//			material[0] = .2;
//			material[1] = .5;
//			material[2] = .5;
//			material[3] = .5;
//
//			R3Point pos = (*m_pMdl2TmpltMapping)[mid]->m_pUndeformedTemplatePnts->PointPosition(p);
//			(*m_pMdl2TmpltMapping)[mid]->m_pMatch->GlobalXForm().ApplyInverse(pos);
//			pos = pos*aveRad+c;
//			int gtp = -1;
//			if (i==0 && mappedPntsTmplt1.find(p)!=mappedPntsTmplt1.end())
//				gtp = mappedPntsTmplt1[p];
//			else if (i==1 && mappedPntsTmplt2.find(p)!=mappedPntsTmplt2.end())
//				gtp = mappedPntsTmplt2[p];
//			double rs = 2.;
//			if (gtp>=0)
//			{
//				window->MapIntToRadicalColor(gtp, material);
//				rs = .5;
//			}
//
//			glEnable(GL_LIGHTING);
//			glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
//			R3Sphere(pos, rad/rs).Draw();
//		}
		
		// draw deformed template
		for (int p=0; p<(int)(*m_pMdl2TmpltMapping)[mid]->m_TemplateInlcuded2AllIDMap.size(); p++)
		{
			int oldp = (*m_pMdl2TmpltMapping)[mid]->m_TemplateInlcuded2AllIDMap[p];
			R3Point pos = (*m_pMdl2TmpltMapping)[mid]->m_pMatch->DeformedTemplateSample(oldp);
			
			
			(*m_pMdl2TmpltMapping)[mid]->m_pMatch->GlobalXForm().ApplyInverse(pos);
			pos = pos*aveRad+c;
			
			material[0] = .2;
			material[1] = .5;
			material[2] = .5;
			material[3] = .5;
			
			int gtp = -1;
			if (i==0 && mappedPntsTmplt1.find(p)!=mappedPntsTmplt1.end())
				gtp = mappedPntsTmplt1[p];
			else if (i==1 && mappedPntsTmplt2.find(p)!=mappedPntsTmplt2.end())
				gtp = mappedPntsTmplt2[p];
			if (gtp>=0)
				window->MapIntToRadicalColor(gtp, material);
			glEnable(GL_LIGHTING);
			glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
			R3Sphere(pos, rad/2).Draw();
			if (gtp>=0)
				window->GetRenderer()->RenderText(pos, VKString("tmplt ")+" "+VKString::number(gtp),
												  1., 1., 1., VisWindowRenderer::TEXT_SIZE_SMALL);
		}
		
		if (i==1)	// render mapped points
		{
			for (int p=0; p<(int)mappedPnts.size(); p++)
			{
				if (mappedPntValid[p])
				{
					window->MapIntToRadicalColor(p, material);
					//material[3] = .2;
					glEnable(GL_LIGHTING);
					glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
					R3Sphere(mappedPnts[p], rad*1.5).Draw();
					window->GetRenderer()->RenderText(mappedPnts[p], VKString("predict ")+VKString::number(p),
													  1., 1., 1., VisWindowRenderer::TEXT_SIZE_SMALL);
				}
			}
		}
	}
}

void TMapDebugApp::DrawFuzzy(DatabaseVisWindow * window)
{
	assert(m_pGtPnts!=NULL && m_pFuzzyPnts!=NULL && m_pV2VMap!=NULL);
	
	int nmodels = (*m_pGtPnts).size();
	assert(window->WindowToModel().size()>=2);
	std::vector<R3Point> mappedPnts;
	std::vector<bool> mappedPntValid;
	
	for (int i=0; i<2; i++)
	{
		int mid = window->WindowToModel()[i];
		VisModel3D * mvis = window->GetDrawableModels()[mid];
		mvis->Draw(window, mvis->SubwindowCameraID(i, window));
		window->LoadStandardSceneProjectionCamera();
		R3Affine camera = mvis->GetCameraTransform(mvis->SubwindowCameraID(i, window));
		camera.Load();
		static GLfloat material[4];
		assert(mid >= 0 && mid < (int)(*m_pGtPnts).size());
		double rad = 0.05 * mvis->GetMeshRadius();
		bool valid;
		
		// render ground truth points
		for (int p=0; p<(int)(*m_pGtPnts)[mid].size(); p++)
		{
			valid = (*m_pGtPnts)[mid][p]!=NULL;
			if (valid)
			{
				R3Point pos = (*m_pGtPnts)[mid][p]->GetPosition();
				window->MapIntToRadicalColor(p, material);
				glEnable(GL_LIGHTING);
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
				R3Sphere(pos, rad).Draw();
				window->GetRenderer()->RenderText(pos, VKString("true ")+VKString::number(p),
												  1., 1., 1., VisWindowRenderer::TEXT_SIZE_SMALL);
			}
		}
		
		if (i==0)	// find mapped points
		{
			//int mid2 = window->GetDrawableModels()[i+1]->GetModel()->ModelID();
			int mid2 = window->WindowToModel()[1];
			
			for (int p=0; p<(int)(*m_pGtPnts)[mid].size(); p++)
			{
				// feature point -> template ID
				valid = (*m_pGtPnts)[mid][p]!=NULL;
				
				if (!valid)
				{
					mappedPnts.push_back(R3Point(0,0,0));
					mappedPntValid.push_back(false);
					continue;	// ignore point: no gt
				}
				
				// predict p1 position (p2)
				R3Point p1 = (*m_pGtPnts)[mid][p]->GetPosition();
				const PointInSet * closest = (*m_pFuzzyPnts)[mid]->GetKdTree()->FindClosest(p1);
				assert(closest!=NULL);
//				bool atvtx = false;
//				int vid = (*m_pFuzzyPnts)[mid]->GetPoint(closest->id).VertexID(&atvtx);
//				assert(atvtx);
				
				std::map<int, R3Point>::iterator iter = m_pV2VMap[mid * nmodels + mid2].find(closest->id);
				assert(iter != m_pV2VMap[mid * nmodels + mid2].end());
				R3Point p2 = iter->second;
				//assert((*m_pFuzzyPnts)[mid2]->GetMesh()!=NULL);
				//R3Point p2 = (*m_pFuzzyPnts)[mid2]->GetMesh()->VertexPosition(iter->second);
				//std::cout<<"NVertices: "<<(*m_pFuzzyPnts)[mid2]->GetMesh()->NVertices()<<std::endl;
//				std::cout<<"p: "<<p<<" -> "<<vid<<" -> "<<iter->second;
//				std::cout<<" -> ["<<p2.X()<<" "<<p2.Y()<<" "<<p2.Z()<<"]"<<std::endl;
				mappedPnts.push_back(p2);
				mappedPntValid.push_back(true);
			}
		}

		if (i==1)	// render mapped points
		{
			for (int p=0; p<(int)mappedPnts.size(); p++)
			{
				if (mappedPntValid[p])
				{
					window->MapIntToRadicalColor(p, material);
					//material[3] = .2;
					glEnable(GL_LIGHTING);
					glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
					R3Sphere(mappedPnts[p], rad*1.5).Draw();
					window->GetRenderer()->RenderText(mappedPnts[p], VKString("predict ")+VKString::number(p),
													  1., 1., 1., VisWindowRenderer::TEXT_SIZE_SMALL);
				}
			}
		}
	}
}



