#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"  
#include "BasicNumerical.h"

#include "RendererOSMesa.h"
#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "AppScreenshots.h"
#include "VisShapeTemplate.h"

#include "STemplate.h"
#include "TMatch.h"
#include "TMatchEnergy.h"
#include "TMatchAlg.h"
#include "PartOfEdge.h"
#include "ConsistentPartsAlg.h"
#include "STemplateNode.h"

GeomShape * ReadModelGeometry(ModelDatabase * db, const VKString & mName, const VKString & geomName,
							  bool readManifold, const VKString & geomValue = "none");
ModelDatabase * ReadGeometries(const VKString & dbDir, std::vector<TMatchEnergy*> & matches, bool readManifold);
void ReadAllMatches(const STemplate * st, const VKString & matchListFile, std::vector<TMatchEnergy*> & matches,
					double fracToProcess, int nMinToProcess, int nMaxToProcess, bool getNewOnly,
					const VKString & filename);
bool IncludeNullFromAllMatches(STemplate * st, const std::vector<TMatchEnergy*> & matches, int methodFlags,
							   std::vector<TemplateNodeDeform*> & newNodes);
void PrintPerPartErrs(STemplate * st,  const std::vector<TMatchEnergy*> & matches);
bool ClusterStyles(STemplate * st,  const std::vector<TMatchEnergy*> & matches);

void EvalCorrs(ModelDatabase * db, std::vector<TMatchEnergy*> & matches, const VKString & evalCorrs);

int main(int argc, char** argv )
{
	if (argc<4)
	{
		std::cout<<"Usage:"<<std::endl;
		std::cout<<"   "<<argv[0]<<" [params]"<<std::endl;;
		std::cout<<"Args: "<<std::endl;
		std::cout<<" -readTemplate file.tmplt - read template"<<std::endl;
		std::cout<<" -writeTemplate file.tmplt - write template"<<std::endl;
		std::cout<<" -readMatches matchList.txt - read matches"<<std::endl;
		std::cout<<" -readGeometries - read geometries"<<std::endl;
		std::cout<<" -updateFromAll - update variance on existing shaepes from all matches (need to have a list)"<<std::endl;
		std::cout<<" -writeUpdateLog logFile - "<<std::endl;
		std::cout<<" -readNMatches F%_minN_maxN - Try reading F % best matches (s.t. at least minN at most maxN)"<<std::endl;
		std::cout<<" -readMatchFileName name - file name (no path, just file)"<<std::endl;
		std::cout<<" -reMatch - match again with null nodes"<<std::endl;
		std::cout<<" -dbDir dirName - used if necessary"<<std::endl;
		std::cout<<"Part clustering:"<<std::endl;
		std::cout<<" -clusterStyles - cluster styles (all non-null points)"<<std::endl;
		std::cout<<" -includeNull - add new parts (from null points)"<<std::endl;
		std::cout<<"Part clustering algorithm flags:"<<std::endl;
		std::cout<<" -initWhole - initial clustering of points: take all points"<<std::endl;
		std::cout<<" -initGT - initial clustering of points: take ground truth"<<std::endl;
		std::cout<<" -initCRF - initial clustering of points: use CRF labeling"<<std::endl;
		std::cout<<" -clusterTakeAll - clustering of parts: take all parts"<<std::endl;
		std::cout<<" -clusterPartsSpectralAgglomerative - agglomerative clustering of parts"<<std::endl;
		std::cout<<" -aveClusterTakeAny - producing ave cluster"<<std::endl;
		std::cout<<" -showGT equivFile "<<std::endl;
		std::cout<<" -evalCorrs outFile - evaluate stats for all pairwise correspondences "<<std::endl;
		std::cout<<" -writeClustered modellist.txt - write files used to update template"<<std::endl;
		std::cout<<" -totalShapes N - total number of shapes - affects some thresholds"<<std::endl;
		std::cout<<"Optional:"<<std::endl;
		std::cout<<" -vis - visualize"<<std::endl;
		exit(0);
	}
	
	VKString readMatchFileName = "deformOnly.match.bin";
	
	int totalShapes = -1;	// unknown
	bool visualize = false;
	bool reMatch = false;
	VKString readTemplate = "none";
	VKString writeTemplate = "none";
	VKString readMatches = "none";
	bool updateFromAll = false;
	int nMinToProcess = -1;
	int nMaxToProcess = -1;
	double fracToProcess = 1.;
	
	int clusteringMethod = 0;
	VKString dbDir = "none";
	bool readGeometries = false;
	bool clusterStyles = false;
	bool includeNull = false;
	bool getNewOnly = false;
	VKString showGT ="none";
	VKString evalCorrs = "none";
	bool readManifold = false;
	VKString writeUpdateLog = "none";
	VKString writeClustered = "none";
	
	for (int i=1; i<argc; i++)
	{
		if (strcmp(argv[i], "-vis")==0)
			visualize = true;
		else if (strcmp(argv[i], "-readTemplate")==0)
			readTemplate = argv[++i];
		else if (strcmp(argv[i], "-writeTemplate")==0)
			writeTemplate = argv[++i];
		else if (strcmp(argv[i], "-updateFromAll")==0)
		{
			updateFromAll = true;
			getNewOnly = true;
		}
		else if (strcmp(argv[i], "-writeUpdateLog")==0)
			writeUpdateLog = argv[++i];
		else if (strcmp(argv[i], "-readNMatches")==0)
		{
			VKStringList updParams = VKString((argv[++i])).split("_");
			assert(updParams.count()==3 && updParams[0].endsWith("%"));
			bool ok;
			fracToProcess = updParams[0].replace("%", "").toDouble(&ok)/100.; 	assert(ok);
			nMinToProcess = updParams[1].toInt(&ok);	assert(ok);
			nMaxToProcess = updParams[2].toInt(&ok);	assert(ok);
		}
		else if (strcmp(argv[i], "-clusterStyles")==0)
			clusterStyles = true;
		else if (strcmp(argv[i], "-includeNull")==0)
			includeNull = true ;
		else if (strcmp(argv[i], "-initWhole")==0)
			clusteringMethod |= ConsistentPartsAlg::CPM_InitWhole;
		else if (strcmp(argv[i], "-initGT")==0)
			clusteringMethod |= ConsistentPartsAlg::CPM_InitFromGT;
		else if (strcmp(argv[i], "-initCRF")==0)
			clusteringMethod |= ConsistentPartsAlg::CPM_InitWithMRF_TakeFurthest;
		else if (strcmp(argv[i], "-clusterTakeAll")==0)
			clusteringMethod |= ConsistentPartsAlg::CPM_ClusterTakeAll;
		else if (strcmp(argv[i], "-clusterPartsSpectralAgglomerative")==0)
			clusteringMethod |= ConsistentPartsAlg::CPM_ClusterSpectralAgglomerative;
		else if (strcmp(argv[i], "-aveClusterTakeAny")==0)
			clusteringMethod |= ConsistentPartsAlg::CPM_AveClusterTakeAny;
		else if (strcmp(argv[i], "-readMatches")==0)
			readMatches = argv[++i];
		else if (strcmp(argv[i], "-readMatchFileName")==0)
			readMatchFileName = argv[++i];
		else if (strcmp(argv[i], "-readGeometries")==0)
			readGeometries = true;
		else if (strcmp(argv[i], "-readManifold")==0)
			readManifold = true;
		else if (strcmp(argv[i], "-reMatch")==0)
			reMatch = true;
		else if (strcmp(argv[i], "-dbDir")==0)
			dbDir = argv[++i];
		else if (strcmp(argv[i], "-showGT")==0)
			showGT = argv[++i];
		else if (strcmp(argv[i], "-evalCorrs")==0)
			evalCorrs = argv[++i];
		else if (strcmp(argv[i], "-writeClustered")==0)
			writeClustered = argv[++i];
		else if (strcmp(argv[i], "-totalShapes")==0)
			totalShapes = atoi(argv[++i]);
		else
			std::cout<<"[WARNING] Unknown arg: "<<argv[i]<<" in UpdateTemplate.cpp"<<std::endl;
	}
	
	TimeProfiler timer;
	timer.startedProcess("UpdateTmplt");
	ModelDatabase * db = NULL;	// for now don't need actual models

	STemplate * st = NULL;
	if (readTemplate!="none")
	{
		std::cout<<"Reading template"<<std::endl;
		st = STemplate::ReadTemplate(readTemplate, db);
	}
	
	std::vector<TMatchEnergy*> matches;
	if (readMatches!="none")
	{
		std::cout<<"Reading matches"<<std::endl;
		ReadAllMatches(st, readMatches, matches, fracToProcess, nMinToProcess, nMaxToProcess, getNewOnly, readMatchFileName);
		if (matches.size()==0)
		{
			std::cout<<"[WARNING] No matches to process. Quitting..."<<std::endl;
			if (writeTemplate!="none" && st!=NULL)
			{
				FileManagement::CreateDirsOnPathToFile(writeTemplate.c_str());
				st->WriteTemplate(writeTemplate);
			}
			exit(0);
		}
	}
	
	if (readGeometries)
		db = ReadGeometries(dbDir, matches, readManifold);
	
	bool templateUpdated = false;
	CoAlignedParts * coAlignedParts = NULL;
	if (matches.size()>0 && st!=NULL)
		coAlignedParts = new CoAlignedParts(st, matches);
	VKString updateString = "";
	
	
	if (updateFromAll)
	{
		std::cout<<"Update From All Matches"<<std::endl;
		updateString = coAlignedParts->UpdateTemplateFromAllMatches(writeUpdateLog, totalShapes);
		templateUpdated = true;
	}
	
	if (clusterStyles)
	{
		std::cout<<"Cluster styles"<<std::endl;
		bool upd = ClusterStyles(st, matches);
		templateUpdated = templateUpdated || upd;
	}
	
	if (includeNull!=0)
	{
		assert(false);
		std::cout<<"Update From Null"<<std::endl;
		assert(dbDir!="none");
		
		std::vector<TemplateNodeDeform*> newNodes;
		bool upd = IncludeNullFromAllMatches(st, matches, clusteringMethod, newNodes);

		std::vector<TMatchEnergy*> newMatches;
		if (reMatch && upd)
		{
			// re-match
			std::cout<<"Re-matching..."<<std::endl;
			assert(matches.size()>0);
			STemplate * st = (STemplate *)matches[0]->GetMatch()->GetSTemplate();
			
			for (int i=0; i<(int)matches.size(); i++)
			{
				TMatch * copy = matches[i]->GetMatch()->CreateCopy(newNodes);
				TMatchAlgParams params;
					// cheap optimization for DEBUGGING
				//params.m_iOptimFlags = TMatchAlgParams::OPT_Pnts | TMatchAlgParams::OPT_PntsEdges;
				newMatches.push_back(TMatchAlg::OptimizeConfigured(st, copy, params));
			}
			
			std::cout<<"Remove existing"<<std::endl;
			// remove existing matches (if template contains them)
			for (int i=0; i<(int)matches.size(); i++)
			{
				TMatch * match = matches[i]->GetMatch();
				if (st->GetCreatedFromModelsList().contains(match->ModelName()))
					st->UpdateAllStats(match, false);
			}

			// add updated matches
			for (int i=0; i<(int)newMatches.size(); i++)
				st->UpdateAllStats(newMatches[i]->GetMatch(), true);

		}
		else
		{
			for (int i=0; i<(int)matches.size(); i++)
				newMatches.push_back(new TMatchEnergy(matches[i]->GetMatch()->CreateCopy(newNodes)));
		}
			
		for (int i=0; i<(int)matches.size(); i++)
			delete matches[i];
		matches.clear();
		matches = newMatches;
		templateUpdated = templateUpdated || upd;
	}
	
	if (writeTemplate!="none" && st!=NULL)
	{
		std::cout<<"Writing Template: "<<writeTemplate.c_str()<<std::endl;
		FileManagement::CreateDirsOnPathToFile(writeTemplate.c_str());
		std::ofstream textStream((writeTemplate+".log.txt").c_str());
		assert(textStream.is_open());
		textStream<<"Previous matches: [";
		for (int j=0; j<st->GetCreatedFromModelsList().count(); j++)
			textStream<<st->GetCreatedFromModelsList()[j].c_str()<<" ";
		textStream<<"]\n";
		textStream<<"Learned from matches: [";
		for (int j=0; j<(int)matches.size(); j++)
			textStream<<matches[j]->GetMatch()->ModelName().c_str()<<" ";
		textStream<<"]\n";
		textStream<<"Writing Template: "<<writeTemplate.c_str()<<std::endl;
		textStream.close();
		st->WriteTemplate(writeTemplate);
	}
	
	if (writeClustered!="none" && st!=NULL)
	{
		const VKStringList & mdls = st->GetCreatedFromModelsList();
		std::ofstream clusteredFile(writeClustered.c_str());
		assert(clusteredFile.is_open());
		for (int i=0; i<mdls.count(); i++)
			clusteredFile<<mdls[i].c_str()<<"\n";
		clusteredFile.close();
		
		if (updateString!="")	// some updates
		{
			// write update string
			VKString updFileName = writeClustered;
			updFileName = updFileName.replace(".txt", "_updSeeds.txt");
			std::ofstream updFileStream(updFileName.c_str());
			assert(updFileStream.is_open());
			updFileStream<<updateString.c_str();
			updFileStream.close();
		}
	}
	timer.finishedProcess("UpdateTmplt");
	
	timer.startedProcess("EvalCorrs");
	if (evalCorrs!="none")
	{
		EvalCorrs(db, matches, evalCorrs);
	}
	timer.finishedProcess("EvalCorrs");
	
	if (writeUpdateLog!="none")// also write time
	{
		VKString timingFName = writeUpdateLog;
		timingFName = timingFName.replace(".txt", ".time.txt");
		std::ofstream textStream(timingFName.c_str());
		assert(textStream.is_open());
		textStream<<timer.profile().c_str();
		textStream.close();
	}
	
	if (visualize)
	{	
		if (st==NULL)
		{
			std::cout<<"[WARNING] No template. Nothing to visualize"<<std::endl;
			exit(0);
		}
#ifdef VISUALIZATION		
		int listFlags = DatabaseVisWindow::LIST_GROUPS_CORRS;
		listFlags |= DatabaseVisWindow::LIST_GROUPS_MODELS;
		
		if (db==NULL)
			db = new ModelDatabase();
		int nWnd = 1;
		//DatabaseVisWindow * dbVis = new DatabaseVisWindow(db, NULL, listFlags, nWnd, 1, 0, 0, 1);
		DatabaseVisWindow * dbVis = new DatabaseVisWindow(db, NULL, listFlags, nWnd, 1, 2, 5, .8);

		VisShapeTemplate * visTemplate = new VisShapeTemplate(db);
		for (int i=0; i<(int)matches.size(); i++)
		{
			TMatch * matched = matches[i]->GetMatch();
			const STemplate * st = matched->GetSTemplate();
			TMatch * seed = st->CreateSeedConfig(matched->SeedID(), NULL, std::vector<int>(), R3identity_affine);
			visTemplate->AddMatchedModel(seed, matches[i]);
		}
		
		visTemplate->SetGTEval(ShapeValue::m_sDefaultSceneGraphLabels, showGT);
		
		if (coAlignedParts!=NULL)
			visTemplate->SetCoAlignedPartsHolder(coAlignedParts);
				
		int flags = (VisShapeTemplate::VISTMPLT_SeedTemplate 
					 | VisShapeTemplate::VISTMPLT_ShowAllMatchesStats
					 | VisShapeTemplate::VISTMPLT_DrawTemplateAsBoxes 
					 | VisShapeTemplate::VISTMPLT_DrawTemplateAsAxes);
		
		visTemplate->SetRenderingFlags(flags);
		dbVis->SetApp(visTemplate);
		
#	ifdef MESA_RENDERING
		std::cout<<"[WARNING] Mesa renderer"<<std::endl;
		VisWindowRenderer * wndMesa = new RendererOSMesa(dbVis);
		wndMesa->UpdateScreen();
#	else
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
#	endif
#else
		std::cout<<"[WARNING] Visualization disabled (compiling flag)"<<std::endl;
#endif
	}
	
	
	return 0;
}


void ReadAllMatches(const STemplate * st, const VKString & matchListFile, std::vector<TMatchEnergy*> & matches,
					double fracToProcess, int nMinToProcess, int nMaxToProcess, bool getNewOnly,
					const VKString & readMatchFileName)
{
	assert(st!=NULL);
	int lastSlash = matchListFile.lastIndexOf("/");
	assert(lastSlash>0);
	VKString dirOnly = matchListFile.left(lastSlash)+"/";
	
	std::ifstream textStream(matchListFile.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Cannot read matches: "<<matchListFile.c_str()<<std::endl;
	assert(textStream.is_open());
	VKStringList toRead;
	int nFiles = 0;
	std::cout<<"Reading models. Previously visited: [";
	const VKStringList & list = st->GetCreatedFromModelsList();
	for (int i=0; i<list.count(); i++)
		std::cout<<list[i].c_str()<<" ";
	std::cout<<"]"<<std::endl;
	
	while (!textStream.eof())
	{
		VKString matchFile;
		matchFile.readLine(textStream);
		VKStringList matchesLine = matchFile.split(" ");
		if (matchesLine.count()>0)
			matchFile = dirOnly + matchesLine[0]+"/"+readMatchFileName;
		else
			continue;
		nFiles++;
		
		if (getNewOnly && st->GetCreatedFromModelsList().contains(matchesLine[0]))
			continue;

		//if (toRead.count() >= nMaxToProcess && nMaxToProcess>=0)
		//	continue;
		
		toRead.push_back(matchFile);
	}
	textStream.close();
	
	std::cout<<"Read matches: "<<toRead.count()<<std::endl;
	
	int nToRead = fracToProcess * nFiles;
	if (nMinToProcess>=0)
		nToRead = vkMax(nToRead, nMinToProcess);
	if (nMaxToProcess>=0)
		nToRead = vkMin(nToRead, nMaxToProcess);
	
	nToRead = vkMin(nToRead, toRead.count());
		
	for (int i=0; i<nToRead; i++)
	{
		if (FileManagement::FileExists(toRead[i].c_str()) && toRead[i]!="")
		{
			FILE * file = fopen(toRead[i].c_str(), "rb");
			assert(file!=NULL);
			matches.push_back(TMatchEnergy::ReadEnergyAndMatch(file, st, NULL));
			fclose(file);
		}
		else
			std::cout<<"[WARNING] Could not find match: "<<toRead[i].c_str()<<std::endl;
	}
}



bool IncludeNullFromAllMatches(STemplate * st, const std::vector<TMatchEnergy*> & matches, int methodFlags,
							   std::vector<TemplateNodeDeform*> & newNodes)
{
	assert(false);
//	assert(matches.size()>0);
//	int commonSeedID = matches[0]->GetMatch()->SeedID(); 	// need to assert all seed id
//	for (int i=0; i<(int)matches.size(); i++)
//		assert(commonSeedID==matches[i]->GetMatch()->SeedID());
//	
//	int nullID = -1;
//	for (int i = 0; i<st->NNodes(); i++)
//		if (st->GetNode(i)->GetPart()->IsType(PartBox::BOXTYPE_NullBox))
//			nullID = i;
//	assert(nullID>=0);
//	
//	std::cout<<"Include Null Nodes"<<std::endl;
//	ConsistentPartsAlg cpAlg(st, matches, methodFlags, false);
//	std::vector<PointsCluster> finalClusters;
//	cpAlg.ExtractConsistentParts(nullID, 0, 1, finalClusters);
//	std::cout<<"Final clusters: "<<finalClusters.size()<<std::endl;
//	
//	// add extracted primitives to the template 
//	for (int i=0; i<(int)finalClusters.size(); i++)
//	{
//		assert(finalClusters[i].m_pGeomFit!=NULL);	// todo..
//		st->InitAndAddNewNode(0, finalClusters[i].m_sLabel, *(finalClusters[i].m_pGeomFit), commonSeedID, &newNodes);
//		STemplateNode * node = (STemplateNode*)st->GetNode(newNodes[(int)newNodes.size()-1]->NodeID());
//		node->PartOfParent()->UpdateNParts(0, 0, true);
//	}
//	return true;
}

GeomShape * ReadModelGeometry(ModelDatabase * db, const VKString & mName,
							  const VKString & geomName, bool readManifold, const VKString & geomValue)
{
	assert(db!=NULL);
	Model3D * m = db->GetModel(mName);
	if (m==NULL)
		std::cout<<"[ERROR] Cannot find model: "<<mName.c_str()<<std::endl;
	assert(m!=NULL);
	VKString geomFile = db->GetModelGeometryFilename(m->ModelName(), geomName);
	VKString mDir = db->GetDBDirname()+"/Models/";
	
	if (m->GetGeomShape(geomName)==NULL)
	{
		if (geomName.endsWith(".pnts"))
		{
			assert(geomName.startsWith(Model3D::m_sMainGeomName));
			ReadModelGeometry(db, mName, Model3D::m_sMainGeomName, readManifold);
			m->ReadGeometry(GeomShape::SHAPE_TYPE_POINT_CLOUD, geomName, geomFile, Model3D::m_sMainGeomName);
		}
		else
		{
			VKString geomFile = db->GetModelGeometryFilename(m->ModelName(), geomName)+".polysoup.off";
			GeomShape::GeomShapeTypes mt = GeomShape::SHAPE_TYPE_TRIANGLE_SOUP;
			if (readManifold)
				mt = GeomShape::SHAPE_TYPE_MANIFOLD;
			
			m->ReadGeometry(mt, geomName, geomFile);
		}
	}
	GeomShape * shape = m->GetGeomShape(geomName);
	assert(shape!=NULL);
	if (geomValue!="none")
	{
		VKString shapeValFile = mDir+db->ShapeValueFilename(m->ModelName(), geomName, geomValue);
		if (shape->GetShapeVal(geomValue)==NULL)
			m->ReadGeomValue("ShapeLabels", shape, geomValue, shapeValFile);
	}
	return shape;
}

ModelDatabase *  ReadGeometries(const VKString & dbDir, std::vector<TMatchEnergy*> & matches, bool readManifold)
{
	VKStringList modelsToLoad;
	for (int i=0; i<(int)matches.size(); i++)
		modelsToLoad.push_back(matches[i]->GetMatch()->ModelName());
	ModelDatabase * db = ModelDatabase::CreateDBFromDir(dbDir, ModelDatabase::DBINPUTDIR_MyFormat, modelsToLoad, 0, 0, dbDir);
	
	TimeProfiler timer;

	std::cout<<"Reading geomtries:"<<std::endl;
	for (int i=0; i<(int)matches.size(); i++)
	{
		timer.WriteProgress("prep", i, (int)matches.size());
		
		TMatch * match = matches[i]->GetMatch();
		VKString valsName = ShapeValue::m_sDefaultSceneGraphLabels;
		VKString valsFile = db->GetModelDirname(match->ModelName()) + "/Geometry/mainVals/"+valsName+".vals";
		if (!FileManagement::FileExists(valsFile.c_str()))
			valsName = "none";
			
		GeomShape * mainMesh = ReadModelGeometry(db, match->ModelName(), Model3D::m_sMainGeomName, readManifold, valsName);
		assert(mainMesh!=NULL);
		GeomShape * pnts = ReadModelGeometry(db, match->ModelName(), match->GeomName(), readManifold);
		assert(pnts!=NULL);
		
		if (valsName!="none")
		{
			ShapeValue * vals = mainMesh->GetShapeVal(ShapeValue::m_sDefaultSceneGraphLabels);
			assert(vals!=NULL && vals->SafeCastToLabels()!=NULL);
			ShapeLabels * labels = ShapeLabels::CopyLabels(vals->SafeCastToLabels(), pnts);
			pnts->AddShapeVal(ShapeValue::m_sDefaultSceneGraphLabels, labels);
		}
		match->ReadGeometry(db);
		match->CreateTemplateSamples();
	}
	std::cout<<std::endl;
	
	return db;
}

void PrintPerPartErrs(STemplate * st,  const std::vector<TMatchEnergy*> & matches)
{
	std::vector<double> avePerPartErr;
	std::vector<int> nParts;
	
	for (int i=0; i<st->NNodes(); i++)
	{
		avePerPartErr.push_back(0);
		nParts.push_back(0);
	}
	
	for (int i=0; i<(int)matches.size(); i++)
	{
		TMatchEnergy * energy = matches[i];
		TMatch * match = matches[i]->GetMatch();
		
		int currTerms = energy->GetTerms();
		//TMatchEnergy::ET_PntEdges |  - currently irrelevant. In Future: part might also be aware of it's
		// typical shape of a cut and penalize accordingly...
		int terms = TMatchEnergy::ET_Pnts | TMatchEnergy::ET_NodeFit2Pnts;
		terms = terms | (currTerms & (TMatchEnergy::ET_PntsDistance | TMatchEnergy::ET_PntsOrientation));
		energy->SetTerms(terms);
		energy->ClearCache();
		
		int NPnts = match->NShapePrimitives();
		std::vector<int> nPerPartPnts;
		for (int j=0; j<match->NTemplateNodes(); j++)
			nPerPartPnts.push_back(0);
		
		for (int p=0; p<match->NShapePrimitives(); p++)
			nPerPartPnts[match->LocalNodeID(match->Loc2GlobPid(p))]++;
		
		std::cout<<"Model: "<<match->ModelName().c_str()<<std::endl;
		for (int j=0; j<match->NTemplateNodes(); j++)
		{
			const TemplateNodeDeform & deform = match->Deformation(j);
			const STemplateNode * node = st->GetNode(deform.NodeID());
			VKString nStr = node->NodeName();
			while(nStr.length() < 10)
				nStr += " ";
			double ne = energy->NodeEnergy(j);
			if (ne>0)
				ne /= nPerPartPnts[j];
			std::cout<<"    "<<deform.NodeID()<<". "<<nStr.c_str()<<": "<<ne<<std::endl;
			
			assert(deform.NodeID()>=0 && deform.NodeID() < (int)avePerPartErr.size());
			assert(deform.NodeID() < (int)nParts.size());
			avePerPartErr[deform.NodeID()] += ne;
			nParts[deform.NodeID()]++;
		}
		energy->SetTerms(currTerms);
	}
	
	std::cout<<"Average Energy Vals:"<<std::endl;
	for (int i=0; i<st->NNodes(); i++)
	{
		if (nParts[i]==0 || st->GetNode(i)->GetPart(0)->IsType(PartBox::BOXTYPE_NullBox))
			continue;
		
		VKString nStr = st->GetNode(i)->NodeName();
		while(nStr.length() < 10)
			nStr += " ";
		
		VKString errStr = VKString::number(avePerPartErr[i] / nParts[i]);
		while(errStr.length() < 10)
			errStr += " ";
		
		std::cout<<"   "<<i<<". "<<nStr.c_str()<<": "<<errStr.c_str()<<" x "<<nParts[i]<<std::endl;
	}
}

bool ClusterStyles(STemplate * st,  const std::vector<TMatchEnergy*> & matches)
{
	int methodFlags = ConsistentPartsAlg::CPM_InitWhole | ConsistentPartsAlg::CPM_ClusterSpectralAgglomerative;
	methodFlags |= ConsistentPartsAlg::CPM_AveClusterTakeAny;
	
	ConsistentPartsAlg consistentParts(st, matches, methodFlags, true);
	std::vector<PointsCluster> finalClusters;
	int nodeID = 2;
	consistentParts.ExtractConsistentParts(nodeID, 0, 1, finalClusters);
	std::cout<<"NodeName = "<<st->GetNode(nodeID)->NodeName().c_str()<<std::endl;
	std::cout<<"Final clusters: "<<finalClusters.size()<<std::endl;
	
//	// add extracted primitives to the template
//	for (int i=0; i<(int)finalClusters.size(); i++)
//	{
//		std::cout<<"Label: "<<finalClusters[i].m_sLabel.c_str()<<std::endl;;
//		st->InitAndAddNewNode(0, finalClusters[i].m_sLabel, *(finalClusters[i].m_pGeomFit), commonSeedID, &newNodes);
//		STemplateNode * node = (STemplateNode*)st->GetNode(newNodes[(int)newNodes.size()-1]->NodeID());
//		node->PartOfParent()->UpdateNParts(0, 0, true);
//	}
	return true;
}

void EvalCorrs(ModelDatabase * db, std::vector<TMatchEnergy*> & matches, const VKString & evalCorrs)
{
	std::cout<<"Evaluating correspondences:"<<matches.size()<<std::endl;
	assert(matches.size()>0);
	//CoAlignedParts coAligned((STemplate*)matches[0]->GetMatch()->GetSTemplate(), matches);
	
	// read gt points / find projections to canonical space
	std::vector<ShapePointCloud * > gtPntSets;
	std::vector<std::vector<int> > gtNearestMatchedPnt;
	std::vector<ShapePointCloud * > projectPntSets;
	std::vector<const ShapePointCloud * > unprojectPntSets;
	std::vector<double> distNorm;
	TimeProfiler timer;
	for (int i=0; i<(int)matches.size(); i++)
	{
		timer.WriteProgress("prep", i, (int)matches.size());
		TMatch * match = matches[i]->GetMatch();
		Model3D * model = db->GetModel(match->ModelName());
		assert(model!=NULL);
		
		// read ground truth feature points
		GeomShape * shape = ReadModelGeometry(db, model->ModelName(), "main_featpnts.pnts", false);
		assert(shape!=NULL && shape->SafeCastToPointCloud()!=NULL);
		ShapePointCloud * featPnts = shape->SafeCastToPointCloud();
		gtPntSets.push_back(featPnts);
		
		// for each ground truth point - find nearest point in matches
		const ShapePointCloud * matchedPnts = match->GetMatchedGeomAsPnts();
		unprojectPntSets.push_back(matchedPnts);
		const R3Kdtree<const PointInSet *> * kdtree = matchedPnts->GetKdTree();
		gtNearestMatchedPnt.push_back(std::vector<int>());
		for (int p=0; p<featPnts->NumPoints(); p++)
		{
			const PointInSet * closest = kdtree->FindClosest(featPnts->PointPosition(p), 0, 10E50);
			assert(closest!=NULL);
			gtNearestMatchedPnt[i].push_back(closest->id);
		}
		
		// project each model to canonical space
		GeomShape * meshShape = model->GetGeomShape(Model3D::m_sMainGeomName);
		assert(meshShape!=NULL && meshShape->SafeCastToMesh()!=NULL);
		ShapePointCloud * alignedPnts = new ShapePointCloud(meshShape->SafeCastToMesh(), "alignedpnts");
		for (int p=0; p<matchedPnts->NumPoints(); p++)
		{
			R3Point pos = match->AlignedToNodePointPosition(match->LocalNodeID(p), p, 0);
			alignedPnts->AddPoint(pos);
		}
		projectPntSets.push_back(alignedPnts);
		
		// distance normalization
		//distNorm.push_back(matchedPnts->AverageRadius(matchedPnts->Centroid()));
		distNorm.push_back(0);
		for (int p1=0; p1<matchedPnts->NumPoints(); p1++)
		for (int p2=p1+1; p2<matchedPnts->NumPoints(); p2++)
			distNorm[i] = vkMax(distNorm[i], R3Distance(matchedPnts->PointPosition(p1), matchedPnts->PointPosition(p2)));
	}
	std::cout<<std::endl;
	
	std::vector<double> errors;
	// measure error for every pair
	for (int i=0; i<(int)matches.size(); i++)
	for (int j=0; j<(int)matches.size(); j++)
	{
		if (i==j)
			continue;
		if (j==0)
			timer.WriteProgress("errs", i, (int)matches.size());
		
		const R3Kdtree<const PointInSet *> * kdtreemj = projectPntSets[j]->GetKdTree();
		
		// go over all feature points on i
		for (int p=0; p < (int)gtNearestMatchedPnt[i].size(); p++)
		{
			// find nearest neighbor
			R3Point alignedMi = projectPntSets[i]->PointPosition(gtNearestMatchedPnt[i][p]);
			const PointInSet * closest = kdtreemj->FindClosest(alignedMi);
			assert(closest!=NULL);
			R3Point predictedPos = unprojectPntSets[j]->PointPosition(closest->id);
			R3Point gtPos = gtPntSets[j]->PointPosition(p);
			errors.push_back(R3Distance(gtPos, predictedPos) / distNorm[j]);
		}
	}
	std::cout<<std::endl;
	
	// generate errors file
//	std::cout<<"Eval Corrs: "<<evalCorrs.c_str()<<std::endl;
//	std::ofstream textStream(evalCorrs.c_str());
//	assert(textStream.is_open());
//	for (int i=0; i<(int)errors.size(); i++)
//		textStream<<errors[i]<<" ";
//	textStream.close();
	
	// generate matlab file?
	std::sort(errors.begin(), errors.end());
	int npercent = 200;
	double maxErr = .25;
	double step = maxErr / npercent;
	std::vector<double> x;
	std::vector<double> y;
	int currErr = 0;
	for (int i=0; i<npercent; i++)
	{
		x.push_back(i * maxErr / npercent);
		while (currErr < (int)errors.size() && errors[currErr] < x[i])
			currErr++;
		y.push_back((double)currErr / (double)((int)errors.size()-1));
	}
	
	VKString evalCorrsPerPoint = evalCorrs;
	evalCorrsPerPoint = evalCorrsPerPoint.replace(".txt", ".perpoint.txt");
	std::ofstream textStream(evalCorrs.c_str());
	assert(textStream.is_open());
	for (int i=0; i<npercent; i++)
		textStream<<y[i]<<" ";
	textStream.close();

//	evalCorrsMFile = evalCorrsMFile.replace(".txt", "")+".m";
//	std::ofstream textStreamMFile(evalCorrsMFile.c_str());
//	assert(textStreamMFile.is_open());
//	
//	textStreamMFile<<"X=[";
//	for (int i=0; i<(int)x.size(); i++)
//		textStreamMFile<<x[i]<<" ";
//	textStreamMFile<<"];\n";
//	
//	textStreamMFile<<"Your=[";
//	for (int i=0; i<(int)y.size(); i++)
//		textStreamMFile<<y[i]<<" ";
//	textStreamMFile<<"];\n";
//	
//	bool addCompare = false;
//	if (evalCorrs.lastIndexOf("FC_Boeing86")!=-1)
//	{
//		textStreamMFile<<"Yfuzzy = [ 0 0.0301642 0.0317829 0.0401961 0.0435021 0.05456 0.0614455 0.0925217 0.102189 0.117214 0.137574 0.14902 0.157182 0.17207 0.181623 0.192043 0.215686 0.230255 0.245326 0.261491 0.273871 0.283698 0.300388 0.308208 0.322959 0.335727 0.346398 0.353511 0.361537 0.372754 0.380529 0.390743 0.40668 0.420178 0.430483 0.440698 0.452941 0.471272 0.484838 0.50212 0.517259 0.528659 0.538258 0.547104 0.562905 0.574897 0.58404 0.594619 0.604492 0.614865 0.628956 0.640538 0.651163 0.671204 0.680848 0.691655 0.701049 0.71083 0.721774 0.726813 0.731965 0.7399 0.750068 0.756726 0.765367 0.773165 0.780278 0.787756 0.794802 0.801756 0.806863 0.810898 0.819745 0.823461 0.828979 0.834724 0.84129 0.847036 0.851847 0.855039 0.85839 0.863019 0.867989 0.870907 0.874122 0.880643 0.883903 0.887232 0.889922 0.891724 0.896284 0.900023 0.902508 0.905016 0.908983 0.912266 0.914934 0.917373 0.919061 0.920885 0.922914 0.925513 0.927132 0.929229 0.931874 0.934245 0.938532 0.939512 0.940356 0.941655 0.943593 0.944277 0.945486 0.947241 0.948062 0.950091 0.950935 0.95155 0.952554 0.952896 0.953602 0.954036 0.954514 0.955221 0.955586 0.956224 0.95643 0.956886 0.957228 0.957706 0.958345 0.958504 0.959667 0.960192 0.961149 0.961628 0.961856 0.962312 0.963794 0.963953 0.964113 0.964318 0.964888 0.965048 0.966142 0.966233 0.96637 0.966621 0.966735 0.966849 0.967009 0.967442 0.967533 0.967693 0.967921 0.968878 0.969015 0.969243 0.969425 0.969608 0.96979 0.969859 0.970018 0.970588 0.970679 0.970862 0.970885 0.970999 0.971181 0.971363 0.971477 0.971569 0.972845 0.972914 0.972937 0.972982 0.973028 0.973096 0.973142 0.973256 0.973301 0.973347 0.973461 0.974578 0.974647 0.974692 0.974692 0.974761 0.974852 0.974875 0.97492 0.974943 0.974966 0.974966 0.975057 0.975103 0.975171 0.975171 0.975194 0.975194 ];\n";
//		//textStreamMFile<<"Ygt = [0.089674 0.092391 0.092391 0.10054 0.10598 0.11141 0.11685 0.125 0.13043 0.13859 0.14674 0.15489 0.16304 0.17391 0.17935 0.19022 0.19565 0.20652 0.20924 0.22011 0.22826 0.23641 0.24457 0.25815 0.26359 0.27174 0.28261 0.29348 0.30707 0.31793 0.33152 0.33424 0.34239 0.35054 0.36413 0.37228 0.38043 0.38587 0.3913 0.39946 0.41033 0.41576 0.42391 0.42935 0.4375 0.44565 0.4538 0.46467 0.47283 0.4837 0.49185 0.5 0.5163 0.52174 0.52989 0.54076 0.54891 0.55978 0.56793 0.57609 0.58424 0.59511 0.60326 0.61141 0.62228 0.62772 0.63587 0.64674 0.65489 0.66304 0.66848 0.67663 0.68478 0.69293 0.69837 0.7038 0.71196 0.71739 0.72554 0.73098 0.73913 0.74457 0.75 0.75543 0.76359 0.76902 0.77446 0.77717 0.78261 0.78804 0.79348 0.79891 0.80435 0.80978 0.8125 0.81793 0.82337 0.83152 0.83424 0.83967 0.83967 0.84239 0.84783 0.85054 0.85598 0.8587 0.86413 0.86957 0.875 0.875 0.87772 0.88043 0.88315 0.88587 0.88859 0.8913 0.89402 0.89674 0.90217 0.90489 0.90761 0.91033 0.91033 0.91304 0.91576 0.91848 0.9212 0.92391 0.92663 0.92663 0.92935 0.92935 0.93207 0.93478 0.93478 0.9375 0.94022 0.94293 0.94293 0.94565 0.94565 0.94837 0.95109 0.95109 0.9538 0.9538 0.95652 0.95652 0.95924 0.95924 0.96196 0.96196 0.96467 0.96467 0.96739 0.96739 0.97011 0.97011 0.97283 0.97283 0.97554 0.97554 0.97826 0.97826 0.98098 0.98098 0.9837 0.9837 0.98641 0.98641];";
//		addCompare = true;
//	}
//	else if (evalCorrs.lastIndexOf("FC_Chair111")!=-1)
//	{
//		textStreamMFile<<"Yfuzzy = [ 0 0.0741769 0.0775758 0.0799427 0.0852826 0.0894922 0.0946192 0.100016 0.105708 0.110115 0.115291 0.123677 0.132039 0.140197 0.148149 0.15285 0.161826 0.167273 0.176806 0.179287 0.187068 0.194889 0.200999 0.206871 0.216183 0.221081 0.227813 0.238468 0.247191 0.260303 0.271016 0.282662 0.287002 0.294644 0.300663 0.31181 0.321024 0.328362 0.332809 0.337952 0.346069 0.356355 0.360172 0.368313 0.37344 0.379206 0.388747 0.394955 0.403857 0.411335 0.420377 0.426478 0.433202 0.446372 0.451949 0.458313 0.468632 0.475586 0.485676 0.491859 0.49855 0.506544 0.519615 0.52679 0.53593 0.545905 0.552711 0.559894 0.5714 0.57959 0.586069 0.595528 0.601163 0.610393 0.619099 0.623358 0.629419 0.636462 0.642703 0.648567 0.653759 0.662342 0.667281 0.671671 0.676609 0.686765 0.693669 0.699001 0.703194 0.70905 0.714029 0.72009 0.725266 0.730352 0.735168 0.738608 0.743301 0.749517 0.75493 0.760287 0.76439 0.768092 0.770655 0.774783 0.77855 0.781032 0.785422 0.790188 0.794586 0.796773 0.80068 0.80362 0.808264 0.810008 0.81421 0.816863 0.820115 0.824488 0.827699 0.830868 0.834259 0.8376 0.840049 0.842957 0.845872 0.848911 0.851057 0.853661 0.85561 0.858256 0.860622 0.862039 0.86457 0.86638 0.868911 0.871753 0.874963 0.876634 0.878141 0.881335 0.884087 0.886151 0.887723 0.889361 0.892048 0.893194 0.895283 0.898444 0.899844 0.901212 0.903292 0.905053 0.907641 0.908755 0.909885 0.910934 0.912301 0.914038 0.915061 0.916143 0.916626 0.918043 0.920393 0.92113 0.922179 0.922957 0.924505 0.925266 0.927019 0.927944 0.928591 0.92955 0.930057 0.931753 0.932375 0.933104 0.933735 0.934398 0.935561 0.936675 0.938018 0.938362 0.938796 0.939468 0.940442 0.940901 0.941237 0.941843 0.942236 0.942465 0.942727 0.943104 0.944644 0.94516 0.945561 0.946036 0.947518 0.947781 0.948935 0.949517];\n";
//		//textStreamMFile<<"Ygt = [ 0 0.0848731 0.0872973 0.0897707 0.096724 0.101794 0.106675 0.113022 0.121261 0.126519 0.133587 0.143628 0.15208 0.160426 0.17086 0.176863 0.187543 0.19217 0.202801 0.205258 0.217109 0.225676 0.233915 0.241409 0.253579 0.260737 0.269075 0.280401 0.290262 0.304218 0.314505 0.328542 0.332048 0.339959 0.346658 0.359877 0.369746 0.376994 0.381835 0.387412 0.396675 0.406388 0.411916 0.420999 0.426675 0.434365 0.444021 0.4505 0.460516 0.469992 0.480557 0.488477 0.4962 0.512989 0.518575 0.526339 0.537993 0.545618 0.556495 0.565708 0.57294 0.580762 0.593595 0.600729 0.608468 0.618919 0.625905 0.633464 0.643497 0.653628 0.659664 0.667273 0.673153 0.682555 0.690156 0.69507 0.701884 0.710704 0.716052 0.722678 0.728681 0.73751 0.743579 0.747002 0.752719 0.760344 0.766323 0.771261 0.775799 0.781687 0.786265 0.792072 0.798165 0.803219 0.80665 0.810811 0.817158 0.822809 0.82837 0.833333 0.837625 0.841654 0.845938 0.849631 0.85412 0.857338 0.862604 0.867477 0.872367 0.874496 0.877076 0.878362 0.881867 0.88303 0.886495 0.889312 0.892916 0.896192 0.899189 0.902056 0.904996 0.907412 0.909296 0.912711 0.915061 0.917477 0.919722 0.921966 0.923792 0.925217 0.927117 0.928583 0.930729 0.932383 0.933817 0.935643 0.938231 0.940516 0.94181 0.943694 0.945152 0.946634 0.948526 0.949762 0.951769 0.95258 0.953849 0.955225 0.956224 0.956921 0.958018 0.959222 0.961229 0.961826 0.962531 0.963186 0.963759 0.964889 0.965389 0.965815 0.966593 0.967232 0.968165 0.969394 0.970393 0.970721 0.971884 0.972219 0.972785 0.97344 0.974218 0.97484 0.975094 0.975708 0.9762 0.976593 0.976888 0.977764 0.978591 0.979099 0.979894 0.980074 0.980377 0.980778 0.981851 0.982064 0.982285 0.982629 0.982858 0.982998 0.983268 0.983964 0.98439 0.984595 0.984758 0.985192 0.985586 0.985799 0.985962 0.986323 ];";
//		addCompare = true;
//	}
//	
//	textStreamMFile<<"set(0,'defaultaxesfontsize',15);\n";
//	textStreamMFile<<"set(0,'defaultaxesfontname','Tahoma');\n";
//	textStreamMFile<<"set(0,'defaultaxesfontweight','bold');\n";
//	textStreamMFile<<"figT1 = figure('Position', [0 0 825 450]);\n";
//	textStreamMFile<<"plot(X, Your*100, 'r', 'LineWidth', 3), hold all;\n";
//	
//	// special gt (ad-hoc)	
//	if (addCompare)
//	{
//		textStreamMFile<<"plot(X, Yfuzzy*100, 'b', 'LineWidth', 3), hold all;\n";
//		textStreamMFile<<"plot(X, Ygt*100, '--m', 'LineWidth', 1), hold all;\n";
//		textStreamMFile<<"legend('Ours', 'Fuzzy', 'GT');\n";
//	}
//	else
//		textStreamMFile<<"legend('Ours');\n";
//	
//	textStreamMFile<<"xlabel('Euclidean Distance');\n";
//	textStreamMFile<<"ylabel('% Correspondences');\n";
//	textStreamMFile<<"set(gca,'XTick',0:.05:.25);\n";
//	textStreamMFile<<"set(gca,'YTick',0:10:100);\n";	
//	textStreamMFile.close();
	
}




