
#include "LinAlg.h"
class LinAlgMatrixComplex;
class LinAlgVectorReal;

#ifndef __LINALG_MATRIX_REAL_H
#define __LINALG_MATRIX_REAL_H

class LinAlgMatrixReal
{
	public:
		friend class LinAlgVectorReal;
		friend class LinAlgMatrixComplex;	
	
	// Initialization	
		LinAlgMatrixReal(int m, int n, double initValue = 0);
		LinAlgMatrixReal(const LinAlgMatrixReal & srcMatrix);		
		virtual ~LinAlgMatrixReal();
	
	// Matrix operations, changing matrix
		LinAlgMatrixReal operator * (LinAlgMatrixReal & m2);	
		virtual void SetValues(double value);
		virtual void MultiplyBy(double value);
		virtual void Add(LinAlgMatrixReal & otherMatrix);	
		LinAlgMatrixReal & operator=(const LinAlgMatrixReal & src);
		double GetVal(int i, int j);
		void AddVal(int i, int j, double val);
		double & operator()(int i, int j);
		const double & operator()(int i, int j) const;
	
	// Matrix operations, save result to a target
		virtual void Transpose(LinAlgMatrixReal & trg);	
		virtual void PointwiseMultiply(LinAlgMatrixReal & m2, LinAlgMatrixReal & trg);
		virtual void Multiply(const LinAlgMatrixReal & m2, LinAlgMatrixReal & trg) const;
		virtual void Multiply(LinAlgVectorReal & v, LinAlgVectorReal & vTrg);
	
	// System of Linear Equations
		virtual void Inverse();
		virtual void Solve(LinAlgVectorReal & b, LinAlgVectorReal &x);
		virtual void SolveLUDestructive(LinAlgVectorReal & b, LinAlgVectorReal &x);
		virtual void LeastSquaresQRDestructive(LinAlgVectorReal & b, LinAlgVectorReal &x);
		
		virtual void DecomposeCholesky();
		virtual void SolveDecomposedCholesky(LinAlgVectorReal & b, LinAlgVectorReal &x);
	
	// eigendecomposition
		virtual void EigendecompositionWithMATLAB(LinAlgVectorReal & eigenValues, LinAlgMatrixReal & eigenVectors,
												  int nEigs, const char * uniquePrefixChar);
		virtual void EigendecompositionSymmetric(LinAlgVectorReal & eigenValues, LinAlgMatrixReal & eigenVectors);
		virtual void EigendecompositionSymmetricSort(LinAlgVectorReal & eigenValues, LinAlgMatrixReal & eigenVectors);
		virtual void SortEigenvectors(LinAlgVectorReal & eigenValues, LinAlgMatrixReal & eigenVectors);
		virtual void Symmetrize();
		virtual void RowNormalize(LinAlgVectorReal * saveRows);

	
	// svd decomposition
		virtual void SVDDecomposeDestructive(LinAlgMatrixReal & U, 
											 LinAlgVectorReal & S, LinAlgMatrixReal & Vt);	

	// Getters	
		virtual double GetMax(int * saveRow=NULL, int * saveCol=NULL);
		virtual double GetMin(int * saveRow=NULL, int * saveCol=NULL);
		virtual double Norm(double p);		//pth norm		
		virtual int Rows() const;
		virtual int Cols() const;
	
	// Visualizing / Type Conversion matrix
		virtual void PrintMatrix(double absMax=0);
		virtual void PrintMatlab();	
		virtual bool WriteMatrixBin(const char * outputFile);
		virtual bool ReadMatrixBin(const char * inputFile);
		virtual bool WriteMatrixASCII(const char * outputFile);
		virtual bool ReadMatrixASCII(const char * inputFile);
			
	protected:		
		bool m_bContainsGarbage;
		int m, n;
		double * m_aRealData;
		bool m_bFreeData;
	
		static std::map<int, std::map<int, double *> > m_LapackWorkspaces;
		static std::map<int, std::map<int, int *> > m_LapackIntWorkspaces;
	
		static double * GetLapackWorkspace(int arraySize, int id);
		static int * GetLapackIntWorkspace(int arraySize, int id);
		
};

#endif
