function [X, D, Y] = mdsdeforms(featuresDir, listName, writeMtx, ndims)
    fid = fopen([listName], 'rt');
    if fid==-1
        disp(['[ERROR] Invalid file ' featuresDir '/' listName]);
        assert(fid ~= -1);
    end
    tline = 'a';
    nPnts = 0;
    nFeat = -1;
    nNodes = -1;
    
    % go over all models - create feature vectors
    counter = 0;
    while ischar(tline)
         counter = counter + 1;
         tline = fgets(fid);
         lnStr = strfind(tline, ' ');
         if tline==-1
             continue;
         end
         mName = strtrim(tline);
         if length(lnStr)>0
             mName = strtrim(tline(1:(lnStr(1)-1)));
         end

         % read deformation
         fName =  [featuresDir '/' mName '.deform.bin'];
         fidDeform = fopen(fName, 'rb');
         if fidDeform==-1
             disp(['[ERROR] Invalid File ' fName]);
             features = -ones(length(features), 1);
             features(2) = rand;
             X = [X ; features'];
             continue;
         end
         % read nodes
         nNodesCurr = fread(fidDeform, 1, 'int');
         
         % read per node values
         nFeatCurr = fread(fidDeform, 1, 'int');
         if nFeat == -1 
             nFeat = nFeatCurr;
             nNodes = nNodesCurr;
         end
         assert(nFeat==nFeatCurr && nNodes==nNodesCurr);
         nPnts = nPnts + 1;
         
         % read feature vector
         features = fread(fidDeform, nFeat * nNodes, 'double')';
         if exist('X')
             X = [X ; features];
         else
             X = features;
         end
         fclose(fidDeform);
    end
    fclose(fid);
        
    % re-weight features
    if nFeat>6
        offsets = [0:(nNodes-1)]*nFeat;
        
        for i = offsets
             X(:, i + [7:nFeat]) = -1;
             %X(:, i + [1:6]) = -1;
        end
    end
    
    % embed using the feature vectors
    D = pdist(X,'euclidean');
    [m n] = size(X);
    if m<4   % no embedding 
        Y = zeros(m, 2);
        Y(:, 1) = (1:m) / m;
    else % mds
        opts = statset('Display', 'iter', 'MaxIter', 50);
        %Y = mdscale(D, ndims, 'Criterion', 'sstress', 'Options', opts);
        Y = cmdscale(D);
        [m, n] = size(Y);
        if n==1
            Y = [Y(:,1)' ; ones(m, n)']';
        else
            Y = Y(:, 1:2);
        end
    end
    disp(['Writing ' writeMtx]);
    writeMatrixDense(Y, writeMtx);
end



