#include "ParamParser.h"
#include <assert.h>

ParamParser::ParamParser(const VKString & inFilename, const char ** argv, int argc)
{
	std::ifstream file(inFilename.c_str());

	if (!file.is_open())
	{
		if (!argc>0)	//all params must be here
		{
			std::cout<<"[ERROR] Param File does not exist and no command line arguments provided: "<<inFilename.c_str()<<std::endl;
			assert(false);
		}
	}
	
	VKString currParam = "General";
		//read file, fill values
	VKStringList commands;
	VKString tempString(file);

	commands = tempString.replace("\n", " ").replace("\t", " ").split(" ");
	for (int i=0; i<argc; i++)	//read arguments, add or refill arguments
		commands.push_back(argv[i]);
	
	int currParamID = -1;
	for (int i=0; i<commands.count(); i++)
	{		
		if (commands[i].startsWith("-") || commands[i].endsWith(":"))
		{
			currParam = commands[i].replace("-", "").replace(":","");
			std::map<VKString, int>::iterator iter = m_ParamModulesIDs.find(currParam);
			if (iter==m_ParamModulesIDs.end())
			{
				m_ModulesContent.push_back(std::map<VKString, VKStringList>());
				m_ModuleAndParamOrder.push_back(VKStringList());
				currParamID = (int)m_ModuleAndParamOrder.size()-1;
				m_ModuleAndParamOrder[currParamID].push_back(currParam);
				m_ParamModulesIDs[currParam] = currParamID;
			}
			else
				currParamID = iter->second;
		}
		else if (currParam!="Invalid")
		{
			VKString field = commands[i++];
			VKString fld = field;
			if (fld.startsWith("+"))
				fld.replace("+", "");
			else
			{
				m_ModulesContent[currParamID][fld].clear();
				assert(currParamID>=0 && currParamID < (int)m_ModuleAndParamOrder.size());
				if (!m_ModuleAndParamOrder[currParamID].contains(fld))
					m_ModuleAndParamOrder[currParamID].push_back(fld);
			}
			if (commands.count()<=i)
			{
				std::cout<<"[ERROR] Cannot parse "<<inFilename.c_str()<<" (add. args: "<<argc<<")"<<std::endl;
				std::cout<<"\ti="<<i<<" / "<<commands.count()<<std::endl;
				std::cout<<"\tCurrParam="<<currParam.c_str()<<std::endl;
				std::cout<<"\tfield="<<field.c_str()<<std::endl;
				assert(commands.count()>i);
			}
			if (commands[i]=="[]")
				m_ModulesContent[currParamID][fld] = VKStringList();
			else if (commands[i].startsWith("["))				//loop over multiple param values
			{
				int level = 0;
				assert(commands.count()>i);
				if (commands[i].length()>1)
					commands[i] = commands[i].right(commands[i].length()-1);
				else
					i++;
					
				while(level>=0)
				{
					assert(i < commands.count());
					
					VKString str = commands[i]; 
					int lenOrig = str.length();
					str = str.replace("[", "");
					int lenNoOpen = str.length();
					str = str.replace("]", "");
					int lenNoBrackets = str.length();
					
					int nOpen = lenOrig - lenNoOpen;
					int nClose = lenNoOpen - lenNoBrackets;
					level += (nOpen - nClose);
					
					if (level < 0 && commands[i].endsWith("]") && commands[i].length()>1)
						commands[i] = commands[i].left(commands[i].length()-1);
					
//					std::cout<<"Arg: "<<fld.c_str()<<" cmd["<<i<<"] = |"<<commands[i].c_str()<<"|"<<std::endl;
//					std::cout<<"\tLevel="<<level<<" nOpen="<<nOpen<<" nClose="<<nClose<<std::endl;
//					std::cout<<"\tstr="<<str.c_str()<<" len=["<<lenOrig<<" "<<lenNoOpen<<" "<<lenNoBrackets<<"]"<<std::endl;
					
					m_ModulesContent[currParamID][fld].push_back(commands[i]);
					
					if (level >= 0)
						i++;
				}
			}
//			else		//set single param value
//			{
//				m_ParamModules[currParam][fld].push_back(commands[i].replace("[", "").replace("]", ""));
//			}
		}
	}
	file.close();
}

void ParamParser::WriteParams(const VKString & outFile)
{
	std::ofstream textStream(outFile.c_str());
	assert(textStream.is_open());
//	for (std::map<VKString, std::map<VKString, VKStringList> >::iterator iter = m_ParamModules.begin();
//		 iter!=m_ParamModules.end(); iter++)
//	{
//		textStream<<iter->first.c_str()<<":\n";
//		for (std::map<VKString, VKStringList>::iterator iter2 = iter->second.begin();
//			 iter2!=iter->second.end(); iter2++)
//		{
//			textStream<<"\t"<<iter2->first.c_str()<<"\t["<<iter2->second.join(", ").c_str()<<"]\n";
//		}
//	}	
	for (int i=0; i<(int)m_ModuleAndParamOrder.size(); i++)
	{
		textStream<<m_ModuleAndParamOrder[i][0].c_str()<<":\n";
		for (int j=1; j<m_ModuleAndParamOrder[i].count(); j++)
		{
			VKString & field = m_ModuleAndParamOrder[i][j];
			VKStringList & vals = GetStrValues(m_ModuleAndParamOrder[i][0], field);
			textStream<<"\t"<<field.c_str()<<"\t["<<vals.join(" ").c_str()<<"]\n";
		}
	}
	textStream.close();
}

void ParamParser::PrintParams()
{
	assert(false);
//	for (std::map<VKString, std::map<VKString, VKStringList> >::iterator iter = m_ParamModules.begin();
//		 iter!=m_ParamModules.end(); iter++)
//	{
//		std::cout<<"Module = "<<iter->first.c_str()<<std::endl;
//		for (std::map<VKString, VKStringList>::iterator iter2 = iter->second.begin();
//			 iter2!=iter->second.end(); iter2++)
//		{
//			std::cout<<"\t'"<<iter2->first.c_str()<<"' = [";
//			std::cout<<iter2->second.join(", ").c_str()<<"]"<<std::endl;
//		}
//	}
}

VKStringList & ParamParser::GetStrValues(const VKString & module, 
										 const VKString & valName, bool & valid)
{
	std::map<VKString, int>::iterator iter1;
	iter1 = m_ParamModulesIDs.find(module);
	valid = (iter1!=m_ParamModulesIDs.end());
	if (!valid)
		return m_EmptyList;
	
	assert(iter1->second < (int)m_ModulesContent.size());
	std::map<VKString, VKStringList>::iterator iter2 = m_ModulesContent[iter1->second].find(valName);
	valid = (iter2 != m_ModulesContent[iter1->second].end());
	
	if (valid)
		return iter2->second;
	else
		return m_EmptyList;
}

VKString ParamParser::GetStrValue(const VKString & module, 
								  const VKString & valName, bool & valid)
{	
	VKStringList & vals = GetStrValues(module, valName, valid);
	assert(vals.count()>0);
	return vals[0];
}

double ParamParser::GetDoubleValue(const VKString & module, 
								   const VKString & valName, bool & valid)
{
	VKStringList & vals = GetStrValues(module, valName, valid);
	assert(vals.count()>0);
	
	if (valid)
		return vals[0].toDouble(&valid);
	else
		return -1.;
}

int ParamParser::GetIntValue(const VKString & module, 
							 const VKString & valName, bool & valid)
{
	VKStringList & vals = GetStrValues(module, valName, valid);
	assert(vals.count()>0);

	if (valid)
		return vals[0].toInt(&valid);
	else
		return -1;
}

std::vector<double> ParamParser::GetDoubleValues(const VKString & module, 
												 const VKString & valName, bool & valid)
{
	VKStringList & vals = GetStrValues(module, valName, valid);
	std::vector<double> newVals;
	if (valid)
	{
		bool currValid;
		for (int i=0; i<vals.count(); i++)
		{
			newVals.push_back(vals[i].toDouble(&currValid));
			valid = currValid && valid;
		}
	}
	return newVals;
}

std::vector<int> ParamParser::GetIntValues(const VKString & module, 
										   const VKString & valName, bool & valid)
{
	VKStringList & vals = GetStrValues(module, valName, valid);
	std::vector<int> newVals;

	if (valid)
	{
		bool currValid;
		for (int i=0; i<vals.count(); i++)
		{
			newVals.push_back(vals[i].toInt(&currValid));
			valid = currValid && valid;
		}
	}
	return newVals;
}


VKStringList & ParamParser::GetStrValues(const VKString & module, 
										 const VKString & valName)
{
	VKStringList & vals = GetStrValues(module, valName, m_Valid);
	AssertParamValid(module, valName, "STRs", "find");
	return vals;
}

VKString ParamParser::GetStrValue(const VKString & module, 
								  const VKString & valName)
{	
	VKStringList & vals = GetStrValues(module, valName, m_Valid);
	assert(vals.count()>0);
	AssertParamValid(module, valName, "STR", "find");	
	return vals[0];
}

double ParamParser::GetDoubleValue(const VKString & module, 
								   const VKString & valName)
{
	VKStringList & vals = GetStrValues(module, valName, m_Valid);
	assert(vals.count()>0);
	AssertParamValid(module, valName, "DBL", "find");	
	double retVal = vals[0].toDouble(&m_Valid);
	AssertParamValid(module, valName, "DBL", "parse");
	return retVal;
}

int ParamParser::GetIntValue(const VKString & module, 
							 const VKString & valName)
{
	VKStringList & vals = GetStrValues(module, valName, m_Valid);
	assert(vals.count()>0);
	AssertParamValid(module, valName, "INT", "find");
	int retVal = vals[0].toInt(&m_Valid);
	AssertParamValid(module, valName, "INT", "parse");
	return retVal;
}

std::vector<double> ParamParser::GetDoubleValues(const VKString & module, 
												 const VKString & valName)
{
	VKStringList & vals = GetStrValues(module, valName, m_Valid);
	AssertParamValid(module, valName, "DBLs", "find");
	std::vector<double> newVals;	
	for (int i=0; i<vals.count(); i++)
	{
		newVals.push_back(vals[i].toDouble(&m_Valid));
		AssertParamValid(module, valName, "DBLs", "parse");
	}
	return newVals;
}

std::vector<int> ParamParser::GetIntValues(const VKString & module, 
										   const VKString & valName)
{
	VKStringList & vals = GetStrValues(module, valName, m_Valid);
	std::vector<int> newVals;
	AssertParamValid(module, valName, "INTs", "find");
	
	for (int i=0; i<vals.count(); i++)
	{
		newVals.push_back(vals[i].toInt(&m_Valid));
		AssertParamValid(module, valName, "INTs", "parse");
	}
	return newVals;
}

void ParamParser::AssertParamValid(const VKString & module, const VKString & paramName,
								   const char * datatype, const char * find_parse)
{
	if (!m_Valid)
	{
		std::cout<<"[ERROR] Cannot "<<find_parse<<" "<<datatype<<" param.";
		std::cout<<"Module="<<module.c_str()<<" ParamName="<<paramName.c_str()<<std::endl;
		assert(m_Valid);
	}	
}

int ParamParser::NModules() const
{
	return (int)m_ModuleAndParamOrder.size();
}

const VKString & ParamParser::ModuleName(int moduleID) const
{
	assert(moduleID >= 0 && moduleID < (int) m_ModuleAndParamOrder.size());
	return m_ModuleAndParamOrder[moduleID][0];
}


void ParamParser::SetParam(const VKString & moduleName, const VKString & paramName, 
						   const VKStringList & vals)
{
	std::map<VKString, int >::iterator iter1 = m_ParamModulesIDs.find(moduleName);
	
	int moduleID = -1;
	if (iter1 == m_ParamModulesIDs.end())
	{
		moduleID = (int)m_ModuleAndParamOrder.size();
		m_ParamModulesIDs[moduleName] = moduleID;
		m_ModuleAndParamOrder.push_back(VKStringList());
		m_ModuleAndParamOrder[moduleID].push_back(moduleName);
		m_ModulesContent.push_back(std::map<VKString, VKStringList>());
	}
	else 
		moduleID = iter1->second;
	
	assert(moduleID < (int)m_ModulesContent.size());
	std::map<VKString, VKStringList>::iterator iter2;
	iter2 = m_ModulesContent[moduleID].find(paramName);
	if (iter2 == m_ModulesContent[moduleID].end())
	{
		assert(!m_ModuleAndParamOrder[moduleID].contains(paramName));
		m_ModuleAndParamOrder[moduleID].push_back(paramName);
		m_ModulesContent[moduleID][paramName] = vals;
	}
	else
		iter2->second = vals;
}



