/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "CorrsCreator.h"
/////
CorrParams::CorrParams(const VKString & corrsParams,
					   const VKString & className,
					   const VKString & paramClassName)
{
	m_sParamsName = corrsParams;
	m_sClassName = className;
	m_sParamClassName = paramClassName;
}


////////////////// CorrsReader //////////////////

std::map<VKString, CorrsCreator*> CorrsCreator::s_AllCreators;

CorrsCreator * CorrsCreator::GetCreator(const VKString & className)
{
	std::map<VKString, CorrsCreator*>::iterator iter;
	iter = s_AllCreators.find(className);
	if (iter == s_AllCreators.end())
	{
		std::cout<<"[ERROR] Cannot find corrs reader for class: "<<className.c_str();
		std::cout<<std::endl;
		assert(false);
	}
	return iter->second;
}

void CorrsCreator::AddCreator(CorrsCreator * creator)
{
	if (s_AllCreators.find(creator->ClassName())==s_AllCreators.end())
		s_AllCreators[creator->ClassName()] = creator;
}

void CorrsCreator::AddCreateParams(CorrParams * algParams)
{
	CorrsCreator * creator = GetCreator(algParams->m_sClassName);
	std::map<VKString, int>::iterator iter;
	iter = creator->m_AlgNameToID.find(algParams->m_sParamsName);
	assert (iter != creator->m_AlgNameToID.end());
	int algID = (int)creator->m_Algs.size();
	creator->m_Algs.push_back(algParams);
	creator->m_AlgNameToID[algParams->m_sParamsName] = algID;
}


CorrsCreator::CorrsCreator(ShapeCorrs::CoreCorrsType tp,
						   const VKString & className)
{
	m_Type = tp;
	m_sClassName = className;
}

CorrsCreator::~CorrsCreator()
{
	for (int i=0; i<(int)m_Algs.size(); i++)
		delete m_Algs[i];
}

CorrParams * CorrsCreator::GetAlgParams(const VKString & algName)
{
	std::map<VKString, int>::iterator iter = m_AlgNameToID.find(algName);
	assert(iter != m_AlgNameToID.end());
	return m_Algs[iter->second];
}

const VKString & CorrsCreator::ClassName() const
{
	return m_sClassName;
}

ShapeCorrs::CoreCorrsType CorrsCreator::ClassType() const
{
	return m_Type;
}

CorrsPair * CorrsCreator::ReadCorrsPair(const VKString & corrName, GeomShape * shapeFrom, 
										GeomShape * shapeTo, FILE * file, DBCorrs * dbCorrs)
{	assert(false);	}

CorrsPairMap * CorrsCreator::ReadCorrsPairMap(const VKString & corrName, GeomShape * shapeFrom, 
											  GeomShape * shapeTo, FILE * file, DBCorrs * dbCorrs)
{	assert(false);	}

CorrsCollection * CorrsCreator::ReadCorrsCollection(const VKString & corrsName, 
													const std::vector<GeomShape*> & shapes, 
													FILE * file, DBCorrs * dbCorrs)
{	assert(false);	}

CorrsCollectionMap * CorrsCreator::ReadCorrsCollectionMap(const VKString & corrsName, 
														  const std::vector<GeomShape*> & shapes, 
														  FILE * file, DBCorrs * dbCorrs)
{	assert(false);	}

void CorrsCreator::CreateCorrsPair(const VKString & corrName, GeomShape * shapeFrom, 
								   GeomShape * shapeTo, CorrParams * params, 
								   std::vector<CorrsPair *> & saveCorrs, DBCorrs * dbCorrs)
{	assert(false);	}

void CorrsCreator::CreateCorrsPairMap(const VKString & corrName, GeomShape * shapeFrom, 
									  GeomShape * shapeTo, CorrParams * params, 
									  std::vector<CorrsPairMap *> & saveCorrs, DBCorrs * dbCorrs)
{	assert(false);	}

void CorrsCreator::CreateCorrsCollection(const VKString & corrsName, 
										 const std::vector<GeomShape*> & shapes, CorrParams * params, 
										 std::vector<CorrsCollection *> & saveCorrs, DBCorrs * dbCorrs)
{	assert(false);	}

void CorrsCreator::CreateCorrsCollectionMap(const VKString & corrsName, 
											const std::vector<GeomShape*> & shapes, CorrParams * params, 
											std::vector<CorrsCollectionMap *> & saveCorrs, DBCorrs * dbCorrs)
{	assert(false);	}

////////////////// CoreCorrsReader //////////////////

void CoreCorrsReader::InitCoreCorrsReaders()
{
	CoreCorrsReader * c1 = new CoreCorrsReader(ShapeCorrs::CORECORRS_CorrsCollection, 
											   "CorrsCollection");
	CoreCorrsReader * c2 = new CoreCorrsReader(ShapeCorrs::CORECORRS_CorrsCollectionMap, 
											   "CorrsCollectionMap");
	CoreCorrsReader * c3 = new CoreCorrsReader(ShapeCorrs::CORECORRS_CorrsPair, 
											   "CorrsPair");
	CoreCorrsReader * c4 = new CoreCorrsReader(ShapeCorrs::CORECORRS_CorrsPairMap, 
											   "CorrsPairMap");
	CorrsCreator::AddCreator(c1);
	CorrsCreator::AddCreator(c2);
	CorrsCreator::AddCreator(c3);
	CorrsCreator::AddCreator(c4);
	
	// todo: v2v, p2p other point-to-point readers...
}

CoreCorrsReader::CoreCorrsReader(ShapeCorrs::CoreCorrsType tp, 
								 const VKString & className)
: CorrsCreator(tp, className)
{
	
}

CorrsPair * CoreCorrsReader::ReadCorrsPair(const VKString & corrName, 
										   GeomShape * shapeFrom, GeomShape * shapeTo,
										   FILE * file, DBCorrs * dbCorrs)
{
	CorrsPair * corrs = new CorrsPair(corrName, shapeFrom, shapeTo, dbCorrs);
	corrs->ReadShapeCorrs(file);
	return corrs;
}

CorrsPairMap * CoreCorrsReader::ReadCorrsPairMap(const VKString & corrName, 
												 GeomShape * shapeFrom, GeomShape * shapeTo,
												 FILE * file, DBCorrs * dbCorrs)
{
	CorrsPairMap * corrs = new CorrsPairMap(corrName, shapeFrom, shapeTo, dbCorrs);	
	corrs->ReadShapeCorrs(file);
	return corrs;
}

CorrsCollection * CoreCorrsReader::ReadCorrsCollection(const VKString & corrsName,
													   const std::vector<GeomShape*> & shapes,
													   FILE * file, DBCorrs * dbCorrs)
{
	CorrsCollection * corrs = new CorrsCollection(corrsName, shapes, dbCorrs);
	corrs->ReadShapeCorrs(file);
	return corrs;

}

CorrsCollectionMap * CoreCorrsReader::ReadCorrsCollectionMap(const VKString & corrsName,
															 const std::vector<GeomShape*> & shapes,
															 FILE * file, DBCorrs * dbCorrs)
{
	//return new CorrsCollectionMap(corrsName, shapes);
	assert(false);
	return NULL;	// TODO
}

