/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __CorrsModelGeoms_H
#define __CorrsModelGeoms_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>
#include "gaps.h"

#include "GeomShape.h"
#include "ShapePoint.h"

#include "CorrsDistProjection.h"

/** 
 * CorrsModelGeoms - map between the same geometries within a model
 * (not fully implemented / used yet)
 */
class CorrsModelGeoms : public ShapeCorrs
{
	public:
		CorrsModelGeoms(const VKString & corrsName, Model3D * model,
						CorrsDistProjection::InterpolationType interp);

		virtual ~CorrsModelGeoms();

		virtual void Domain(std::vector<GeomShape *> & domainShapes );
		virtual void Range(std::vector<GeomShape *> & rangeShapes );

	
		virtual int GetCorrs(const ShapePoint & pnt,
							 std::vector<ShapePoint> & corrs, 
							 const VKString & corrValName="none",
							 std::vector<double> *corrVals = NULL);

		virtual int PairwiseCorrs(const ShapePoint & pnt,
								  GeomShape * toShape,
								  std::vector<ShapePoint> & corrs, 
								  const VKString & corrValName = "none",
								  std::vector<double> * corrValue = NULL);	

		virtual ShapePoint BestCorr(const ShapePoint & pnt,
									GeomShape * toShape,
									const VKString & corrValName = "none",
									double * corrValue = NULL);

//		virtual void WriteShapeCorrs(FILE * file);
//		virtual void ReadShapeCorrs(FILE * file);

	protected:
		Model3D * m_pModel;
		CorrsDistProjection::InterpolationType m_Interp;
		
};

#endif

