/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __CorrsPair_H
#define __CorrsPair_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "VKString.h"

#include "GeomShape.h"
#include "ShapePoint.h"
#include "ShapeCorrs.h"
#include "ShapePointCloud.h"

/** 
 * CorrsPair - fuzzy pairwise
 */
class CorrsPair : public ShapeCorrs
{
	public:
//		static void WriteHeader(FILE * file, PairMapParams params);
//		static PairMapParams ReadHeader(FILE * file);
	
		CorrsPair(const VKString & corrName, GeomShape * shapeFrom, 
				  GeomShape * shapeTo, DBCorrs * corrs=NULL, 
				  CoreCorrsType tp = ShapeCorrs::CORECORRS_CorrsPairMap);
		virtual ~CorrsPair();
		virtual VKString GetCorrsClassName() const;
	
//		const VKString & ParentCollectionName(int * saveID) const;
		virtual GeomShape * ShapeFrom();
		virtual GeomShape * ShapeTo();
		virtual ShapePointCloud * SafeCastPntsFrom();
		virtual ShapePointCloud * SafeCastPntsTo();
		virtual const Model3D * ModelFrom() const;
		virtual const Model3D * ModelTo() const;
		virtual CorrsPair * Inverse();
		
		virtual void Domain(std::vector<const GeomShape *> & domainShapes ) const ;
		virtual void Range(std::vector<const GeomShape *> & rangeShapes ) const ;
	
		virtual void WriteShapeCorrs(FILE * file);
		virtual void ReadShapeCorrs(FILE * file);
		void WriteShapeCorrsAsCorrsPair(FILE * file);
		void ReadShapeCorrsAsCorrsPair(FILE * file);
	
		virtual void SetFromShape(GeomShape * shape);
		virtual void SetToShape(GeomShape * shape);
	
		virtual int GetCorrs(const ShapePoint & pnt,
							 std::vector<ShapePoint> & corrs, 
							 const VKString & corrValName,
							 std::vector<double> *corrVals);

	// NOTE: only 'default' and 'none' corr values are allowed (todo other).
		void AutoFillCache(const VKString & corrValueName = "default");
	// also use InitCache to clear cache
		void InitCache(const VKString & corrValueName = "default");
	
	// NOTE: incremental caching is not implemented. todo
		bool SetCachedCorrs(int fromPntID, const std::vector<ShapePoint> & corrs);
		bool SetCachedVals(int fromPntID, const std::vector<double> & vals);
		bool GetCachedCorrs(int fromPntID, std::vector<ShapePoint> & corrs);
		bool GetCachedVals(int fromPntID, std::vector<double> & vals);
		void SetCacheFilled(bool filled);	
		bool CacheFilled() const;
		int GetCachedCorrs(const ShapePoint & pnt,
							std::vector<ShapePoint> & corrs, 
							const VKString & corrValName,
							std::vector<double> *corrVals);
		
	protected:
		GeomShape * m_pShapeFrom;
		GeomShape * m_pShapeTo;
	
	private:
		bool m_bCacheFilled;
		int m_iNCachedFromPnts;
		std::vector<ShapePoint> * m_aCachedCorrs;
		std::vector<double> * m_aCachedCorrVals;
};

#endif

