/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "CorrsPairMap.h"

#include "ShapeMesh.h"
#include "Model3D.h"

CorrsPairMap::CorrsPairMap(const VKString & corrName,
						   GeomShape * shapeFrom, 
						   GeomShape * shapeTo, DBCorrs * corrs)
: CorrsPair(corrName, shapeFrom, shapeTo, corrs, ShapeCorrs::CORECORRS_CorrsPairMap)
{
	m_GenPnts1 = NULL;
	m_GenPnts2 = NULL;
}

CorrsPairMap::~CorrsPairMap()
{
	
}

VKString CorrsPairMap::GetCorrsClassName() const
{
	return "CorrsPairMap";
}

int CorrsPairMap::GetCorrs(const ShapePoint & pnt,
					   std::vector<ShapePoint> & corrs, 
					   const VKString & corrValName,
					   std::vector<double> *corrVals)
{
	if (corrVals==NULL)
		corrs.push_back(MapPoint(pnt, corrValName));
	else
	{
		corrVals->push_back(0);
		corrs.push_back(MapPoint(pnt, corrValName, &((*corrVals)[0])));
	}
	return 1;
}

int CorrsPairMap::PairwiseCorrs(const ShapePoint & pnt,
								GeomShape * toShape,
								std::vector<ShapePoint> & corrs, 
								const VKString & corrValName,
								std::vector<double> * corrVals)
{
	assert(toShape == ShapeTo());
	if (corrVals==NULL)
		corrs.push_back(MapPoint(pnt, corrValName));
	else
	{
		corrVals->push_back(1.);
		corrs.push_back(MapPoint(pnt, corrValName, &((*corrVals)[0])));
	}
}

ShapePoint CorrsPairMap::BestCorr(const ShapePoint & pnt,
							  GeomShape * toShape,
							  const VKString & corrValName,
							  double * corrValue)
{
	assert(toShape == ShapeTo());
	if (corrValue==NULL)
		return MapPoint(pnt, corrValName);
	else
	{
		return MapPoint(pnt, corrValName, corrValue);
	}
}

void CorrsPairMap::WriteShapeCorrs(FILE * file)
{
	WriteShapeCorrsAsCorrsPairMap(file);
}

void CorrsPairMap::ReadShapeCorrs(FILE * file)
{
	ReadShapeCorrsAsCorrsPairMap(file);
}

void CorrsPairMap::WriteShapeCorrsAsCorrsPairMap(FILE * file)
{
	assert(false);
}

void CorrsPairMap::ReadShapeCorrsAsCorrsPairMap(FILE * file)
{
	assert(false);
}

void CorrsPairMap::AutoFillMapCache(const VKString & corrValueName)
{
	assert(false);	
}

void CorrsPairMap::InitMapCache(int nFromPnts,
								const VKString & corrValueName)
{
	assert(false);
}

bool CorrsPairMap::SetCachedMapCorrs(int fromPntID, const ShapePoint & corr)
{
	assert(false);
}

bool CorrsPairMap::SetCachedMapVals(int fromPntID, double val)
{
	assert(false);
}

ShapePoint CorrsPairMap::GetCachedMapCorrs(int fromPntID)
{
	assert(false);
}

double CorrsPairMap::GetCachedMapVals(int fromPntID)
{
	assert(false);
}

void CorrsPairMap::SetMapCacheFilled(bool filled)
{
	assert(false);
}

bool CorrsPairMap::MapCacheFilled() const
{
	assert(false);
}

ShapePoint CorrsPairMap::MapPoint(const ShapePoint & pnt, 
								  const VKString & corrValName,
								  double * corrValue)
{
	assert(false);
}

void CorrsPairMap::RememberGenerators(const ShapePointCloud * pnts1, 
									  const ShapePointCloud * pnts2,
									  const std::vector<int> &corrs1,
									  const std::vector<int> &corrs2)
{
	assert(m_Corrs1.size()==0 && m_Corrs2.size()==0);
	m_GenPnts1 = pnts1;
	m_GenPnts2 = pnts2;
	m_Corrs1 = corrs1;
	m_Corrs2 = corrs2;
}

bool CorrsPairMap::SameGenPnts(const ShapePointCloud * pnts1, 
							   const ShapePointCloud * pnts2) const
{
	return m_GenPnts1==pnts1 && m_GenPnts2==pnts2;
	
}

void CorrsPairMap::GetGenerators(std::vector<int> & corrs1,
								 std::vector<int> & corrs2) const
{
	corrs1 = m_Corrs1;
	corrs2 = m_Corrs2;
}

void CorrsPairMap::WriteGenerators(FILE * file)
{
	// no need to waste space w/ set names
//	VKString set1Name = "None";
//	if (m_GenPnts1!=NULL)
//		set1Name = m_GenPnts1->GeometryName();
//	
//	VKString set2Name = "None";
//	if (m_GenPnts2!=NULL)
//		set2Name = m_GenPnts2->GeometryName();
//	VKString::WriteString(set1Name, file);
//	VKString::WriteString(set2Name, file);
	
	assert(m_Corrs1.size()==m_Corrs2.size());
	int nCorrs = (int)m_Corrs1.size();
	fwrite(&nCorrs, sizeof(int), 1, file);
	if (nCorrs>0)
	{
		int * corrs = new int[2*nCorrs];
		for (int i=0; i<nCorrs; i++)
		{
			corrs[2*i+0] = m_Corrs1[i];
			corrs[2*i+1] = m_Corrs2[i];
		}
		fwrite(corrs, sizeof(int), 2*nCorrs, file);
		delete [] corrs;
	}
}

void CorrsPairMap::WriteVtx2VtxMap(const VKString & outFile)
{
	std::ofstream textStream(outFile.c_str());
	assert(textStream.is_open());
	ShapeMesh * mesh = ShapeFrom()->SafeCastToMesh();
	assert(mesh!=NULL);
	for (int i=0; i<mesh->NVertices(); i++)
	{
		ShapePoint p = MapPoint(ShapePoint(mesh, i));
		int fID;
		double b1, b2;
		const ShapeMesh * meshSave;
		p.GetBarycentricCoords(&fID, &b1, &b2, &meshSave);
		double b3 = 1. - b1 - b2;
		int vID = 0;
		if (b1 > b2 && b1>b3)
			vID = meshSave->VertexOnFace(fID, 0);
		else if (b2 > b3)
			vID = meshSave->VertexOnFace(fID, 1);
		else
			vID = meshSave->VertexOnFace(fID, 2);
		textStream<<vID<<"\n";	
	}
	textStream.close();
}

void CorrsPairMap::WriteVtx2PntMap(const VKString & outFile)
{
	std::ofstream textStream(outFile.c_str());
	assert(textStream.is_open());
	ShapeMesh * mesh = ShapeFrom()->SafeCastToMesh();
	assert(mesh!=NULL);
	for (int i=0; i<mesh->NVertices(); i++)
	{
		ShapePoint p = MapPoint(ShapePoint(mesh, i));
		int fID;
		double b1, b2;
		const ShapeMesh * meshSave;
		p.GetBarycentricCoords(&fID, &b1, &b2, &meshSave);
		double b3 = 1. - b1 - b2;
		textStream<<fID<<" "<<b1<<" "<<b2<<" "<<b3<<"\n";
	}
	textStream.close();
}

double CorrsPairMap::DefaultIntegratedValue(const VKString & evalPnts) const
{
	CorrsPairMap * ncThis = (CorrsPairMap*)this;
	const Model3D * model = ncThis->ShapeFrom()->GetModel();
	assert(model!=NULL);
	const GeomShape * shape = model->GetGeomShape(evalPnts);
	assert(shape!=NULL);
	const ShapePointCloud * pnts = shape->SafeCastToPointCloud();
	assert(pnts!=NULL);

	double sumVal = 0;
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		double val;
		ncThis->MapPoint(pnts->GetPoint(i), "default", &val);
		sumVal += val;
	}
	return sumVal / pnts->NumPoints();
}

void CorrsPairMap::ReadGenerators(FILE * file)
{
	//  no need to waste space w/ set names
//	VKString set1Name = VKString::ReadString(file);
//	VKString set2Name = VKString::ReadString(file);
//	if (set1Name!="None")
//	{
//		GeomShape * shape = ShapeFrom()->GetModel()->GetGeomShape(set1Name);
//		assert(shape!=NULL);
//		m_GenPnts1 = shape->SafeCastToPointCloud();
//		assert(m_GenPnts1!=NULL);
//	}
//	if (set2Name!="None")
//	{
//		GeomShape * shape = ShapeTo()->GetModel()->GetGeomShape(set2Name);
//		assert(shape!=NULL);
//		m_GenPnts2 = shape->SafeCastToPointCloud();
//		assert(m_GenPnts2!=NULL);
//	}
	
	int nCorrs;
	fread(&nCorrs, sizeof(int), 1, file);
	if (nCorrs>0)
	{
		int * corrs = new int[2*nCorrs];
		fread(corrs, sizeof(int), 2*nCorrs, file);
		for (int i=0; i<(int)nCorrs; i++)
		{
			m_Corrs1.push_back(corrs[2*i+0]);
			m_Corrs2.push_back(corrs[2*i+1]);
		}
		delete [] corrs;
	}
}

