/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __DBCorrs_H
#define __DBCorrs_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "GeomShape.h"
#include "ShapeCorrs.h"
#include "CollectionPairwise.h"

/** 
 * Stores all correspondences in a collection of 3d models
 */
class DBCorrs
{
	public:
		static const unsigned int s_iCorrsCollHeader;
		static const unsigned int s_iCorrsPairHeader;
		static const unsigned int s_iCorrsHeader;
	
		enum EntryType
		{
			DBCORRS_Unknown,
			DBCORRS_PairwiseColl,
			DBCORRS_Corrs
		};
		DBCorrs(ModelDatabase * db = NULL);
		virtual ~DBCorrs();
	
		virtual void AddPairColl(const VKString & pairwiseCorrName,
								 CollectionPairwise * pairwise);
		virtual void AddCorrs(const VKString & corrName, ShapeCorrs * corrs);

		virtual int NumPairColls() const;
		virtual int NumCorrs() const;
	
		virtual CollectionPairwise * GetPairColl(int id);
		virtual ShapeCorrs * GetCorrs(int id);

		virtual CollectionPairwise * GetPairColl(const VKString & collName);
		virtual ShapeCorrs * GetCorrs(const VKString & corrName);
		virtual EntryType CorrsType(const VKString & corrName);
	
	// readers
		static CollectionPairwise * ReadPairwiseCollection(const VKString & corrName,
														   ModelDatabase * db, 
														   const VKString & filename,
														   DBCorrs * dbCorrs);
		
	protected:
		std::vector<ShapeCorrs*> m_Corrs;
		std::vector<CollectionPairwise*> m_PairwiseCollections;
	
		std::map<VKString, int> m_PairwiseNameToID;
		std::map<VKString, int> m_CorrsNameToID;
		
		ModelDatabase * m_pDB;
		
};

#endif

