/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __Model3D_H
#define __Model3D_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "GeomShape.h"
#include "ParamParser.h"
#include "Model3DParser.h"

class ModelDatabase;

class Model3D
{
	public:
		static const VKString m_sMainGeomName;
		static const VKString m_sThumbGeomName;
	
		// easy model creation: keep as is
		static Model3D * CreateFromFile(const VKString & filename, int readingFlags, int creatingFlags,
										const VKString & modelName = "none",  int modelID = -1);
	
		static void AddParser(const VKString & pName, Model3DParser * parser);
	
		Model3D(const VKString & modelName, int modelID = -1);
		virtual ~Model3D();
	
		virtual void SetMetaData(const VKString & infoFile, ParamParser * metaData);
		virtual VKStringList & GetMetaData(const VKString & field, bool * exists);
		virtual void CreateMetaField(const VKString & field);
		virtual void WriteMetaData();
		
		virtual int NGeometries();
		virtual GeomShape * GetGeomShape(int shapeID);
		virtual GeomShape * GetGeomShape(const VKString & shapeName);
		virtual const GeomShape * GetGeomShape(int shapeID) const;
		virtual const GeomShape * GetGeomShape(const VKString & shapeName) const;
		virtual bool GeometryExists(const VKString & shapeName) const;
		virtual void AddGeomShape(const VKString & shapeName, GeomShape * shape);
		virtual void RemoveGeomShape(const VKString & name);
		virtual void RearrangeAxes(char newOrder[3], int newSigns[3]);
	
		virtual void SetModelName(const VKString & modelName);
		
		virtual void WriteModelDBFormat(const VKString & modelDir,
										const VKString & filename,
										int writingFlags = 0xFFFFFFFF,
										const VKString & writeArg = "none");

		virtual GeomShape * ReadGeometry(GeomShape::GeomShapeTypes tp, const VKString & geomName,
										 const VKString & geomFile, const VKString & pntsCloudMesh = "none");

		virtual GeomShape * CreateGeometry(GeomShape::GeomShapeTypes tp, const VKString & geomName,
										   const VKString & pntCloudMeshName="none");
		virtual ShapeValue * ReadGeomValue(const VKString & className, GeomShape * shape, 
										   const VKString & valueName, const VKString & shapeValFile);
		virtual ShapeValue * CreateGeomValue(const VKString & className, 
											 GeomShape * shape, const VKString & valueName);

//		virtual void ReadModelColladaFormat(const VKString & directory,
//											const VKString & filename);
		virtual bool ReadUnrecognizedFormat(const VKString & directory, 
											const VKString & filename);
		virtual void ReadModelPSBSegmentedFormat(const VKString & directory,
												 const VKString & filename);
		virtual void ReadModelPSBFormat(const VKString & directory,
										const VKString & filename);
		virtual void ReadModelDBFormat(const VKString & directory, 
									   const VKString & filename, int readingFlags);
		
		virtual const VKString & ModelName() const ;
		virtual const VKString & ModelFile() const ;
		virtual int ModelID() const;
		virtual void UpdateID(int newID);
	
//		virtual int ModelUniqueID() const;
//		virtual void SetUniqueID();

	protected:
		ParamParser * m_pMetaData;
		VKString m_sInfoFile;
		static VKStringList m_sEmptyStrList;
	
			// surface, point clouds, etc...
		std::vector<GeomShape *> m_GeomShapes;
		std::map<VKString, int> m_GeomShapeToID;
	
		int m_ModelID;
		VKString m_ModelName;
	
		static std::map<VKString, Model3DParser*>  m_TypeToModelParser;
};

#endif

