/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ShapeMaterials_H
#define __ShapeMaterials_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "ShapeValue.h"

#include "GeomShape.h"
class ColorOrTexture
{
	public:
		enum Type
		{
			BRDF_Invalid,
			BRDF_Color, 
			BRDF_Texture
		};
		ColorOrTexture();
		
		bool IsTexture();
		bool IsColor();
		void WriteColorOrTexture(FILE * file);
		void ReadColorOrTexture(FILE * file);
		
		Type m_Type;
		int m_iTextureID;
		RNRgb m_Color;
	
};

class MaterialBRDF
{
	public:
		MaterialBRDF();
		virtual ~MaterialBRDF();
	
	// NOTE: these are standard material properties in COLLADA. 
	// Most of them are not implemented / supported by our simple visualization tools
		ColorOrTexture m_EmitClr;
		ColorOrTexture m_DiffuseClr;
		ColorOrTexture m_AmbientClr;
		ColorOrTexture m_SpecularClr;
		double m_fShininess;
		ColorOrTexture m_ReflectiveClr;
		double m_Reflectivity;
		ColorOrTexture m_OpacityClr;
		double m_IndexOfRefraction;
		
		int m_iMaterialID;
};

class ShapeMaterials : public ShapeValue
{
	public:
		ShapeMaterials(GeomShape * shape, const VKString & valueName);
		virtual ~ShapeMaterials();
		virtual VKString GetClassName() const;
		virtual ShapeMaterials * SafeCastToMaterials();
		virtual const ShapeMaterials * SafeCastToMaterials() const;
		
		virtual int AddMaterial(MaterialBRDF * material);
		virtual int AddTextureImg(R2Image * img);
		
		virtual void SetFaceMaterials(std::vector<R2Point> & uvCoords,
									  std::vector<int> & perFaceMaterialIDs,
									  std::vector<int> & perFaceUVIds);
		virtual void FaceMaterial(int fID, double * u1, double * v1, double * u2, double * v2, 
								  double * u3, double * v3, int * materialID);

		virtual R2Image * GetTexture(int textureID);
		virtual MaterialBRDF * GetMaterial(int materialID);
		virtual int NTextures() const;
		virtual int NMaterials() const;
		virtual const std::vector<int> & GetAllFaces(int material);
	
		virtual void WriteShapeValues(const VKString & filename);
		virtual void ReadShapeValues(const VKString & filename);

		virtual void RemoveFace(int f);
		virtual void RemoveFaces(int f1, int f2);
	
	protected:	
		std::vector<int> * m_FacesPerMaterial;
	
		std::vector<MaterialBRDF *> m_Materials;
		std::vector<R2Image*> m_Textures;
};

#endif

