/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ShapeManifold_H
#define __ShapeManifold_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "ShapeMesh.h"

class MeshElementTriangle
{
	public:
		int v[3];
		int tID;
};

class MeshElementVertex
{
	public:
		double p[3];
		int vID;
		std::vector<int> faces;
		R3Vector normal;
		bool validNorm;
};



class ShapeTriangleSoup : public ShapeMesh
{
	public:
		enum OffFormat
		{
			OFF_AUTO,		// figure out from the header
			OFF_ASCII,		// ascii format
			OFF_BIN_SINGLE,	// floating pt single precision
			OFF_BIN_DOUBLE  // floating pt - double precision
		};
		ShapeTriangleSoup(const VKString & geomFile = "none", const VKString & geomName = "none", Model3D * model = NULL);
		virtual ~ShapeTriangleSoup();
		virtual ShapeTriangleSoup * SafeCastToTriangleSoup();

		virtual void WriteGeometryFile(const VKString & geomFile);
		virtual void ReadGeometryFile(const VKString & geomFile);

		virtual void ReadOff(const VKString & filename, OffFormat format = OFF_AUTO);
		virtual void WriteOff(const VKString & filename, OffFormat format = OFF_BIN_SINGLE);
		virtual void WriteObj(const VKString & filename);
		
		virtual void ReadPly(const VKString & filename, const VKString & readVals);
		virtual R3Mesh * CreateR3Mesh() const;
		
		virtual int AddVertex(double x, double y, double z, double nx=0, double ny=0, double nz=0);		
		virtual int AddFace(int v1, int v2, int v3);
		virtual void RemoveFace(int f);
		virtual void RemoveFaces(int f1, int f2);
		virtual void RemoveVertex(int v);

		// vertices
		virtual int NVertices() const;
		virtual R3Point VertexPosition(int vID) const;
		virtual R3Vector VertexNormal(int vID) const;
		virtual int FaceOnVertex(int vID) const;
		virtual int NAdjacentFaces(int vID) const;
		virtual int FaceOnVertex(int vID, int nFace) const;
		virtual std::vector<int> AdjacentVertices(int vID) const;
		
		// nedges 
		virtual int NEdges() const;
	
		// faces
		virtual int NFaces() const;
		virtual int VertexOnFace(int fID, int vID) const;
		virtual R3Point FacePoint(int fID, double b1, double b2) const;
		virtual R3Point FaceBarycentric(int fID, const R3Point & p) const;
		virtual R3Vector FaceNormal(int fID) const;
		virtual double FaceArea(int fID) const;
	
		virtual void RearrangeAxes(char newOrder[3], int newSigns[3]);
	
	protected:
		virtual void ReadOffASCII(const VKString & filename);
		virtual void ReadOffBinary(const VKString & filename, bool single);
		virtual void WriteOffASCII(const VKString & filename);
		virtual void WriteOffBinary(const VKString & filename, bool single);

		virtual R3MeshSearchTree * CreateSearchTree(R3Mesh ** mesh,
													std::vector<int> & triangleMap,
													std::vector<int> & vertexMap) const ;

	
		std::vector<MeshElementTriangle> m_Triangles;
		std::vector<MeshElementVertex> m_Vertices;
};

#endif

