/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifdef OPEN_COLLADA
#ifndef __Model3DColladaParser_H
#define __Model3DColladaParser_H

#include <vector>
#include <iostream>
#include <assert.h>

#include "COLLADAFW.h"
#include "COLLADASaxFWLLoader.h"

#include "VKString.h"
#include "ParamParser.h"
#include "ShapeMaterials.h"
#include "ShapeLabels.h"
#include "Model3DParser.h"

class Model3D;
class ShapeMesh;

class Model3DColladaParserStatic : public Model3DParser
{
	public:
		virtual void LoadModelFile(Model3D * model, const VKString & colladaFile,
								   const VKString & geometryName, int colladaReadFlags);

};

class Model3DColladaParser : public COLLADAFW::IWriter
{
	public:
		class NodeHierarchy
		{
			public:
				VKString m_sNodeName;
				
				COLLADAFW::UniqueId m_NodeID;
							
				R4Matrix m_XForm;
				
				std::vector<COLLADAFW::UniqueId> m_NodeIDs;
				std::vector<COLLADAFW::UniqueId> m_GeometryIDs;
				std::vector<COLLADAFW::UniqueId> m_ChildrenIDs;
				std::vector<std::map<COLLADAFW::MaterialId, COLLADAFW::UniqueId> > m_GeomBindinginInfo;
		};
		
		class GeometryIDs
		{
			public:
				struct Face
				{
					Face(int v1, int v2, int v3);
					int & operator [](int i);
					int pos[3];		// filled
					int uv[3];		// filled - but some todos
					int norm[3];	// filled
					int rgb[3];		// todo
					COLLADAFW::MaterialId materialID;	// filled
				};
			
			// note: norms, uvCoords, positions are not related
			// it's something that is decided at face level
				std::vector<R3Vector> m_Normals;
				std::vector<R3Point> m_Positions;
				std::vector<R2Point> m_UVCoords;
				std::vector<RNRgb> m_Colors;
				std::vector<Face> m_Faces;
			
				int m_iGeomListID;
				VKString m_sName;
		};
	
		class ColladaImage
		{
			public:
				R2Image * m_pImg;
				COLLADAFW::UniqueId m_ImgID;
		};
	
		class ColladaEffect
		{
			public:
				COLLADAFW::UniqueId m_EffectID;
				std::vector<COLLADAFW::UniqueId> m_Images;	
			
				COLLADAFW::EffectCommon::ShaderType m_Shader;
				COLLADAFW::ColorOrTexture m_EmitClr;
				COLLADAFW::ColorOrTexture m_DiffuseClr;
				COLLADAFW::ColorOrTexture m_AmbientClr;
				COLLADAFW::ColorOrTexture m_SpecularClr;
				double m_fShininess;
				COLLADAFW::ColorOrTexture m_ReflectiveClr;
				double m_Reflectivity;
				COLLADAFW::ColorOrTexture m_OpacityClr;
				double m_IndexOfRefraction;
		};
	
		Model3DColladaParser(const VKString & colladaDir, int colladaReadFlags);
		virtual ~Model3DColladaParser();
		
		virtual ShapeMesh * ColladaGeometry();
		virtual ShapeMaterials * ColladaMaterial();
		virtual ShapeLabels * ColladaSceneGraphLabels();
	
		virtual void AddColorOrTexture(const COLLADAFW::ColorOrTexture & cOrImg,
									   const ColladaEffect & effect,
									   ColorOrTexture & myCOrImg,
									   ShapeMaterials & materialList, 
									   std::map<COLLADAFW::UniqueId, int> & img2myID);
	
	// todo: read other objects (e.g. labels, textures)
		int m_iColladaReadFlags;
		ShapeMesh * m_pGeometry;
		ShapeMaterials * m_pMaterial;	// add material
		ShapeLabels * m_pSceneGraphLabels;	// default segmentation
		VKString m_ColladaDir;
	
		VKStringList m_UnimplementedStuffCalled;
		void UnimplementedCall(const VKString & name);
	
		virtual COLLADAFW::UniqueId AddNode(const COLLADAFW::Node * node);
		virtual SceneGraphLabel * TransformNode(const COLLADAFW::UniqueId nodeID,
												const R4Matrix & matrixStack);
		virtual int TransformGeometry(const COLLADAFW::UniqueId geometryID,
									std::map<COLLADAFW::MaterialId, COLLADAFW::UniqueId> & matMap,
									  const R4Matrix & matrixStack, int segmentID);
		
		virtual void CreateMaterials();
		virtual void ParseNodeHierarchy();

		std::vector<GeometryIDs> m_AllGeometries;
		std::map<COLLADAFW::UniqueId, int> m_mIDToGeometry;
		std::map<COLLADAFW::UniqueId, NodeHierarchy> m_mIDToNodeHierarchy;
		std::map<COLLADAFW::UniqueId, ColladaEffect> m_mEffectIDToMaterial;
		std::map<COLLADAFW::UniqueId, COLLADAFW::UniqueId> m_mMaterialUniqueIDToEffectID;

		std::map<COLLADAFW::UniqueId, ColladaImage> m_mIDToImage;
	
		std::vector<COLLADAFW::UniqueId> m_RootVisualNodes;
		std::vector<COLLADAFW::UniqueId> m_RootLibraryNodes;
	
		std::vector<R2Point> m_GlobalUVs;
		std::vector<R3Vector> m_GlobalNormals;
		std::vector<RNRgb> m_GlobalColors;
		std::vector<COLLADAFW::UniqueId> m_FaceToMaterial;
		std::vector<int> m_FaceToUVs;
		std::vector<int> m_FaceToNormals;
		std::vector<int> m_FaceToColors;
	
		std::vector<int> m_FaceToSceneGraphLeaf;
		std::vector<SceneGraphLabel*> m_AllSceneGraphNodes;
	
	// inherited:
		virtual void cancel(const COLLADAFW::String& errorMessage);		
		virtual void start();
		virtual void finish();
		virtual bool writeGlobalAsset ( const COLLADAFW::FileInfo* asset );		
		virtual bool writeScene ( const COLLADAFW::Scene* scene );		
		virtual bool writeVisualScene ( const COLLADAFW::VisualScene* visualScene );
		virtual bool writeLibraryNodes ( const COLLADAFW::LibraryNodes* libraryNodes );
		virtual bool writeGeometry ( const COLLADAFW::Geometry* geometry );
		virtual bool writeMaterial( const COLLADAFW::Material* material );
		virtual bool writeEffect( const COLLADAFW::Effect* effect );
		virtual bool writeCamera( const COLLADAFW::Camera* camera );
		virtual bool writeImage( const COLLADAFW::Image* image );
		virtual bool writeLight( const COLLADAFW::Light* light );
		virtual bool writeAnimation( const COLLADAFW::Animation* animation );
		virtual bool writeAnimationList( const COLLADAFW::AnimationList* animationList );
		virtual bool writeSkinControllerData( const COLLADAFW::SkinControllerData* skinControllerData );
		virtual bool writeController( const COLLADAFW::Controller* controller );
		virtual bool writeFormulas( const COLLADAFW::Formulas* formulas );
		virtual bool writeKinematicsScene( const COLLADAFW::KinematicsScene* kinematicsScene );	
	
};

#endif
#endif
