/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __CorrsAffineAlign_H
#define __CorrsAffineAlign_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "VKString.h"

#include "GeomShape.h"
#include "ShapePoint.h"
#include "ShapeCorrs.h"
#include "CorrsPair.h"
#include "CorrsCreator.h"

class AffineCorrsCreator : public CorrsCreator
{
	public:
		AffineCorrsCreator();
		virtual CorrsPair * ReadCorrsPair(const VKString & corrName, 
										  GeomShape * shapeFrom, GeomShape * shapeTo, 
										  FILE * file, DBCorrs * dbCorrs);
		virtual void CreateCorrsPair(const VKString & corrName, GeomShape * shapeFrom, GeomShape * shapeTo, 
									 CorrParams * params, std::vector<CorrsPair *> & saveCorrs, 
									 DBCorrs * dbCorrs);

};

class ParamsAffineAlign : public CorrParams
{
	public:
		ParamsAffineAlign(const VKString & paramSettingName = "default");
	
		R3Vector * m_UpVector;
			// distance for default corr. value and radius of corr. points
		double m_fMaxDist;
		bool m_bICP_Translation, m_bICP_Rotation, m_bICP_Scale;
		bool m_bPCA_Translation, m_bPCA_Rotation, m_bPCA_Scale;
};

/** 
 * CorrsAffineAlign
 */
class CorrsAffineAlign : public CorrsPair
{
	public:
	// static functions: create affine alignment
		static void CreateCorrs(const VKString & corrName,
								ShapePointCloud * pnts1, ShapePointCloud * pnts2,
								ParamsAffineAlign * params, 
								std::vector<CorrsAffineAlign* > & corrs);
	
		/** NOTE: upVector can be NULL */
		static void CreateCorrsFromPCA(const VKString & corrName,
									   ShapePointCloud * pnts1, ShapePointCloud * pnts2,
									   double distNorm, R3Vector * upVector,
									   bool pca_translation, bool pca_rotation, bool pca_scale,
									   bool icp_translation, bool icp_rotation, bool icp_scale,
									   std::vector<CorrsAffineAlign* > & corrs);

		static void GetPCAAlignments(const R3Point & center1, double rad1, const R3Triad & axes1, 
									 const R3Point & center2, double rad2, const R3Triad & axes2, 
									 bool pca_translation, bool pca_rotation, bool pca_scale,
									 int upAxis, std::vector<R3Affine> & affine12s);

		static void GetPCAAlignments(ShapePointCloud * pnts1, ShapePointCloud * pnts2,
									 bool pca_translation, bool pca_rotation, bool pca_scale,
									 R3Vector * upVector, std::vector<R3Affine> & affine12s);

		static bool RefineWithICP(ShapePointCloud * pnts1, ShapePointCloud * pnts2, R3Affine & affine12,
								  bool icp_translation, bool icp_rotation, bool icp_scale, R3Vector * upVector);
	
		static double RMSD(ShapePointCloud * pnts1, ShapePointCloud * pnts2,
						   const R3Affine & affine12, const R3Affine & affine21);
	
		static R4Matrix R3AlignInXY(int npoints, R3Point *points1, R3Point *points2, RNScalar *weights,
									RNBoolean align_center, RNBoolean align_rotation, int align_scale);
		
	
		static int CreateP2PCorrs(ShapePointCloud * pnts1, ShapePointCloud * pnts2,
								  const R3Affine& affine12, 
								  R3Point *correspondences1, R3Point *correspondences2, 
								  int max_correspondences); 
		static R4Matrix InverseMatrix(const R4Matrix & affine);
		static R3Affine InverseAffine(const R3Affine & affine);

	// members: use once alignment is known
		// this is for reading
		CorrsAffineAlign(const VKString & corrName,
						 GeomShape * shapeFrom, GeomShape * shapeTo);

		// this class 
		CorrsAffineAlign(const VKString & corrName,
						 GeomShape * shapeFrom, GeomShape * shapeTo, 
						 double distNorm, R3Affine affine12 = R4identity_matrix);

		virtual ~CorrsAffineAlign();
		
		virtual VKString GetCorrsClassName() const;
		
		virtual int GetCorrs(const ShapePoint & pnt,
							 std::vector<ShapePoint> & corrs, 
							 const VKString & corrValName="none",
							 std::vector<double> *corrVals = NULL);

		virtual ShapePoint BestCorr(const ShapePoint & pnt,
									GeomShape * toShape,
									const VKString & corrValName = "default",
									double * corrValue = NULL);	
		
		virtual double DefaultIntegratedValue(const VKString & evalPnts = "none") const;
	
		const R3Affine & GetTransformation() const;
	
	
		//CorrsAffineAlign * CreateInverse() const;
		virtual void WriteShapeCorrs(FILE * file);
		virtual void ReadShapeCorrs(FILE * file);
	

	protected:
		double m_fIntegratedValue;
		double m_fDistNorm;
		R3Affine m_Affine12;
	
		static R3Point * m_CorrBuffer1;
		static R3Point * m_CorrBuffer2;
		static int m_iCorrBufferSize;
};

#endif

