#include "ConsistentPartsAlg.h"
#include "Clustering.h"
#include "STemplateNode.h"


PointsCluster::PointsCluster()
{
	m_iModelID = -1;
	m_iClusterID = -1;
	m_iLocalNodeID = -1;
	
	//m_pSubsetKdTree = NULL;
	m_Pnts2Tmplt = R3identity_affine;
	m_pAllPnts = NULL;
	m_pGeomFit = NULL;
	m_sLabel = "none";
	m_pAlignedSubsetPoints = NULL;
	m_pLocalOrient = NULL;
}

PointsCluster::PointsCluster(int modelID, int clusterID, int localNodeID, const VKString & label, const ShapePointCloud * pnts)
{
	m_iModelID = modelID;
	m_iClusterID = clusterID;
	m_iLocalNodeID = localNodeID;
	m_sLabel = label;
	
	m_pAllPnts = pnts;
	//m_pSubsetKdTree = NULL;
	m_pGeomFit = NULL;
	m_pAlignedSubsetPoints = NULL;

	m_Pnts2Tmplt = R3identity_affine;
	m_pLocalOrient = NULL;
}

PointsCluster::~PointsCluster()
{
//	if (m_pAlignedSubsetPoints!=NULL)
//		delete m_pAlignedSubsetPoints;
//	if (m_pLocalOrient!=NULL)
//		delete m_pLocalOrient;
}

void PointsCluster::CreateAlignedKdTree(const std::vector<int> & subset, const ShapePointCloud * pnts, TMatch * match)
{
	assert(m_pLocalOrient==NULL && m_pAlignedSubsetPoints==NULL && pnts!=NULL && match!=NULL);
	m_pAlignedSubsetPoints = new ShapePointCloud(NULL, "alignedpnts");
	for (int i=0; i<(int)subset.size(); i++)
	{
		R3Point pnt = match->AlignedToNodePointPosition(m_iLocalNodeID, subset[i], 0);
		m_pAlignedSubsetPoints->AddPoint(pnt);
	}
	//std::cout<<"SubsetSize = "<<subset.size()<<std::endl;
	m_pLocalOrient = new ShapeLocalPCA((GeomShape*)m_pAlignedSubsetPoints, "locpca", .2,
									   PartLocalGeometry::GetNhdSizeConstant(), true);
	m_pLocalOrient->ComputePntsPCA();
}

ConsistentPartsAlg::ConsistentPartsAlg(STemplate * st, const std::vector<TMatchEnergy*> & matches,
								 int methodFlags, bool verb)
: m_pST(st), m_Matches(matches), m_iMethodFlags(methodFlags), m_bVerbose(verb)
{
	
}

ConsistentPartsAlg::~ConsistentPartsAlg()
{
	
}


void ConsistentPartsAlg::MRFPointClustering(int matchID, int localNodeID, std::vector<int> & allpnts)
{
	assert(matchID>=0 && matchID < (int)m_ClusteredPoints.size());
	TMatchEnergy * energy = m_Matches[matchID];
	TMatch * match = energy->GetMatch();
	ShapePointCloud * pnts = (ShapePointCloud*)match->GetMatchedGeomAsPnts();
	assert(pnts!=NULL);
	
	ShapeLocalPCA * locPCA = new ShapeLocalPCA(pnts, "locPCA", .1, PartLocalGeometry::GetNhdSizeConstant(), true, allpnts);
	locPCA->ComputePntsPCA();
	R3Kdtree<const PointInSet *> * kdTree = pnts->CreateSubsetKdTree(allpnts);
	
	int NPoints = (int)allpnts.size();
	if (NPoints==0)
		return;
	int NClusters = -1;
	
	std::map<int, int> global2nullpnt;
	for (int i=0; i<NPoints; i++)
		global2nullpnt[allpnts[i]] = i;
	std::vector<int> clusterCenters;
	std::vector<std::vector<int> > clusterCenterPoints;

	if (m_iMethodFlags & CPM_InitWithMRF_TakeFurthest)	// take furthest to initialize clusters
	{
		assert(false);	// keep it around possibly useful for topological changes
//		NClusters = NPoints/10;
//		
//		// select NClusters farthest points to seed initial clusters
//		std::map<int, bool> picked;
//		clusterCenters.push_back(0);
//		picked[0] = true;
//		for (int i=0; i<NClusters; i++)
//		{
//			double maxDist = -1;
//			int pID = -1;
//			for (int p=0; p<(int)allpnts.size(); p++)
//			{
//				if (picked.find(p)!=picked.end())
//					continue;
//				double minDist = -1;
//				for (int c=0; c<(int)clusterCenters.size(); c++)
//				{
//					R3Point cp = pnts->PointPosition(allpnts[clusterCenters[c]]);
//					R3Point np = pnts->PointPosition(allpnts[p]);
//					double dist = R3Distance(cp, np);
//					minDist = vkMin(dist, minDist);
//				}
//				if (pID==-1 || minDist > maxDist)
//				{
//					maxDist = minDist;
//					pID = p;
//				}
//			}
//			clusterCenters.push_back(pID);
//			picked[pID] = true;
//		}
	}
	else if (m_iMethodFlags & (CPM_InitWithMRF_AssignToNonNull | CPM_InitGreedy_AssignToNonNull))
	{
		std::map<int, int> nodeID2ClusterID;
		for (int i=0; i<match->NTemplateNodes(); i++)
		{
			if (localNodeID !=i)
			{
				nodeID2ClusterID[i] =(int)clusterCenters.size();
				clusterCenters.push_back(i);
				clusterCenterPoints.push_back(std::vector<int>());
			}
		}
		
		for (int i=0; i<match->NShapePrimitives(); i++)
		{
			int lnid = match->LocalNodeID(match->Loc2GlobPid(i));
			std::map<int, int>::iterator iter = nodeID2ClusterID.find(lnid);
			if (iter != nodeID2ClusterID.end())
				clusterCenterPoints[iter->second].push_back(match->Loc2GlobPid(i));
		}
		NClusters = (int)match->NTemplateNodes() - 1;
	}
	else
		assert(false);
	
	
	// nclusters
	assert(NClusters>=0);
	if (NClusters==0)
		return;
	else if (NClusters==1)
	{
		assert(m_ClusteredPoints[matchID].size()==0);
		
		VKString label = VKString("NewNode")+VKString::number(matchID)+".0";
		m_ClusteredPoints[matchID].push_back(PointsCluster(matchID, 0, localNodeID, label, pnts));
		
		for (int i=0; i<(int)allpnts.size(); i++)
			m_ClusteredPoints[matchID][0].m_aSubset.push_back(allpnts[i]);
		return;
	}
	
	// Fill MRF problem
	// data term = distance to N furthest points
	double e1[3];
	double e2[3];
	double * data = new double[NPoints * NClusters];
	for (int i=0; i<NPoints; i++)
	for (int j=0; j<NClusters; j++)	
	{
		assert(j < (int)clusterCenters.size());
		int p1 = allpnts[i];
		R3Triad o1 = locPCA->GetPCA(pnts->GetPoint(p1), e1);
		int p2 = -1;
		if (m_iMethodFlags & CPM_InitWithMRF_TakeFurthest)
			p2 = allpnts[clusterCenters[j]];
		else if (m_iMethodFlags & (CPM_InitWithMRF_AssignToNonNull|CPM_InitGreedy_AssignToNonNull))
		{
			// find nearest point in cluster
			assert(j<(int)clusterCenterPoints.size());
			double mindist = -1;
			for (int k=0; k<(int)clusterCenterPoints[j].size(); k++)
			{
				double dist = R3Distance(pnts->PointPosition(p1), pnts->PointPosition(clusterCenterPoints[j][k]));
				if (p2 == -1 || dist < mindist)
				{
					p2 = clusterCenterPoints[j][k];
					mindist = dist;
				}
			}
		}
		
		if (p2>=0)
		{
			R3Triad o2 = locPCA->GetPCA(pnts->GetPoint(p2), e2);
			data[i * NClusters + j] = pow(R3Distance(pnts->PointPosition(p1), pnts->PointPosition(p2)), 2.);
		}
		else
			data[i * NClusters + j] = 100.;	// vk: const - kind of infinity
		assert(data[i * NClusters + j]>=0);
		//std::cout<<"Data ("<<i<<" "<<j<<") = "<<data[i * NClusters + j]<<std::endl;
		//data[i * NClusters + j] = energy->ComputeEPnt(pnts->PointPosition(p1), o1, e1, pnts->PointPosition(p2), o2, e2, 1.);
	}
	
		
	// smoothness - orientation similarity
	
	const VKGraph * smoothness = NULL;
	const double * labels2labels = NULL;
	if ((m_iMethodFlags & (CPM_InitWithMRF_AssignToNonNull|CPM_InitWithMRF_TakeFurthest))!=0)
	{
		assert((m_iMethodFlags & CPM_InitWithMRF_TakeFurthest)==0);
		smoothness = energy->GetEdgeEnergy();
		labels2labels = energy->GetLabelEnergy();
	}

		
	MRFProblemArray * mrfProblem = NULL;
	MRFOptimizer * mrfOptimizer = NULL;
	
	if (NPoints>1 && (m_iMethodFlags & (CPM_InitWithMRF_AssignToNonNull|CPM_InitWithMRF_TakeFurthest)))
	{
		mrfProblem = new MRFProblemArray(NPoints, NClusters, smoothness, data, labels2labels, 1000.);
		mrfOptimizer = new MRFOptimizer(mrfProblem);
		// Optimize w/ graph cuts
		mrfOptimizer->Optimize();
	}

	
	// fill m_ClusteredPoints
	std::map<int, int> labelID2clusterID;
	for (int i=0; i<NPoints; i++)
	{
		int lID = -1;
		if (NPoints>1 && (m_iMethodFlags & (CPM_InitWithMRF_AssignToNonNull|CPM_InitWithMRF_TakeFurthest)))
		{
			assert(mrfOptimizer!=NULL);
			lID = mrfOptimizer->Label(i);
		}
		else if (NPoints==1 || (m_iMethodFlags & CPM_InitGreedy_AssignToNonNull))
		{
			for (int c=0; c<NClusters; c++)
				if (lID == -1 || data[i * NClusters + c] < data[i * NClusters + lID] )
					lID = c;
		}
		else
			assert(false);
		assert(lID>=0);
		
		int cID = -1;
		if (labelID2clusterID.find(lID)==labelID2clusterID.end())
		{
			labelID2clusterID[lID] = (int)m_ClusteredPoints[matchID].size();
			cID = (int)m_ClusteredPoints[matchID].size();
			VKString label = VKString("NewNode")+VKString::number(matchID)+"."+VKString::number(cID);
			
			if (m_iMethodFlags & (CPM_InitWithMRF_AssignToNonNull|CPM_InitGreedy_AssignToNonNull))	// create indicative label
			{
				assert(lID>=0 && lID<(int)clusterCenters.size());
				const TemplateNodeDeform & deform = match->Deformation(clusterCenters[lID]);
				label = VKString("AttachedTo_")+VKString::number(clusterCenters[lID])+"_"+VKString::number(deform.NodeID())+"_";
				label += VKString::number(deform.PartID())+"_"+VKString::number(deform.NParts())+"_";
				label += match->GetSTemplate()->GetNode(deform.NodeID())->NodeName();
			}
			
			m_ClusteredPoints[matchID].push_back(PointsCluster(matchID, cID, localNodeID, label, pnts));
		}
		else
			cID = labelID2clusterID[lID];
		m_ClusteredPoints[matchID][cID].m_aSubset.push_back(allpnts[i]);
	}
	
	// a bit ad-hoc - but remove too small clusters (should re-cluster those points)
	if (m_iMethodFlags & CPM_InitWithMRF_TakeFurthest)
	{
		int minpnts = 5;	// vk: const
		
		for (int i=0; i<(int)m_ClusteredPoints[matchID].size(); i++)
		{
			if (m_ClusteredPoints[matchID][i].m_aSubset.size() < minpnts)
			{
				m_ClusteredPoints[matchID].erase(m_ClusteredPoints[matchID].begin() + i);
				i--;
			}
		}
	}
//	delete smoothness;
	delete data;
//	delete labels2labels;
	if (mrfProblem!=NULL)
		delete mrfProblem;
	if (mrfOptimizer!=NULL)
		delete mrfOptimizer;
	delete locPCA;
	delete kdTree;
}

void ConsistentPartsAlg::InitializeClusters(int nodeID, int partID, int nPartsForSelectedNode)
{
	// remember points w/ NULL label
	std::vector<std::vector<int> > pntsInCluster;
	std::vector<TMatchEnergy*> & ncmatches = (std::vector<TMatchEnergy*> &)m_Matches;
	
	std::vector<int> localIDs;
	
	// find points in given cluster
	for (int i=0; i<(int)m_Matches.size(); i++)
	{
		localIDs.push_back(-1);
		pntsInCluster.push_back(std::vector<int>());
		TMatch * match = m_Matches[i]->GetMatch();
		assert(m_pST == match->GetSTemplate());
		const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
		assert(pnts!=NULL);
		for (int p=0; p<pnts->NumPoints(); p++)
		{
			int lnid = match->LocalNodeID(p);
			assert(lnid>=0);
			const TemplateNodeDeform & deform = match->Deformation(lnid);

			if (deform.NodeID()==nodeID && deform.NParts()==nPartsForSelectedNode && deform.PartID()==partID)
			{
				pntsInCluster[i].push_back(p);
				if (localIDs[i] == -1)
					localIDs[i] = lnid;
				else
					assert(localIDs[i] == lnid);
			}
		}
	}
		
	// find cluster nodeID points
	int nParts = 0;
	for (int i=0; i<(int)m_Matches.size(); i++)
	{
		TMatch * match = m_Matches[i]->GetMatch();
		const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
		assert(pnts!=NULL);
		m_ClusteredPoints.push_back(std::vector<PointsCluster>());
		if (m_iMethodFlags & CPM_InitWhole)
		{
			if (pntsInCluster[i].size()>0)
			{
				m_ClusteredPoints[i].push_back(PointsCluster(i, 0, localIDs[i], m_pST->GetNode(nodeID)->NodeName(), pnts));
				for (int p=0; p<(int)pntsInCluster[i].size(); p++)
					m_ClusteredPoints[i][0].m_aSubset.push_back(pntsInCluster[i][p]);
			}
		}
		else if (m_iMethodFlags & CPM_InitFromGT)
		{
			std::map<int, int> gtID2cluster;
			const ShapeValue * gtLabelsVals = pnts->GetShapeVal(ShapeValue::m_sDefaultSceneGraphLabels);
			assert(gtLabelsVals!=NULL);
			const ShapeLabels * gtLabels = gtLabelsVals->SafeCastToLabels();
			assert(gtLabels!=NULL);
			
			for (int p=0; p<(int)pntsInCluster[i].size(); p++)
			{
				int gtID = (int)gtLabels->Value(pnts->GetPoint(pntsInCluster[i][p]), 0);
				int clusterID = -1;
				std::map<int, int>::iterator iter = gtID2cluster.find(gtID);
				if (iter == gtID2cluster.end())
				{
					clusterID = (int)m_ClusteredPoints[i].size();;
					gtID2cluster[gtID] = clusterID;
					m_ClusteredPoints[i].push_back(PointsCluster(i, clusterID, localIDs[i], gtLabels->NodeLabel(gtID), pnts));
				}
				else
					clusterID = iter->second;
				assert(clusterID>=0);
				
				assert(clusterID < (int)m_ClusteredPoints[i].size());
				m_ClusteredPoints[i][clusterID].m_aSubset.push_back(pntsInCluster[i][p]);
			}
		}
		else if (m_iMethodFlags & (CPM_InitWithMRF_TakeFurthest
								   | CPM_InitWithMRF_AssignToNonNull | CPM_InitGreedy_AssignToNonNull))
			MRFPointClustering(i, localIDs[i], pntsInCluster[i]);
		else
		{
			std::cout<<"[ERROR] Unknown method for initializing points / models for clustering"<<std::endl;
			assert(false);
		}
		
		nParts += (int)m_ClusteredPoints[i].size();
		for (int j=0; j<(int)m_ClusteredPoints[i].size(); j++)
		{
			m_MatchIDsForParts.push_back(i);
			m_ClusterIDsForParts.push_back(j);
		}
	}
		
	//ShapeLocalPCA * lpca = NULL;
	
	// fit primitives
	for (int i=0; i<(int)m_ClusteredPoints.size(); i++)
	{
		TMatch * match = m_Matches[i]->GetMatch();
		const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
		assert(pnts!=NULL);
		
		for (int c=0; c<(int)m_ClusteredPoints[i].size(); c++)
		{
			assert(pnts==m_ClusteredPoints[i][c].m_pAllPnts);
//			R3Point center;
//			R3Triad orient;
//			R3Vector scale;
//			double fitErr;

			// if need distances....
			if (m_iMethodFlags & CPM_ClusterSpectralAgglomerative)
			{
				std::vector<int> & subset = m_ClusteredPoints[i][c].m_aSubset;
				assert(subset.size()>0);
				m_ClusteredPoints[i][c].CreateAlignedKdTree(subset, pnts, match);
				assert(m_ClusteredPoints[i][c].m_pAlignedSubsetPoints != NULL);
			}
//			PartBox::FindOptimalBoxFit(pnts, m_ClusteredPoints[i][c].m_aSubset, center, orient, scale, fitErr);
//			if (fitErr>=0.05)	// vk const
//			{
//				m_ClusteredPoints[i].erase(m_ClusteredPoints[i].begin()+c);
//				
//				for (int rmID = 0; rmID < (int)m_MatchIDsForParts.size(); rmID++)
//				{
//					if  (m_MatchIDsForParts[rmID]==i && m_ClusterIDsForParts[rmID] == c)
//					{
//						m_MatchIDsForParts.erase(m_MatchIDsForParts.begin()+rmID);
//						m_ClusterIDsForParts.erase(m_ClusterIDsForParts.begin()+rmID);
//						rmID--;
//					}
//					else if (m_MatchIDsForParts[rmID]==i && m_ClusterIDsForParts[rmID] > c)
//					{
//						m_ClusterIDsForParts[rmID]--;
//					}
//				}
//				
//				c--;
//				continue;
//			}
//			m_ClusteredPoints[i][c].m_pGeomFit = new PartBox(((STemplate*)match->GetSTemplate())->GetAllParts());
//			m_ClusteredPoints[i][c].m_pGeomFit->InitBox(center, orient, scale, fitErr);
//			m_ClusteredPoints[i][c].m_pGeomFit->ApplyTransform(match->GlobalXForm());
		}
	}
}

void ConsistentPartsAlg::ExtractConsistentParts(int nodeID, int partID, int nParts, std::vector<PointsCluster> & finalClusters)
{
	TimeProfiler timer;
	timer.startedProcess("initc");
	InitializeClusters(nodeID, partID, nParts);
	timer.finishedProcess("initc");
	
	if ((m_iMethodFlags & CPM_ClusterTakeAll)!=0)
	{
		if (m_bVerbose)
			std::cout<<"Extracting Consistent parts: take all"<<std::endl;

			// simple debug: add all new parts
		for (int i=0; i<(int)m_ClusteredPoints.size(); i++)
		for (int c=0; c<(int)m_ClusteredPoints[i].size(); c++)
			finalClusters.push_back(m_ClusteredPoints[i][c]);
	}
	else if ((m_iMethodFlags & CPM_ClusterSpectralAgglomerative) != 0)	// agglomerative spectral clustering
	{
		if (m_bVerbose)
			std::cout<<"Extracting Consistent parts: agglomerative clustering"<<std::endl;

		int nParts = (int)m_ClusterIDsForParts.size();

		LinAlgVectorReal * evals = NULL;
		LinAlgMatrixReal * evecs = NULL;
		LinAlgMatrixReal * simmtx = NULL;
		bool row_normalize = true;
		ComputeEigs(&evals, &evecs, &simmtx, row_normalize);
		if (m_bVerbose)
		{
			std::cout<<"Computed eigs for spectral agglomerative clustering of parts"<<std::endl;
			std::cout<<"Eigs = [ ";
			for (int i=0; i<(*evals).Rows(); i++)
				std::cout<<(*evals)(i)<<" ";
			std::cout<<"] "<<std::endl;
		}
		
		int nBlocks;
		for (nBlocks=0; nBlocks<nParts && (*evals)(nBlocks)>=.9; nBlocks++);
		nBlocks--;
		nBlocks = vkMin(vkMax(nBlocks, 3), nParts);
		//nBlocks = vkMin(nParts, 20);
		assert(nBlocks>0);

		if (m_bVerbose)
			std::cout<<"nParts = "<<nParts<<" nBlocks="<<nBlocks<<std::endl;

		// spectral clustering w/ k-means
			// define problem
		double diffTime = 5;
		int nFeatures = vkMin(nBlocks*2, nParts-1);
		ClusteringProblem cProblem(nParts, nFeatures);
		double * features = new double[nFeatures];
		for (int i=0; i<nParts; i++)
		{
			for (int f=0; f<nFeatures; f++)
				features[f] = (*evecs)(i, f+1) * pow((*evals)(f+1), diffTime);
			cProblem.SetFeatures(i, features);
		}
		
		int nExpectedClusters = vkMin(2*nBlocks, nParts);

		ClusterNearest cAlg(nExpectedClusters);
		cAlg.Cluster(cProblem);
		
		int nEls = 1;
		for (int i=0; i<cProblem.NClusters(); i++)
		{
			if (m_bVerbose)
				std::cout<<"Cluster["<<i<<"] = "<<cProblem.ClusterSize(i)<<std::endl;
			if (cProblem.ClusterSize(i) >= nEls)
			{
				const std::vector<ClusteredNode> & cluster = cProblem.GetCluster(i);
				std::vector<int> cids;
				for (int c=0; c<cluster.size(); c++)
					cids.push_back(cluster[c].id);
				int f0 = (int)finalClusters.size();
				AverageParts(cids, finalClusters);
				for (int j=f0; j<(int)finalClusters.size(); j++)
					finalClusters[j].m_iClusterID = i;
			}
		}
	}
	else
	{
		std::cout<<"[ERROR] Clustering algorithm is not specified. Flags="<<m_iMethodFlags<<std::endl;
		assert(false);
	}
	
	m_ClusteredPoints.clear();
	m_MatchIDsForParts.clear();
	m_ClusterIDsForParts.clear();
}

void ConsistentPartsAlg::EraseClusters(std::vector<int> & cids)
{
	std::sort(cids.begin(), cids.end());
	while(cids.size()>0)
	{
		int lid = (int)cids.size()-1;
		int rmid = cids[lid];
		cids.erase(cids.begin()+lid);
		m_MatchIDsForParts.erase(m_MatchIDsForParts.begin()+rmid);
		m_ClusterIDsForParts.erase(m_ClusterIDsForParts.begin()+rmid);
	}
}

void ConsistentPartsAlg::ComputeEigs(LinAlgVectorReal ** evals, LinAlgMatrixReal ** evecs,
								  LinAlgMatrixReal ** simmtx, bool row_normalize)
{
	int nParts = (int)m_ClusterIDsForParts.size();
	// create affinity matrix
	if (nParts > 1000)
		std::cout<<"[WARNING] Too many parts, solving matrix might be slow..."<<std::endl;
	assert(evals!=NULL && evecs!=NULL && simmtx!=NULL);

	if (*evals == NULL || (**evals).Rows() != nParts)
	{
		if (*evals!=NULL)
			delete *evals;
		*evals = new LinAlgVectorReal(nParts);
	}
	
	if (*evecs == NULL || (**evecs).Rows() != nParts || (**evecs).Cols() != nParts)
	{
		if (*evecs!=NULL)
			delete *evecs;
		*evecs = new LinAlgMatrixReal(nParts, nParts);
	}
	
	if (*simmtx == NULL || (**simmtx).Rows() != nParts || (**simmtx).Cols() != nParts)
	{
		if (*simmtx!=NULL)
			delete *simmtx;
		*simmtx = new LinAlgMatrixReal(nParts, nParts);
	}
	
	assert(*evals!=NULL && *evecs!=NULL && *simmtx!=NULL);
	
	std::vector<double> rowsum;
	double mindist = 10e10;
	double aveDist = 0;
	std::vector<double> dists;
	// pairwise distances
	for (int i=0; i<nParts; i++)
	for (int j=i+1; j<nParts; j++)
	{
		int m1 = m_MatchIDsForParts[i];
		int c1 = m_ClusterIDsForParts[i];
		int m2 = m_MatchIDsForParts[j];
		int c2 = m_ClusterIDsForParts[j];
		
		double dist = ClusterDistance(m_Matches[m1], m_ClusteredPoints[m1][c1], m_ClusteredPoints[m2][c2]);
	
		if (dist>=0)
		{
			(**simmtx)(i,j) = dist*dist;
			(**simmtx)(j,i) = dist*dist;
			mindist = vkMin(mindist, dist*dist);
			aveDist += dist;
			dists.push_back(dist);
		}
		else
		{
			(**simmtx)(i,j) = -1;
			(**simmtx)(j,i) = -1;
		}
	}
	for (int i=0; i<nParts; i++)
		(**simmtx)(i,i) = mindist;

	assert(dists.size()>0);
	aveDist = dists[(int)dists.size()/2];
	if (m_bVerbose)
		std::cout<<"Ave Dist = "<<aveDist<<std::endl;
	
	assert(aveDist>0);
	//aveDist /= nParts*nParts;
	double sigma2 = pow(aveDist, 2.);
	
	// pairwise affinity matrix
	for (int i=0; i<nParts; i++)
	for (int j=0; j<nParts; j++)
	{
		if ((**simmtx)(i,j)>=0)
			(**simmtx)(i,j) = exp(-(**simmtx)(i,j) / sigma2);
		else
			(**simmtx)(i,j) = 0;
	}
	
	if (row_normalize)
		(**simmtx).RowNormalize(NULL);
	
	(**simmtx).EigendecompositionWithMATLAB(**evals, **evecs, nParts, "deleteme");
}

void ConsistentPartsAlg::AverageParts(std::vector<int> & partIDs, std::vector<PointsCluster> & finalClusters)
{
	if (m_iMethodFlags & CPM_AveClusterTakeAny)
	{
		assert(partIDs.size()>0);
		int lid =(int)finalClusters.size();
		finalClusters.push_back(PointsCluster());
		finalClusters[lid] = m_ClusteredPoints[m_MatchIDsForParts[partIDs[0]]][m_ClusterIDsForParts[partIDs[0]]];
		
		if (m_bVerbose)
		{
			for (int i=0; i<(int)partIDs.size(); i++)
			{
				int id = m_ClusteredPoints[m_MatchIDsForParts[partIDs[i]]][m_ClusterIDsForParts[partIDs[i]]].m_iModelID;
				std::cout<<"    "<<m_Matches[id]->GetMatch()->ModelName().c_str();
			}
			std::cout<<std::endl;
		}
	}
	else if (m_iMethodFlags & CPM_AveClusterTakeAll)
	{
		assert(partIDs.size()>0);
		for (int i=0; i<partIDs.size(); i++)
		{
			int lid =(int)finalClusters.size();
			finalClusters.push_back(PointsCluster());
			finalClusters[lid] = m_ClusteredPoints[m_MatchIDsForParts[partIDs[i]]][m_ClusterIDsForParts[partIDs[i]]];
		}

	}
	else
	{
		std::cout<<"[ERROR] Unknown clustering method "<<std::endl;
		assert(false);
//		for (int i=0; i<partIDs.size(); i++)
//			finalClusters.push_back(m_ClusteredPoints[m_MatchIDsForParts[partIDs[i]]][m_ClusterIDsForParts[partIDs[i]]]);
	}
}


double ConsistentPartsAlg::ClusterDistance(const TMatchEnergy * e, const PointsCluster & c1, const PointsCluster & c2)
{
	if (c1.m_iModelID==c2.m_iModelID)
		return -1;
		
	assert(c1.m_pAlignedSubsetPoints!=NULL);
	assert(c2.m_pAlignedSubsetPoints!=NULL);

	double d1 = OneWayDistance(e, c1, c2);
	double d2 = OneWayDistance(e, c2, c1);
	return d1 * .5 + d2 * .5;
}

double ConsistentPartsAlg::OneWayDistance(const TMatchEnergy * e, const PointsCluster & c1, const PointsCluster & c2)
{
	// distance: c1 -> c2
	double dist = 0;
	double eigs1[3];
	double eigs2[3];
	
	TMatch * m1 = m_Matches[c1.m_iModelID]->GetMatch();
	TMatch * m2 = m_Matches[c2.m_iModelID]->GetMatch();
	
	assert(c2.m_pAlignedSubsetPoints!=NULL && c1.m_pAlignedSubsetPoints!=NULL);
	assert(c1.m_pLocalOrient!=NULL && c2.m_pLocalOrient!=NULL);
	const R3Kdtree<const PointInSet *> * kdtree2 = c2.m_pAlignedSubsetPoints->GetKdTree();
	
	assert((int)c1.m_aSubset.size()==c1.m_pAlignedSubsetPoints->NumPoints());
	for (int i=0; i<(int)c1.m_aSubset.size(); i++)
	{
		R3Point p1 = c1.m_pAlignedSubsetPoints->PointPosition(i);
		R3Triad o1 = c1.m_pLocalOrient->GetPCA(c1.m_pAlignedSubsetPoints->GetPoint(i), eigs1);

		const PointInSet * p2c = kdtree2->FindClosest(p1);
		assert(p2c!=NULL && p2c->id>=0 && p2c->id < (int)c2.m_aSubset.size());
		assert(p2c->id >=0 && p2c->id < c2.m_pAlignedSubsetPoints->NumPoints());
		R3Point p2 = c2.m_pAlignedSubsetPoints->PointPosition(p2c->id);
		R3Triad o2 = c2.m_pLocalOrient->GetPCA(p2c->GetModelPoint(), eigs2);
		assert(false);
		//dist += e->ComputeEPnt(p1, o1, eigs1, p2, o2, eigs2);
	}
	return dist / (int)c1.m_aSubset.size();
}

void ConsistentPartsAlg::FindConsistentParts(const LinAlgMatrixReal & pairwiseSim,
										  const LinAlgVectorReal & evals, const LinAlgMatrixReal & evecs,
										  std::vector<int> & saveIDs)
{
	int nParts = (int)m_ClusterIDsForParts.size();
	int eig0 = 0;
	int neigs = 1;
	neigs =vkMin(eig0 + neigs, nParts);
	int nAdd = -1;
	
//	std::cout<<"E-vals: [";
//	for (int i=0; i<nParts; i++)
//		std::cout<<evals(i)<<" ";
//	std::cout<<"]"<<std::endl;

//	std::cout<<"ev=[";
//	for (int i=0; i<4; i++)
//		std::cout<<evals(i)<<" ";
//	std::cout<<"... ] "<<std::endl;
	for (int e = eig0; e<neigs; e++)
	//int e = 0;
	{
		std::vector<Sortable> sortedMatches;
		for (int i=0; i<nParts; i++)
			sortedMatches.push_back(Sortable(-vkAbs(evecs(i, e)), NULL, i));
		std::sort(sortedMatches.begin(), sortedMatches.end());
		
		std::cout<<"     evec = ["<<e<<"] = [";
		for (int v=0; v<10; v++)
			std::cout<<evecs(sortedMatches[v].id, e)<<" ";
		std::cout<<"... ]"<<std::endl;
		
		//int nAdd = 1;
		if (nAdd>0)
		{
			for (int v=0; v<vkMin(nAdd, nParts); v++)
				saveIDs.push_back(sortedMatches[v].id);
		}
		else
		{
			int seedid = sortedMatches[0].id;
			double bestSim = pairwiseSim(seedid, seedid);
			assert(bestSim>0);
			int v = 0;
			std::vector<Sortable> candidates;
			candidates.push_back(Sortable(-bestSim, NULL, seedid));
			while ((v+1) < (int)sortedMatches.size() && sortedMatches[++v].value/sortedMatches[0].value > .05)
				candidates.push_back(Sortable(0, NULL, sortedMatches[v].id));
			
//			std::cout<<"     Candidates: "<<candidates.size()<<std::endl;
			
			/*//double simthresh = bestSim * .3;
			double simthresh = 0;
			double minsim = 0;
			while (candidates.size()>0)
			{
				//std::cout<<"      -- adding candidate: "<<candidates[0].value<<" <? "<<simthresh<<std::endl;
				if (-candidates[0].value < simthresh)
					break;
				
				saveIDs.push_back(candidates[0].id);
				for (int cnd=1; cnd<(int)candidates.size(); cnd++)
					candidates[cnd].value = vkMin(candidates[cnd].value, -pairwiseSim(candidates[0].id, candidates[cnd].id));
				
				candidates.erase(candidates.begin());
				std::sort(candidates.begin(), candidates.end());
			}*/
			for (int cid=0; cid<(int)candidates.size(); cid++)
				saveIDs.push_back(candidates[cid].id);
			assert(saveIDs.size()>0);
		}
	}
}




