/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __PartOfEdge_H
#define __PartOfEdge_H

#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapeDescriptor.h"

#include "ProbDistribution.h"
#include "PartTransforms.h"
#include "Sortable.h"
#include "STemplate.h"

class STemplateNode;

/** 
 * Helper class for ShapeTemplate
 *
 * Edge that describes scene graph hierarchy. 
 */
class PartOfEdge
{
	public:
		PartOfEdge(STemplateNode * parent, STemplateNode * child, STemplate * st);
		virtual ~PartOfEdge();
	
		virtual PartOfEdge * CreateCopy(STemplate * newST, std::map<int, int> & old2newnode,
										std::map<int, int> & new2oldnode, int configoffset) const;
	
	// get link info
		virtual STemplateNode * ParentTemplate();
		virtual const STemplateNode * ParentTemplate() const;
		virtual STemplateNode * ChildTemplate();
		virtual const STemplateNode * ChildTemplate() const;
	
	// stats on the link
		virtual int NConfigs() const;
		virtual const DiscreteProbDistribution * GetProbNParts(int configID) const;

		// mean part transformation
		virtual const double * GetMeanTransform(int configID, int nParts, int partID, int & flags, int & ndims) const;
		// log probability
		virtual double LogProb(int configID, int nParts, int partID, const double * sample, int xformflags) const;
		// variance (for visualization)
		virtual R3Vector PosVariance(int configID, int nParts, int partID, int & ancestor) const;
	
	
	// updating link
		virtual void SetNPartsAlwaysOne();
		virtual void UpdateNParts(int configID, int nparts, bool include);
		virtual void UpdateXForm(int configID, const STemplateNode * parent,
								 const STemplateNode * child, int nParts, int partID);
		virtual void UpdateXForm(int configID, const TemplateNodeDeform & deform, bool include);
		virtual void Reset();
		virtual void Normalize(const R3Point & centroid, double aveRadius);

	// read / write
		virtual void WritePartOfEdge(FILE * file) const;
		static PartOfEdge * ReadPartOfEdge(FILE * file, STemplate * nodes, STemplate::STemplateVersions version);
		virtual void DuplicateConfigs(int from);

//	protected:
		virtual const PartTransforms * GetTransform(int configID, int nParts, int nMinSamples, int * ancestor) const;
	
	// parents (part holonym) and children (part meronym)
		STemplateNode * m_pParent;
		STemplateNode * m_pChildTemplate;
	
		STemplate * m_pST;	// need to understand relationship between configs
		
	// part probabilities
		// number of parts
		std::vector<DiscreteProbDistribution *> m_apPartProb;
		DiscreteProbDistribution * m_pAlwaysOnePart;
	
		// part transformations
		std::vector<std::map<int, PartTransforms*> > m_apPartTransforms;
};

#endif

