/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __STemplate_H
#define __STemplate_H

#include "PartBox.h"
#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapeDescriptor.h"
#include "PartLocalGeometry.h"

class TMatch;
class STemplateNode;
class PartsCollection;
class TemplateNodeDeform;

/**
 *
 * NOTE: This a sandbox class for on-going research. 
 * 
 * ------
 * This class represents a template for a collection of 3D models 
 * or a part at some level of hierarchy. The main purpose of this 
 * template is to capture a deformation space, and describe small 
 * collections with a few degrees of freedom...
 * 
 */ 
class STemplate
{
	public:
		enum STemplateVersions
		{
			STMPLT_V0_1 = 1,
			STMPLT_V0_2 = 2,
			STMPLT_V0_3 = 3,
			STMPLT_V0_4 = 4,
		};
		// initialize template from a single segmentation / scene graph
		static STemplate * CreateTemplate(ShapeLabels * labels, bool trustLabels, bool deformTemplate, double fUseMeshThresh,
										  ShapeLabels * meshLabels);

		STemplate(ShapeLabels * labels);	// labels can be NULL
		virtual ~STemplate();
		virtual void MergeTemplate(const STemplate * st);
	
		virtual void UseMeshInParts(double thresh);
		const ShapeLabels * TrainingLabels() const;
		//virtual void Normalize();
		virtual void Normalize(const R3Point & centroid, double aveRad );
	
	// nodes
		virtual int NNodes() const;
		virtual const STemplateNode * GetNode(int nodeID) const;
		virtual void InitAndAddNewNode(int parentID, const VKString & label, const PartBox & geometry, int configID,
									   std::vector<TemplateNodeDeform*> * newDeforms = NULL);
		virtual void RemoveNode(int nodeID);

	// template configuration
		virtual void DeformEllipsoid(const TemplateNodeDeform & deform, PartBox & deformMe) const;
		virtual PartsCollection * GetAllParts();
	
	// seed (high-probability) template configurations (configID==0 => single most probable)
		virtual int NSeedConfigs() const;
		virtual TMatch * CreateSeedConfig(int configID, GeomShape * shape, const std::vector<int> & subset,
										  const R3Affine & shape2root) const;
	
		virtual R3Point TemplatePointInCanonicalSpace(const TMatch * match, int tmpltPntID) const;
		virtual R3Point TemplatePointInOtherConfig(int configID, const TMatch * match, int tmpltPntID) const;
	
		virtual const VKString & ConfigName(int configID) const;
		virtual int MaxDescribedInstances(int configID) const;
		
	// improve stats
		static LocalGeometryTrainingSet * CreateTrainingSet(const TMatch * match, std::vector<const PartLocalGeometry *> & cleanup);
		virtual void UpdateAllStats(const TMatch * match, bool include);
		virtual void AddOrReplaceSeed(int seedID, const VKString & name, TMatch * match, int seedParent);
		virtual void RemoveSeed(int seedID);
		virtual void RemoveExcludedDeformsFromSeeds();
		virtual const VKStringList & GetCreatedFromModelsList() const;
		
	// write template
		virtual void WriteTemplate(const VKString & filename) const;
		static STemplate * ReadTemplate(const VKString & filename, ModelDatabase * db);
		
		virtual void SetDeformAsRestPose(TMatch * match);
	
		virtual int NSamples(int configID) const;
		virtual int ConfigParent(int configID) const;
		virtual const std::vector<int> & GetConfigHierarchy() const;
	
		
		virtual void TrainClassifiers();
		virtual void TrainClassifier(int configID);
		virtual const double * LogProbsLocalDescriptors(int pntID, const TMatch * match) const;
		virtual const LocalGeometryClassifier * GetLocalClassifier(int configID) const;
	
	//protected:
		std::vector<LocalGeometryClassifier *> m_apLocalClassifiers;
	
		int m_iInitCoreSeeds;
	
		int m_iNTopologyChanges;
		//int m_iNDescribedInstances;
	
		std::vector<STemplateNode*> m_AllNodes;
		ShapeLabels * m_pGeomLabels;
		VKStringList m_CreatedFromModels;
	
		PartsCollection * m_pAllParts;
		double m_fUseMeshThresh;
	
		VKStringList m_sSeedNames;
		std::vector<TMatch *> m_SeedDeformations;
	
	// configuration hierarchy
		std::vector<int> m_ConfigParents;
	
	// root match for co-alignment
		TMatch * m_pRootMatch;
		std::vector<int> m_NodeID2RootMatchDeform;
	
		std::vector<TMatch*> m_CanonicalMatches;
};


#endif


