/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __STemplateNode_H
#define __STemplateNode_H

#include "PartBox.h"
#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapeDescriptor.h"
#include "STemplate.h"

class PartConnectorEdge;
class PartOfEdge;
class TMatch;
class TMatchAlg;
class TemplateNodeDeform;
class STemplate;

/**
 *
 * NOTE: This a sandbox class for on-going research. 
 * 
 * ------
 * This class represents a single node in a template for a collection of 3D models 
 * 
 */ 
class STemplateNode
{
	public:
	// initialize empty template (maybe a part)
		STemplateNode(const VKString & mainName, int nodeID, STemplate * st);
		virtual ~STemplateNode();
		virtual STemplateNode * CreateCopyNoEdges(STemplate * st, int partoffset, int configoffset, int newNodeID);
		virtual void CopyEdges(int partoffset, int configoffset, std::map<int, int> & old2newnode,
							   std::map<int, int> & new2oldnode, STemplate * stCopyFrom);
	
		
	// node info
		virtual const VKString & NodeName() const;
		virtual int NodeID() const;
		virtual void SetNodeID(int nodeID);
		virtual void SetName(const VKString & newName);	// used by algorithm / manual interaction
	
	// node geometry
		virtual const PartBox * GetPart(int configID) const;
		virtual void InitNullGeometry();
		virtual bool IsNull() const;
		//virtual const PartBox * PartGeometry(int geometryID) const;
		
	// update statistics (from another box)
		virtual void InitShapeAndXForm(int configID, const STemplateNode * parent, int nParts, int partID,
									   const ShapePointCloud * pnts, const std::vector<int> & ptsIDs, 
									   double fUseMeshThresh, const std::vector<int> & subsetFaces);
		virtual void InitShapeAndXForm(int configID, const STemplateNode * parent, int nParts, int partID,
									   const PartBox & geomPart);
		virtual void UpdateShapeAndXForm(int configID, const TemplateNodeDeform & deform, bool include);
		virtual void Normalize(const R3Point & centroid, double aveRadius);
		virtual void Reset();
		
		virtual void AddPartOfChild(STemplateNode * child, STemplate * st, int configID, bool nullbox);
		virtual int NPartOfChildren() const;
		virtual void RemoveChild(STemplateNode * child);	// USED ONLY IN INTERACTIVE MODE (to remove node)
		virtual const PartOfEdge * PartOfParent() const;
		virtual const PartOfEdge * PartOfChild(int cid) const;
		virtual PartOfEdge * PartOfParent();
		virtual PartOfEdge * PartOfChild(int cid);
	
		virtual int NDescribedInstances(int configID) const;
		virtual R3Vector GetMeanScale(int configID) const;
		virtual double LogProb(int configID, const R3Vector & sample) const;
		virtual R3Vector ScaleVariance(int configID, int & ancestor) const;
	
	// read / write
		virtual void WriteNode(FILE * file) const;
		static STemplateNode * ReadNode(FILE * file, STemplate * st, STemplate::STemplateVersions v);
		virtual void DuplicateConfigs(int from);
		virtual void WriteSubparts(FILE * file) const;
		virtual void ReadSubparts(FILE * file, STemplate * st, STemplate::STemplateVersions v);

//	protected:
		virtual const ContProbDistribution * GetScaleDistr(int configID, int nMinSamples, int * ancestor) const;
		STemplate * m_pST;
	
	// meta data
		VKString m_NodeName;
		int m_iNodeID;

	// geometry
		int m_iNDescribedInstances;
		bool m_bIsNull;	
		std::vector<PartBox *> m_apEllipsoids;
		std::vector<ContProbDistribution*> m_apPartScales;

	// relationships
		std::vector<PartConnectorEdge *> m_Connectors;	// part similarities (e.g. symmetry)
		std::vector<PartOfEdge *> m_SubpartTypes;	// part-of relationships (e.g. part hierarchy)
		PartOfEdge * m_pParentPart;	// part hierarchy
};


#endif


