/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __TemplateNodeDeform_H
#define __TemplateNodeDeform_H

#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapePointCloud.h"
#include "ShapeLocalPCA.h"

#include "PartBox.h"


/**
 * This class describes mapping for a single segment to a template tree
 */
class TemplateNodeDeform
{
	friend class STemplate;
		
	public:
		TemplateNodeDeform(int nodeID, int nParts, int partID,
						   const double * xform, int ndim, int type, const R3Vector & scale);
		TemplateNodeDeform(const TemplateNodeDeform & deform, std::map<int, int> * old2newNode);
		virtual ~TemplateNodeDeform();
	
		virtual TemplateNodeDeform * CreateCopy(std::map<int, int> * old2newNode = NULL) const;
	
	// for optimizer: deformation parameters
		virtual int NodeID() const;
		virtual int PartID() const;
		virtual int NParts() const;
		virtual void SetNodeID(int nodeID);// do not use (except in interactive updates)
	
		virtual const double * GetTransform(int & xformDims, int & xFormFlags) const;
		virtual void SetTransform(const double * newTransform);
		virtual bool IsIncluded() const;
		virtual void SetIncluded(bool included);
		virtual const R3Vector & GetScale() const;
		virtual void SetScale(const R3Vector & scale);
		virtual void SetDirtyFlag();
		
	// for space deformation: explain warp
		virtual R3Point Orig2Deformed(const R3Point & pnt) const;
		virtual R3Point Deformed2Orig(const R3Point & pnt) const;
		
		virtual R3Vector Orig2Deformed(const R3Vector & vec) const;
		virtual R3Vector Deformed2Orig(const R3Vector & vec) const;
		
		virtual R3Triad Orig2Deformed(const R3Triad & axes) const;
		virtual R3Triad Deformed2Orig(const R3Triad & axes) const;

	// read / write
		virtual void WriteNodeDeform(FILE * file) const;
		static TemplateNodeDeform * ReadNodeDeform(FILE * file);
	
//	protected:
	// DEFORMATION PARAMETERS
		// node id: never changes
		int m_iNodeID;
		int m_iNParts;
		int m_iPartID;
		
		// transformation
		int m_iXFormNDim; 
		int m_TransformType;
		double * m_fTransform;
		
		// part scale
		R3Vector m_PCAScale;
		
		// quick exclusion
		bool m_bIncluded;
		
	// DEFORMATION OF SPACE (cached)
		// parameters
		bool m_bXFormChanged;
		R3Affine m_AffineXForm;
};

class IdentityTemplateDeform : public TemplateNodeDeform
{
	public:
		IdentityTemplateDeform();
	// disable all setters
		virtual void WriteNodeDeform(FILE * file) const;
		virtual void SetTransform(double * newTransform);
		virtual void SetIncluded(bool included);
		virtual void SetScale(const R3Vector & scale);
	
		static IdentityTemplateDeform s_IdentityDeform;

};

#endif
