#include "RendererGLUT.h"

RendererGLUT * RendererGLUT::s_MainGLUTWindow = NULL;

RendererGLUT::RendererGLUT(int argc, char ** argv, VisWindow * window)
: VisWindowRenderer(window)
{
#ifdef MESA_RENDERING
	std::cout<<"[WARNING] Mesa rendering is enabled - do cannot use glut..."<<std::endl;
#else
	assert(s_MainGLUTWindow==NULL);
	s_MainGLUTWindow = this;
	m_GLUTwindowID = 0;
	m_GLUTmouse[0] = 0;		m_GLUTmouse[1] = 0;	
	m_GLUTmodifiers = 0;
	
	// Open window 
	glutInit(&argc, argv);
	
	glutInitWindowPosition(100, 100);
	glutInitWindowSize(m_pWindow->Width(), m_pWindow->Height());
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	m_GLUTwindowID = glutCreateWindow("Large3DDB Visualization (GLUT)");
	
	m_pWindow->InitializeGL();
	
	// Initialize GLUT callback functions
	glutDisplayFunc(GLUTRedraw);
	glutReshapeFunc(GLUTResize);
	glutKeyboardFunc(GLUTKeyboard);
	glutSpecialFunc(GLUTSpecial);
	glutMouseFunc(GLUTMouse);
	glutMotionFunc(GLUTMotion);
	
	glutMainLoop();
#endif
}

void RendererGLUT::RenderText(R2Point pnt, const VKString & text, 
							  double r, double g, double b, TextSize size) const
{
#ifdef MESA_RENDERING
	static bool warnOnce = true;
	if (warnOnce)
	{
		warnOnce = false;
		std::cout<<"[WARNING] Mesa rendering is enabled - cannot use glut..."<<std::endl;
	}
#else
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
//	glDisable(GL_DEPTH_TEST);
	
	// TODO: convert to screen coords
	// Draw text string s and position p
	glColor3d(r, g, b);
	glRasterPos2d(pnt[0], pnt[1]);
	
	const char * s = text.c_str();
	if (size==TEXT_SIZE_SMALL)
		while (*s) glutBitmapCharacter(	GLUT_BITMAP_HELVETICA_10 , *(s++));	
	else if (size==TEXT_SIZE_MEDIUM)
		while (*s) glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, *(s++));	
	else if (size==TEXT_SIZE_LARGE)
		while (*s) glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *(s++));	
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
#endif
}


void RendererGLUT::RenderText(R3Point pnt, const VKString & text, 
							  double r, double g, double b, TextSize size) const
{
#ifdef MESA_RENDERING
	static bool warnOnce = true;
	if (warnOnce)
	{
		warnOnce = false;
		std::cout<<"[WARNING] Mesa rendering is enabled - do cannot use glut..."<<std::endl;
	}
#else
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	
	// Draw text string s and position p
	glColor3d(r, g, b);
	glRasterPos3d(pnt[0], pnt[1], pnt[2]);
	
	const char * s = text.c_str();
	if (size==TEXT_SIZE_SMALL)
		while (*s) glutBitmapCharacter(	GLUT_BITMAP_HELVETICA_10 , *(s++));	
	else if (size==TEXT_SIZE_MEDIUM)
		while (*s) glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, *(s++));	
	else if (size==TEXT_SIZE_LARGE)
		while (*s) glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *(s++));	
#endif
	glEnable(GL_DEPTH_TEST);
}

void RendererGLUT::TerminateWindow()
{
#ifndef MESA_RENDERING
	glutDestroyWindow(m_GLUTwindowID);
#endif
}

void RendererGLUT::UpdateScreen()
{
#ifndef MESA_RENDERING
	glutPostRedisplay();
#endif
}

void RendererGLUT::GLUTRedrawCallback()
{
	m_pWindow->Draw();
#ifndef MESA_RENDERING
	glutSwapBuffers();
#endif
}

void RendererGLUT::GLUTResizeCallback(int w, int h)
{
	m_pWindow->Resize(w,h);
}

int RendererGLUT::GetSpecialKeys()
{
	int retVal = 0;
#ifndef MESA_RENDERING
	if ((m_GLUTmodifiers & GLUT_ACTIVE_SHIFT) != 0)
		retVal = retVal | VisWindow::SPECIAL_SHIFT;
	if ((m_GLUTmodifiers & GLUT_ACTIVE_CTRL) != 0)
		retVal = retVal | VisWindow::SPECIAL_CTRL;
	if ((m_GLUTmodifiers & GLUT_ACTIVE_ALT) != 0)
		retVal = retVal | VisWindow::SPECIAL_ALT;
#endif
	return retVal;
}

void RendererGLUT::GLUTKeyboardCallback(unsigned char key, int x, int y)
{
#ifndef MESA_RENDERING
	// Invert y coordinate
	y = m_pWindow->Height() - y;
	
	// Remember mouse position 
	m_GLUTmouse[0] = x;
	m_GLUTmouse[1] = y;
	
	// Remember modifiers 
	m_GLUTmodifiers = glutGetModifiers();
	m_pWindow->SetSpecialKey(GetSpecialKeys());
	m_pWindow->KeyPressed(key, x, y);	
#endif
}

void RendererGLUT::GLUTSpecialCallback(int key, int x, int y)
{
#ifndef MESA_RENDERING
	// Invert y coordinate
	y = m_pWindow->Height() - y;
	
	// Process keyboard button event 
	
	// Remember mouse position 
	m_GLUTmouse[0] = x;
	m_GLUTmouse[1] = y;
	
	// Remember modifiers 
	m_GLUTmodifiers = glutGetModifiers();
	m_pWindow->SetSpecialKey(GetSpecialKeys());
#endif
}

void RendererGLUT::GLUTMouseCallback(int button, int state, int x, int y)
{
#ifndef MESA_RENDERING
	bool valid;
	// Invert y coordinate
	y = m_pWindow->Height() - y;
	
	// Remember modifiers 
	m_GLUTmodifiers = glutGetModifiers();
	
	// Remember mouse position 
	m_GLUTmouse[0] = x;
	m_GLUTmouse[1] = y;
	
	// mouse button
	VisWindow::MouseButton bt;
	if (button == GLUT_LEFT_BUTTON)
		bt = VisWindow::MOUSE_LEFT;
	else if (button == GLUT_MIDDLE_BUTTON)
		bt = VisWindow::MOUSE_MIDDLE;
	else if (button == GLUT_RIGHT_BUTTON)
		bt = VisWindow::MOUSE_RIGHT;
	
	// mouse state
	VisWindow::MouseButtonState st;
	if (state == GLUT_DOWN)
		st = VisWindow::MOUSE_BT_DOWN;
	else
		st = VisWindow::MOUSE_BT_UP;
	m_pWindow->SetSpecialKey(GetSpecialKeys());
	m_pWindow->MouseButtonEvent(x, y, bt, st);
#endif
}

void RendererGLUT::GLUTMotionCallback(int x, int y)
{
#ifndef MESA_RENDERING
	// Invert y coordinate
	y = m_pWindow->Height() - y;
	
	// Compute mouse movement
	int dx = x - m_GLUTmouse[0];
	int dy = y - m_GLUTmouse[1];
	
	// Remember mouse position 
	m_GLUTmouse[0] = x;
	m_GLUTmouse[1] = y;
	m_pWindow->SetSpecialKey(GetSpecialKeys());
	m_pWindow->MouseMotionEvent(x, y, dx, dy);
#endif
}


///////// STATIC GLUT FUNCTIONS /////////
void GLUTRedraw()
{
	RendererGLUT::s_MainGLUTWindow->GLUTRedrawCallback();
}

void GLUTResize(int w, int h)
{
	RendererGLUT::s_MainGLUTWindow->GLUTResizeCallback(w, h);
}

void GLUTMotion(int x, int y)
{
	RendererGLUT::s_MainGLUTWindow->GLUTMotionCallback(x, y);
}

void GLUTMouse(int button, int state, int x, int y)
{
	RendererGLUT::s_MainGLUTWindow->GLUTMouseCallback(button, state, x, y);
}

void GLUTSpecial(int key, int x, int y)
{
	RendererGLUT::s_MainGLUTWindow->GLUTSpecialCallback(key, x, y);
}

void GLUTKeyboard(unsigned char key, int x, int y)
{
	RendererGLUT::s_MainGLUTWindow->GLUTKeyboardCallback(key, x, y);
}

