#include "RendererOSMesa.h"

RendererOSMesa::RendererOSMesa(VisWindow * window)
: VisWindowRenderer(window)
{
#ifdef MESA_RENDERING
	assert(m_pWindow!=NULL);
	
#if OSMESA_MAJOR_VERSION * 100 + OSMESA_MINOR_VERSION >= 305
	/* specify Z, stencil, accum sizes */
	m_MesaContext = OSMesaCreateContextExt( OSMESA_RGBA, 16, 0, 0, NULL );
#else
	m_MesaContext = OSMesaCreateContext( OSMESA_RGBA, NULL );
#endif
	if (!m_MesaContext) 
		std::cout<<"[ERROR] OSMesaCreateContext failed!"<<std::endl;;
	assert(m_MesaContext);
	int w = m_pWindow->Width();
	int h = m_pWindow->Height();
	m_aOffscreenBuffer = new unsigned char[4 * w * h];
	bool ok = OSMesaMakeCurrent(m_MesaContext, m_aOffscreenBuffer, GL_UNSIGNED_BYTE, w, h );
	assert(ok);
	
//	// DO NOT NEED z, s, a  - just checking
//	{
//		int z, s, a;
//		glGetIntegerv(GL_DEPTH_BITS, &z);
//		glGetIntegerv(GL_STENCIL_BITS, &s);
//		glGetIntegerv(GL_ACCUM_RED_BITS, &a);
//		printf("Depth=%d Stencil=%d Accum=%d\n", z, s, a);
//	}

	m_pWindow->InitializeGL();
#endif
}

void RendererOSMesa::UpdateScreen()
{
	m_pWindow->Draw();
}

void RendererOSMesa::TerminateWindow()
{
#ifdef MESA_RENDERING
	OSMesaDestroyContext( m_MesaContext );
	delete [] m_aOffscreenBuffer;
	m_aOffscreenBuffer = NULL;
#endif
}

void RendererOSMesa::CreateBuffer(BufferID & bufferID)
{
	bufferID.m_iWidth = m_pWindow->Width();
	bufferID.m_iHeight = m_pWindow->Height();
	bufferID.m_aColorBuffer = new unsigned char[4 * bufferID.m_iWidth * bufferID.m_iHeight];
}

void RendererOSMesa::LoadBuffer(BufferID & bufferID)
{
#ifdef MESA_RENDERING
	bool ok = OSMesaMakeCurrent(m_MesaContext, bufferID.m_aColorBuffer, GL_UNSIGNED_BYTE, 
								bufferID.m_iWidth, bufferID.m_iHeight);
	assert(ok);
#endif
}

void RendererOSMesa::FillDrawBuffer(BufferID & bufferID)
{
#ifdef MESA_RENDERING
	// should be ok...
#endif
}









