/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "DatabaseVisWindow.h"
#include "DBApp.h"
#include "ShapeValue.h"

VKString DatabaseVisWindow::m_sPairwisePrefix = "Pair_";
VKString DatabaseVisWindow::m_sCorrPrefix = "Corr_";

int DatabaseVisWindow::LIST_GROUPS_MODELS = (DatabaseVisWindow::LISTGP_Models
											 | DatabaseVisWindow::LISTGP_Geoms
											 | DatabaseVisWindow::LISTGP_GeomVals
											 | DatabaseVisWindow::LISTGP_MeshFlags);

int DatabaseVisWindow::LIST_GROUPS_CORRS = (DatabaseVisWindow::LISTGP_Corrs
											| DatabaseVisWindow::LISTGP_CorrVals
											| DatabaseVisWindow::LISTGP_CorrFlags);

DatabaseVisWindow::DatabaseVisWindow(ModelDatabase * db, DBCorrs * dbCorrs,
									 int listGPFlags,
									 int mainWndGridX, int mainWndGridY,
									 int nAdditionalGridX, int nAdditionalGridY, 
									 double xSeparator)
{	
	m_iDesignFlags = RENDER_DESIGN_BOXES | RENDER_MODEL_LIST | RENDER_BOXID;	

	m_iOrderedModelPage = 0;
	m_pApp = NULL;
	m_pDB = db;
	m_pDBCorrs = dbCorrs;
	m_pCorrsVis = NULL;
	m_iMainWndGridX = mainWndGridX;
	m_iMainWndGridY = mainWndGridY;
	m_iAdditionalWndGridX = nAdditionalGridX;
	m_iAdditionalWndGridY = nAdditionalGridY;
	m_fXSeparator = xSeparator;
	
	m_iFromWindowID = 0;
	int nTotal = (mainWndGridX*mainWndGridY + nAdditionalGridX*nAdditionalGridY);
	if (nTotal==1)
		m_iToWindowID = 0;
	else if (nTotal==2)
		m_iToWindowID = 1;
	else 
		m_iToWindowID = mainWndGridX*mainWndGridY + nAdditionalGridX*nAdditionalGridY/2;

	// create buttons menu
	m_pLists = new VisButtonsMenu();
	VisButtonsGroup::ButtonsGroupType radio = VisButtonsGroup::BUTGROUP_RadioButtons;
	VisButtonsGroup::ButtonsGroupType check = VisButtonsGroup::BUTGROUP_CheckboxButtons;
	
	bool nogeoms = false;
		// create geometries
	if (m_pDB!=NULL)
	{
		// list all geometries and models
		VKStringList geoms;
		VKStringList mdls;
		for (int i=0; i<m_pDB->NumModels(); i++)
		{
			Model3D * m = m_pDB->GetModel(i);
			m_DrawableModels.push_back(new VisModel3D(m, 0));
			//m_sAllModels.push_back(m->ModelName());
			mdls.push_back(m->ModelName());
			if (m->NGeometries()==0)
				nogeoms = true;
			for (int j=0; j<m->NGeometries(); j++)
			{
				GeomShape * geom = m->GetGeomShape(j);
				VKString geomName = geom->GeometryName();
				if (!geoms.contains(geomName))
					geoms.push_back(geomName);
			}
		}
		
		if (listGPFlags & LISTGP_Models)
		{
			AddGroup(LISTGP_Models, "m", radio);
			m_pLists->GetGroup(GroupID(LISTGP_Models))->AddOnOffButtons(mdls);
		}
		
		if (listGPFlags & LISTGP_Geoms)
		{
			AddGroup(LISTGP_Geoms, "g", check);
			m_pLists->GetGroup(GroupID(LISTGP_Geoms))->AddOnOffButtons(geoms);
		}
		if (listGPFlags & LISTGP_MeshFlags)
		{
			AddGroup(LISTGP_MeshFlags, "gf", check);
			VKStringList meshFlags;
			meshFlags.push_back("Draw Mesh Faces");			// DRAW_FACES
			meshFlags.push_back("Draw Mesh Edges");			// DRAW_EDGES
			meshFlags.push_back("Draw Mesh Vertices");		// DRAW_VERTICES
			meshFlags.push_back("Draw Pnts As Dots");		// DRAW_PDOTS
			meshFlags.push_back("Draw Pnts As Spheres");	// DRAW_PSPHERES
			meshFlags.push_back("Draw IDs");				// DRAW_IDs
			meshFlags.push_back("Draw Normals");			// DRAW_NORMALS
			meshFlags.push_back("Draw Materials");			// DRAW_MATERIALS
//			meshFlags.push_back("Draw Thumbnails");			// DRAW_THUMBNAILS
			meshFlags.push_back("Draw Meshes");				// DRAW_MESHES
			meshFlags.push_back("Draw Points");				// DRAW_POINTS
			meshFlags.push_back("Draw Flat");				// DRAW_FLAT
			m_pLists->GetGroup(GroupID(LISTGP_MeshFlags))->AddOnOffButtons(meshFlags);
			m_pLists->ChangeState(GroupID(LISTGP_MeshFlags), 0, 1);
			m_pLists->ChangeState(GroupID(LISTGP_MeshFlags), meshFlags.count()-3, 1);
			m_pLists->ChangeState(GroupID(LISTGP_MeshFlags), meshFlags.count()-2, 1);
			m_pLists->ChangeState(GroupID(LISTGP_MeshFlags), meshFlags.count()-1, 1);
			for (int i=0; i<(int)m_DrawableModels.size(); i++)
			{
				m_DrawableModels[i]->SetFlag(VisModel3D::DRAW_FACES, true);
				m_DrawableModels[i]->SetFlag(VisModel3D::DRAW_MESHES, true);
				m_DrawableModels[i]->SetFlag(VisModel3D::DRAW_POINTS, true);
				//m_DrawableModels[i]->SetFlag(VisModel3D::DRAW_FLAT, true);
			}
		}
		if (listGPFlags & LISTGP_GeomVals)
		{
			AddGroup(LISTGP_GeomVals, "gv", radio);
			VKStringList geomVals;
			std::vector<int> maxStates;
			for (int i=0; i<m_pDB->NumModels(); i++)
			{
				Model3D * m = m_pDB->GetModel(i);
				
				//m_DrawableModels.push_back(new VisModel3D(m, 0));
				//m_sAllModels.push_back(m->ModelName());
				mdls.push_back(m->ModelName());
				if (m->NGeometries()==0)
					nogeoms = true;
				
				for (int j=0; j<m->NGeometries(); j++)
				{
					GeomShape * geom = m->GetGeomShape(j);
					for (int k=0; k<geom->NShapeVals(); k++)
					{
						ShapeValue * shapeVal = geom->GetShapeVal(k);
						assert(shapeVal!=NULL);
						VKString shapeValName = shapeVal->GetShapeValueName();
						int id = geomVals.containsGetID(shapeValName);
						if (id==-1)
						{
							id = geomVals.count();
							geomVals.push_back(shapeValName);
							maxStates.push_back(2);
						}
						if (shapeVal->SafeCastToLabels()!=NULL)
						{
							int mDepth = shapeVal->SafeCastToLabels()->MaxDepth();
							maxStates[id] = vkMax(maxStates[id], mDepth+3);
						}
					}
				}
			}
			for (int i=0; i<geomVals.count(); i++)
				m_pLists->GetGroup(GroupID(LISTGP_GeomVals))->AddButton(geomVals[i], maxStates[i]);
			//m_pLists->GetGroup(GroupID(LISTGP_GeomVals))->AddOnOffButtons(geomVals);
		}
//		if (listGPFlags & LISTGP_GeomDists)
//		{
//			AddGroup(LISTGP_GeomDists, "d", radio);
//			// todo...
//		}
	}

		// create corrs	
	if (m_pDBCorrs!=NULL)
	{			
		int drawFlags = 0;
		if (listGPFlags & LISTGP_Corrs)
			AddGroup(LISTGP_Corrs, "c", radio);
		if (listGPFlags & LISTGP_CorrVals)
			AddGroup(LISTGP_CorrVals, "cv", radio);
		if (listGPFlags & LISTGP_CorrFlags)
		{
			AddGroup(LISTGP_CorrFlags, "cf", check);
			
			VKStringList corrFlags;
			corrFlags.push_back("Draw Corrs: Colored Points");
			corrFlags.push_back("Draw Corrs: Connect Lines");
			corrFlags.push_back("Draw Corrs: Value Transfer");
			corrFlags.push_back("Draw Corrs: Affine Alignment");
			corrFlags.push_back("Draw Corrs: Fuzzy Values");
			m_pLists->GetGroup(GroupID(LISTGP_CorrFlags))->AddOnOffButtons(corrFlags);
			m_pLists->ChangeState(GroupID(LISTGP_CorrFlags), 0, 1);
			m_pLists->ChangeState(GroupID(LISTGP_CorrFlags), 1, 1);
			drawFlags = (VisCorrs::DRAW_COLORED_PNTS | VisCorrs::DRAW_LINES);
		}
		m_pCorrsVis = new VisCorrs(m_pDBCorrs, drawFlags);
	}
	
	m_WindowListDims[0] = 0.016 * (int)m_GroupToType.size();
	m_WindowListDims[1] = .3;

	if (m_pDB!=NULL)
	{
		if ((listGPFlags & LISTGP_Geoms) && !nogeoms && m_pDB->NumModels()>0)
			ToggleGeometrySelection(0);
		int nMaxModels = m_pDB->NumModels()-1;
		
		for (int i=0; i<nTotal; i++)
			m_WindowToModel.push_back( (i > nMaxModels ) ? -1 : i);
//		m_WindowToModel[0] = 0;
		m_iSelectedWindow = -1;
		if (m_pDB->NumModels()>0)
		{
			SelectModelWindow(0);
			SelectModelInWindow(0);
		}
		for (int i=0; i<m_pDB->NumModels(); i++)
			m_ModelOrdering.push_back(i);
		
		while (m_EmptyWindowXForms.size() < m_WindowToModel.size())
			m_EmptyWindowXForms.push_back(new VisModel3D(NULL, 0));
	}
	
	m_iBackgroundColorID = 1;
}
								   
DatabaseVisWindow::~DatabaseVisWindow()
{
	for (int i=0; i<(int)m_DrawableModels.size(); i++)
		 delete m_DrawableModels[i];
}

VKString DatabaseVisWindow::GetClassName() const
{
	return "DatabaseVisWindow";
}

void DatabaseVisWindow::SetBackgroundColor(int bk)
{
	m_iBackgroundColorID = bk;
}

void DatabaseVisWindow::GetBackgroundColor(double * r, double * g, double * b) const
{
	if (m_iBackgroundColorID==0)
	{
		*r = 222. / 255.;
		*g = 222. / 255.;
		*b = 222. / 255.;
	}
	else if (m_iBackgroundColorID==1)
	{
		*r = 1.;
		*g = 1.;
		*b = 1.;
	}
	else if (m_iBackgroundColorID==2)
	{
		*r = 0.;
		*g = 0.;
		*b = 0.;
	}
	else
		std::cout<<"[WARNING] Unknown background colorID: "<<m_iBackgroundColorID<<std::endl;
}

void DatabaseVisWindow::SetApp(DBApp* app)
{
	m_pApp = app;
}

void DatabaseVisWindow::KeyPressed(const char key, int mouseX, int mouseY)
{
	if (m_pApp!=NULL && m_pApp->KeyPressed(this, key, mouseX, mouseY))
	{
		m_pRenderer->UpdateScreen();
		return;
	}
	
	switch (key)
	{
		case 27:
			exit(0);
		case 'a':
			if (m_pApp!=NULL)
				m_pApp->ToggleDisable();
			break;
		case '`':
			std::cout<<"Rendering Flags:"<<std::endl;
			std::cout<<"\tf\tShow/Hide model list"<<std::endl;
			std::cout<<"\tF\tHide/Show design boxes"<<std::endl;
			std::cout<<"\tb\tChange background color: gray, white, black"<<std::endl;
			std::cout<<"Other:"<<std::endl;
			std::cout<<"\t`\tPrint this message"<<std::endl;
			std::cout<<"\t~\tSelect point (by iD)"<<std::endl;
			std::cout<<"\tESC\texit the app"<<std::endl;
			break;
		case '\"':
		case '\'':
			ChangePage((key=='\"') ? -1 : 1);
			break;
		case 'f':
			ToggleDesignFlag(RENDER_MODEL_LIST);
			break;
		case 'F':
			ToggleDesignFlag(RENDER_DESIGN_BOXES);
			break;
		case '~':	// select point
		{
			int pntID;
			if (m_iSelectedWindow>=0 && m_iSelectedWindow < (int)m_DrawableModels.size())
			{
				ShapePointCloud * pnts = NULL;
				for (int i=0; i<m_DrawableModels[m_iSelectedWindow]->NumRenderableGeometries() && pnts==NULL; i++)
					if (m_DrawableModels[m_iSelectedWindow]->GetRenderableGeometry(i)!=NULL)
						pnts = m_DrawableModels[m_iSelectedWindow]->GetRenderableGeometry(i)->SafeCastToPointCloud();
				if (pnts!=NULL)
				{
					VKString pName = pnts->GeometryName().c_str();
					std::cout<<"Input Point ID in set "<<pName.c_str()<<" ("<<pnts->NumPoints()<<"): "<<std::flush;
					scanf("%d", &pntID);
					if (pntID>=0 && pntID < pnts->NumPoints())
						m_DrawableModels[m_iSelectedWindow]->SelectPoint(pnts->GetPoint(pntID));
					else
						std::cout<<"Invalid point ID"<<std::endl;
				}
				else
					std::cout<<"Could not find points geometry "<<std::endl;
			}
			break;
		}
		case 'b':
			m_iBackgroundColorID++;
			if (m_iBackgroundColorID>=3)
				m_iBackgroundColorID = 0;
			break;
		case 'w':
			WriteAllCameras();
			break;
		case 'r':
			ReadAllCameras();
			break;
		case ' ':
		{
			int wndID = ClickInWindow(mouseX, mouseY);
			if (wndID>=0 && wndID < NumWindows())
			{
				SelectModelWindow(wndID);
				
				// select point on a shape in subwindow
				int mID = m_WindowToModel[m_iSelectedWindow];
				if (mID >= 0)
				{
					assert(mID>=0 && mID < (int)m_DrawableModels.size());
					VisModel3D * mvis = m_DrawableModels[mID];
					int cameraID = mvis->SubwindowCameraID(m_iSelectedWindow, this);
					mvis->SelectPoint(cameraID, mouseX, mouseY, this);
					break;
				}
			}
		}
		default:
			std::cout<<"[WARNING] No action for key '"<<key<<"' (#"<<(int)key<<")"<<std::endl;
	}
	m_pRenderer->UpdateScreen();
}

void DatabaseVisWindow::MouseButtonEvent(int x, int y, MouseButton mouseButtonID, 
										 MouseButtonState newState)
{
	UpdateMouseButtons(mouseButtonID, newState);
	if (m_pApp!=NULL && m_pApp->MouseButtonEvent(this, x, y, mouseButtonID, newState))
	{
		m_pRenderer->UpdateScreen();
		return;
	}

	bool update = false;
	// TODO: app might capture the event
		
	if (mouseButtonID==MOUSE_LEFT && newState==MOUSE_BT_DOWN)
	{	// left-click: list selection, subwindow selection
		double xWnd, yWnd;
			// clicked on the list of buttons (top-left corner)
		if (InSubWindow(x, y, &xWnd, &yWnd,
						m_WindowListDims[0], m_WindowListDims[1],
						(VisWindow::WND_TOP | VisWindow::WND_LEFT)))
		{
			ProcessClickInModelListSubwindow(xWnd, yWnd);
			update = true;
		}
		else	// clicked on the subwindow
		{
			// select model / button list items
			int wndID = ClickInWindow(x, y);
			if (wndID>=0 && wndID < NumWindows())
			{
				SelectModelWindow(wndID);
				update = true;
			}
		}
	}
	else if (mouseButtonID==MOUSE_RIGHT && newState==MOUSE_BT_DOWN)
	{	// right-click: point selection
		int wndID = ClickInWindow(x, y);
		if (wndID>=0 && wndID < NumWindows())
		{
			SelectModelWindow(wndID);
			
			// select point on a shape in subwindow
			int mID = m_WindowToModel[m_iSelectedWindow];
			if (mID >= 0)
			{
				assert(mID>=0 && mID < (int)m_DrawableModels.size());
				VisModel3D * mvis = m_DrawableModels[mID];
				int cameraID = mvis->SubwindowCameraID(m_iSelectedWindow, this);
				mvis->SelectPoint(cameraID, x, y, this);
				
				update = true;
			}
		}
	}
	if (update)
		m_pRenderer->UpdateScreen();
}

void DatabaseVisWindow::ToggleDesignFlag(RenderDesignFlags flag)
{
	if (IsDesignFlagOn(flag))
		m_iDesignFlags = (m_iDesignFlags & ~(flag));
	else
		m_iDesignFlags = (m_iDesignFlags | flag);
}

bool DatabaseVisWindow::IsDesignFlagOn(RenderDesignFlags flag)
{
	return (m_iDesignFlags & flag)!=0;
}

void DatabaseVisWindow::ToggleGeometryValueSelection(int geomValID, int dir)
{
	int gpID = GroupID(LISTGP_GeomVals);
	assert(gpID>=0);
	
	m_pLists->ChangeState(gpID, geomValID, dir);
	
	VKString geomStr = m_pLists->ButtonName(gpID, geomValID);
	int bState = m_pLists->State(gpID, geomValID);
		
	for (int i=0; i<(int)m_DrawableModels.size(); i++)
	{
		m_DrawableModels[i]->RemoveRenderableGeomValue("all");
		if (bState>0)
			m_DrawableModels[i]->AddRenderableGeomValue(geomStr, bState-1);
	}
}

void DatabaseVisWindow::ToggleGeometrySelection(int geomID)
{
	int gpID = GroupID(LISTGP_Geoms);
	assert(gpID>=0);
	m_pLists->ChangeState(gpID, geomID, 1);
	VKString geomStr = m_pLists->ButtonName(gpID, geomID);
	int bState = m_pLists->State(gpID, geomID);
	
	for (int i=0; i<(int)m_DrawableModels.size(); i++)
	{
		if (bState==1)
			m_DrawableModels[i]->AddRenderableGeometry(geomStr);
		else
			m_DrawableModels[i]->RemoveRenderableGeometry(geomStr);			
	}
}

void DatabaseVisWindow::ProcessClickInModelListSubwindow(double wndX, double wndY)
{
	int gID, bID, dir;
	if (m_pLists->ClickInWindow(wndX, wndY, gID, bID, dir))
	{
		ListGroups gType = GroupType(gID);
		if (gType==LISTGP_Corrs)
		{
			m_pLists->ChangeState(gID, bID, dir);
			
			VKString buttonName = m_pLists->ButtonName(gID, bID);
			int bl = buttonName.length();
			VKString corrName = buttonName;
			if (buttonName.startsWith(m_sPairwisePrefix))
			{
				corrName = corrName.right(bl - m_sPairwisePrefix.length());
				int corrID = m_pLists->State(gID, bID) - 1;
				m_pCorrsVis->SelectCorr(DBCorrs::DBCORRS_PairwiseColl, corrName, corrID);
			}
			else if (buttonName.startsWith(m_sCorrPrefix))
			{
				corrName = corrName.right(bl - m_sCorrPrefix.length());
				int corrID = m_pLists->State(gID, bID)-1;
				m_pCorrsVis->SelectCorr(DBCorrs::DBCORRS_Corrs, corrName, corrID);
			}
		}
		else if (gType==LISTGP_CorrVals)
		{
			std::cout<<"[WARNING] Cannot selct corr vals: not implemented"<<std::endl;
		}
		else if (gType==LISTGP_CorrFlags)
		{
			m_pLists->ChangeState(gID, bID, dir);
			
			bool enable = m_pLists->State(gID, bID)==1;
			VKString buttonName = m_pLists->ButtonName(gID, bID);
			VisCorrs::DrawFlags flag = VisCorrs::DRAW_NONE;
			
			if (buttonName=="Draw Corrs: Colored Points")
				flag = VisCorrs::DRAW_COLORED_PNTS;
			else if (buttonName=="Draw Corrs: Connect Lines")
				flag = VisCorrs::DRAW_LINES;
			else if (buttonName=="Draw Corrs: Value Transfer")
				flag = VisCorrs::DRAW_VALUE_TRANSFER;
			else if (buttonName=="Draw Corrs: Affine Alignment")
				flag = VisCorrs::DRAW_AFFINE_ALIGN;
			else if (buttonName=="Draw Corrs: Fuzzy Values")
				flag = VisCorrs::DRAW_FUZZY_VALS;
			else
				std::cout<<"[WARNING] Unknown DrawFlag button: "<<buttonName.c_str()<<std::endl;

			if (flag!=VisCorrs::DRAW_NONE)
			{
				for (int i=0; i<(int)m_DrawableModels.size(); i++)
					m_pCorrsVis->SetFlag(flag, enable);
			}
		}
		else if (gType==LISTGP_Models)
		{
			if (m_iSelectedWindow>=0)
				SelectModelInWindow(bID);
		}
		else if (gType==LISTGP_Geoms)
			ToggleGeometrySelection(bID);
		else if (gType==LISTGP_GeomVals)
			ToggleGeometryValueSelection(bID, dir);
		else if (gType==LISTGP_GeomDists)
			std::cout<<"[WARNING] Cannot selct geom dists: not implemented"<<std::endl;			
		else if (gType==LISTGP_MeshFlags)
		{
			m_pLists->ChangeState(gID, bID, dir);
			bool enable = m_pLists->State(gID, bID)==1;
			VKString buttonName = m_pLists->ButtonName(gID, bID);
			VisModel3D::DrawFlags flag = VisModel3D::DRAW_NONE;
			if (buttonName=="Draw Mesh Faces")
				flag = VisModel3D::DRAW_FACES;
			else if (buttonName=="Draw Mesh Edges")
				flag = VisModel3D::DRAW_EDGES;
			else if (buttonName=="Draw Mesh Vertices")
				flag = VisModel3D::DRAW_VERTICES;
			else if (buttonName=="Draw IDs")
				flag = VisModel3D::DRAW_IDs;
			else if (buttonName=="Draw Pnts As Dots")
				flag = VisModel3D::DRAW_PDOTS;
			else if (buttonName=="Draw Pnts As Spheres")
				flag = VisModel3D::DRAW_PSPHERES;
			else if (buttonName=="Draw Normals")
				flag = VisModel3D::DRAW_NORMALS;
			else if (buttonName=="Draw Materials")
				flag = VisModel3D::DRAW_MATERIALS;
//			else if (buttonName=="Draw Thumbnails")
//				flag = VisModel3D::DRAW_THUMBNAILS;
			else if (buttonName=="Draw Meshes")
				flag = VisModel3D::DRAW_MESHES;
			else if (buttonName=="Draw Points")
				flag = VisModel3D::DRAW_POINTS;
			else if (buttonName=="Draw Flat")
				flag = VisModel3D::DRAW_FLAT;
			else
				std::cout<<"[WARNING] Unknown DrawFlag button: "<<buttonName.c_str()<<std::endl;

			if (flag!=VisModel3D::DRAW_NONE)
			{
				for (int i=0; i<(int)m_DrawableModels.size(); i++)
					m_DrawableModels[i]->SetFlag(flag, enable);
			}
		}
	}
}

int DatabaseVisWindow::GroupID(ListGroups gp)
{
	std::map<ListGroups, int>::iterator iter = m_ListToGroupID.find(gp);
	if (iter==m_ListToGroupID.end())
		return -1;
	return iter->second;
}

DatabaseVisWindow::ListGroups DatabaseVisWindow::GroupType(int groupID)
{
	assert(groupID>=0 && groupID < (int)m_GroupToType.size());
	return m_GroupToType[groupID];
}

void DatabaseVisWindow::AddGroup(ListGroups gp, const VKString & gpName,
								 VisButtonsGroup::ButtonsGroupType tp)
{
	m_ListToGroupID[gp] = (int)m_GroupToType.size();
	m_GroupToType.push_back(gp);
	m_pLists->AddButtonGroup(new VisButtonsGroup(gpName, tp));
}

void DatabaseVisWindow::SelectModelWindow(int wndID)
{
	if (m_iSelectedWindow==wndID)
		return;
	m_iSelectedWindow = wndID;
	if (m_iSelectedWindow >= 0 && m_WindowToModel[m_iSelectedWindow]>=0)
		SelectModelInWindow(m_WindowToModel[m_iSelectedWindow]);
}

void DatabaseVisWindow::SelectModelInWindow(int modelID)
{
	if (m_iSelectedWindow >= 0)
	{
//		if (m_WindowToModel[m_iSelectedWindow]==modelID)
//			return;

		int gID = GroupID(LISTGP_Models);
		if (gID>=0)
			m_pLists->ChangeState(gID, modelID, 1);
		m_WindowToModel[m_iSelectedWindow] = modelID;
		
		if (m_pDBCorrs!=NULL && m_pCorrsVis!=NULL
			&& (m_iSelectedWindow==m_iFromWindowID || m_iSelectedWindow==m_iToWindowID))
		{
			assert(m_iToWindowID>=0 && m_iToWindowID < (int)m_WindowToModel.size());
			assert(m_iFromWindowID>=0 && m_iFromWindowID < (int)m_WindowToModel.size());
			int mFromID = m_WindowToModel[m_iFromWindowID];
			int mToID = m_WindowToModel[m_iToWindowID];
			VKString m1Name = m_pDB->GetModel(mFromID)->ModelName();
			VKString m2Name = m_pDB->GetModel(mToID)->ModelName();

			VKString cachedM1Name = m_pCorrsVis->ModelFromName();
			VKString cachedM2Name = m_pCorrsVis->ModelToName();

			if (m1Name == cachedM1Name && m2Name == cachedM2Name)
				return;
			
			// clear corrs and corr vals (model changed)
			int gcID = GroupID(LISTGP_Corrs);
			int gvID = GroupID(LISTGP_CorrVals);
			VisButtonsGroup * gc = NULL;
			VisButtonsGroup * gv = NULL;
			if (gcID>=0)
			{
				m_pLists->ClearGroup(gcID);
				gc = m_pLists->GetGroup(gcID);
			}

			if (gvID>=0)
			{
				m_pLists->ClearGroup(gvID);
				gv = m_pLists->GetGroup(gvID);
			}

			VKStringList corrValList;
			corrValList.push_back("default");
			
			int camera1 = m_DrawableModels[mFromID]->SubwindowCameraID(m_iFromWindowID, this);
			int camera2 = m_DrawableModels[mToID]->SubwindowCameraID(m_iToWindowID, this);
			m_pCorrsVis->SelectPair(m1Name, m_DrawableModels[mFromID], camera1,
									m2Name, m_DrawableModels[mToID], camera2);
			
			if (gc!=NULL)
			{
				// find all corrs: pairwise
				for (int i=0; i<m_pDBCorrs->NumPairColls(); i++)
				{
					CollectionPairwise * pairColls = m_pDBCorrs->GetPairColl(i);
					int nCorrs = pairColls->NumCorrsForPair(m1Name, m2Name);
					gc->AddButton(m_sPairwisePrefix + pairColls->CollectioName(), nCorrs+1);
				}
				
				// find all corrs: 
				for (int i=0; i<m_pDBCorrs->NumCorrs(); i++)
				{
					ShapeCorrs * corrs = m_pDBCorrs->GetCorrs(i);
					gc->AddButton(m_sCorrPrefix + corrs->GetCorrsName(), 2);
				}
			}
			
			// add corr values
			if (gv!=NULL)
				gv->AddOnOffButtons(corrValList);
		}
	}
}

void DatabaseVisWindow::MouseMotionEvent(int x, int y, int dx, int dy)
{
	if (m_pApp!=NULL && m_pApp->MouseMotionEvent(this, x, y, dx, dy))
	{
		m_pRenderer->UpdateScreen();
		return;
	}
	
	bool update = false;
//	if (m_DrawableModels.size()==0)
//		return;
	
	if (m_MouseStates[0] == MOUSE_BT_DOWN)
	{
		if (m_iSelectedWindow>=0 && m_iSelectedWindow < NumWindows())
		{
			int mID = m_WindowToModel[m_iSelectedWindow];
			VisModel3D * mvis = NULL;
			if (mID<0)
				mvis = m_EmptyWindowXForms[m_iSelectedWindow];
			else
			{
				assert(mID < (int)m_DrawableModels.size());
				mvis = m_DrawableModels[mID];
			}

			int cameraID = mvis->SubwindowCameraID(m_iSelectedWindow, this);
			if (!Ctrl() && !Alt() && !Shift())	// rotate
				mvis->Rotate(cameraID, x, y, dx, dy, this);
			else if (Ctrl())
				mvis->Scale(cameraID, x, y, dx, dy, this);
			else if (Shift())
				mvis->Translate(cameraID, x, y, dx, dy, this);		  
			update = true;
		}
	}
	if (update)
		m_pRenderer->UpdateScreen();
}

void DatabaseVisWindow::Resize(int w, int h)
{
	// not implemented
}

void DatabaseVisWindow::Draw()
{
	if (m_pApp!=NULL && m_pApp->Draw(this))
		return;

	double br, bg, bb;
	GetBackgroundColor(&br, &bg, &bb);
	glClearColor(br, bg, bb, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT 
			| GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	// draw all models:
	for (int i=0; i<(int)m_WindowToModel.size(); i++)
	{
		int modelID = m_WindowToModel[i];
		if (modelID<0 || modelID >= (int)m_DrawableModels.size())
			continue;
		int cameraID = m_DrawableModels[modelID]->SubwindowCameraID(i, this);
		m_DrawableModels[modelID]->Draw(this, cameraID);
	}
	
	// draw corrs
	if (m_pCorrsVis!=NULL)
		m_pCorrsVis->Draw(this);
	
	if (IsDesignFlagOn(RENDER_DESIGN_BOXES))
	{
		for (int i=0; i<(int)m_WindowToModel.size(); i++)
			DrawSubwindow(i);
	}
	
	if (IsDesignFlagOn(RENDER_MODEL_LIST))
		DrawModelListSubwindow();
}

void DatabaseVisWindow::NumMainWindows(int & x, int & y) const
{
	x = m_iMainWndGridX;
	y = m_iMainWndGridY;
}

void DatabaseVisWindow::NumAdditionalWindows(int & x, int & y) const
{
	x = m_iAdditionalWndGridX;
	y = m_iAdditionalWndGridY;
}

int DatabaseVisWindow::NumWindows()
{
	return (int)m_WindowToModel.size();
}

int DatabaseVisWindow::NumMainWindows() const
{
	return m_iMainWndGridX * m_iMainWndGridY;
}

int DatabaseVisWindow::NumAdditionalWindows() const
{
	return m_iAdditionalWndGridX * m_iAdditionalWndGridY;
}

int DatabaseVisWindow::ClickInWindow(int x, int y)
{
	double fx = (double) x / (double)Width();
	double fy = (double) y / (double)Height();
	
	fx = fx * 2. - 1.;
	fy = fy * 2. - 1.;
	for (int i=0; i < NumWindows(); i++)
	{
		double cx, cy, scale, dx, dy;
		GetWindowParams(i, &cx, &cy, &scale, &dx, &dy);
		double minX = (cx - dx/2.) / Width();
		double maxX = (cx + dx/2.) / Width();
		double minY = (cy - dy/2.) / Height();
		double maxY = (cy + dy/2.) / Height();
		if (fx >= minX && fx <= maxX && fy>=minY && fy<=maxY)
			return i;
	}
	return -1;
}

void DatabaseVisWindow::InitializeGL()
{
	// Initialize background color 
	//glClearColor(150./255.0, 150./255.0, 150./255.0, 1.0);
	
	// Initialize lights
	static GLfloat lmodel_ambient[] = { 0.2, 0.2, 0.2, 1.0 };
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
	
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	static GLfloat light0_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
	glEnable(GL_LIGHT0);
	static GLfloat light1_diffuse[] = { 0.5, 0.5, 0.5, 1.0 };
	glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
	glEnable(GL_LIGHT1);
	glEnable(GL_NORMALIZE);
	glEnable(GL_LIGHTING); 
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	
	// Initialize graphics modes  
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);	
	
	// initialize alpha-blending
	glEnable (GL_BLEND); 
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
}

void DatabaseVisWindow::DrawModelListSubwindow()
{
	glLineWidth(1.);
	StartRenderSubwindow(true, true, m_WindowListDims[0], m_WindowListDims[1]);
	m_pLists->Draw(this);
}

void DatabaseVisWindow::DrawSubwindow(int wndID)
{
	glLineWidth(1.);
	double r=1., g=0., b=0.;
	if (wndID == m_iSelectedWindow)
	{
		r = 0.;	g = 0.;	b = 1.;
	}
	double textW=.03;
	double textH=.04;
	
	double cx, cy, scale, dx, dy;
	GetWindowParams(wndID, &cx, &cy, &scale, &dx, &dy);
	
	glColor3d(r, g, b);	
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glDisable(GL_LIGHTING);
	glLineWidth(1.);

	if (IsDesignFlagOn(RENDER_BOXID))
	{
		double textX = (cx-dx/2.) / Width() + textW;
		double textY = (cy) / Height() - textH;
		m_pRenderer->RenderText(R2Point(textX, textY), VKString::number(wndID), 
								r, g, b, VisWindowRenderer::TEXT_SIZE_MEDIUM);
		
		glBegin(GL_LINE_STRIP);
		glVertex2d(textX - textW*.15, textY - textH*.15);
		glVertex2d(textX - textW*.15, textY + textH);
		glVertex2d(textX + textW, textY + textH);
		glVertex2d(textX + textW, textY  - textH*.15);
		glVertex2d(textX - textW*.15, textY - textH*.15);
		glEnd();
	}
	
	double minX = (cx - dx/2.) / Width();
	double maxX = (cx + dx/2.) / Width();
	double minY = (cy - dy/2.) / Height();
	double maxY = (cy + dy/2.) / Height();
	
	double off = 0.01;
	glBegin(GL_LINE_STRIP);
	glVertex2d(minX + off, minY + off);
	glVertex2d(minX + off, maxY - off);
	glVertex2d(maxX - off, maxY - off);
	glVertex2d(maxX - off, minY + off);
	glVertex2d(minX + off, minY + off);
	glEnd();
	glEnable(GL_DEPTH_TEST);
}

void DatabaseVisWindow::GetWindowParams(int wndID, double * cx, double * cy, double * scale,
										double * dx, double * dy)
{
	double minX = (double)-Width();
	double maxX = (double)Width();
	double minY = (double)-Height();
	double maxY = (double)Height();
	
	double mainWndSizeX = (maxX - minX) * m_fXSeparator;
	double mainWndSizeY = (maxY - minY);
	
	int nMainWndX = m_iMainWndGridX;
	int nMainWndY = m_iMainWndGridY;
	if (wndID < nMainWndX*nMainWndY)		// main window
	{
		double wndInMainDX = mainWndSizeX / (double)nMainWndX;
		double wndInMainDY = mainWndSizeY / (double)nMainWndY;
		int r = nMainWndY - wndID / nMainWndX - 1;
		int c = wndID % nMainWndX;
		
		*cy = minY + wndInMainDY/2. + r*wndInMainDY;
		*cx = minX + wndInMainDX/2. + c*wndInMainDX;
		
		*scale = m_fXSeparator * 1. / vkMax(nMainWndX, nMainWndY);
		
		if (dx!=NULL)
			*dx = wndInMainDX;
		if (dy!=NULL)
			*dy = wndInMainDY;
	}
	else 
	{
		wndID -= nMainWndX*nMainWndY;
		assert(wndID>=0);
		double addWndSizeX = (maxX - minX) * (1.-m_fXSeparator);
		double addWndSizeY = (maxY - minY);
		
		double wndInAddDX = addWndSizeX / m_iAdditionalWndGridX;
		double wndInAddDY = addWndSizeY / m_iAdditionalWndGridY;
		
		int r = m_iAdditionalWndGridY - wndID / m_iAdditionalWndGridX - 1;
		int c = wndID % m_iAdditionalWndGridX;

		*cy = minY + wndInAddDY/2. + r*wndInAddDY  ;
		*cx = minX + mainWndSizeX + wndInAddDX/2. + c*wndInAddDX;
		
		*scale = (1.-m_fXSeparator) * 1. / vkMax(m_iAdditionalWndGridX, m_iAdditionalWndGridY);
		*scale *= 3;
		
		if (dx!=NULL)
			*dx = wndInAddDX;
		if (dy!=NULL)
			*dy = wndInAddDY;
	}
}

std::vector<VisModel3D *> & DatabaseVisWindow::GetDrawableModels()
{
	return m_DrawableModels;
}

int DatabaseVisWindow::SelectedWindow() const
{
	return m_iSelectedWindow;
}

R3Affine DatabaseVisWindow::GetEmptySubwindowCamera(int subwindow)
{
	assert(subwindow<(int)m_EmptyWindowXForms.size());
	int camID = m_EmptyWindowXForms[subwindow]->SubwindowCameraID(subwindow, this);
	return m_EmptyWindowXForms[subwindow]->GetCameraTransform(camID);
}

VisModel3D * DatabaseVisWindow::GetEmptyWindowXForm(int subwindow)
{
	assert(subwindow<(int)m_EmptyWindowXForms.size());
	return m_EmptyWindowXForms[subwindow];
}

R3Affine DatabaseVisWindow::GetSubwindowCamera(int subwindow)
{
	int mID = ModelID(subwindow);
	if (mID<0)
	{
		assert(subwindow<(int)m_EmptyWindowXForms.size());
		int camID = m_EmptyWindowXForms[subwindow]->SubwindowCameraID(subwindow, this);
		return m_EmptyWindowXForms[subwindow]->GetCameraTransform(camID);
	}
	else
	{
		assert (mID<(int)m_DrawableModels.size());
		int camID = m_DrawableModels[mID]->SubwindowCameraID(subwindow, this);
		return m_DrawableModels[mID]->GetCameraTransform(camID);
	}
	assert(false);
}

int DatabaseVisWindow::ModelID(int subwindow )
{
	assert(subwindow>=0 && subwindow<(int)m_WindowToModel.size());
	return m_WindowToModel[subwindow];
}

std::vector<int> & DatabaseVisWindow::WindowToModel()
{
	return m_WindowToModel;
}

void DatabaseVisWindow::ChangePage(int change)
{
	if (m_ModelOrdering.size()==0)
		return;

	m_iOrderedModelPage += change;
	UpdatePageAndOrdering();
	std::cout<<"Page: "<<m_iOrderedModelPage<<". NModels: "<<m_ModelOrdering.size()<<std::endl;
}

int & DatabaseVisWindow::GetOrderedPage()
{
	return m_iOrderedModelPage;
}

std::vector<int> & DatabaseVisWindow::GetModelOrdering()
{
	return m_ModelOrdering;
}

void DatabaseVisWindow::UpdatePageAndOrdering()
{
	int nAdditional = m_iAdditionalWndGridX * m_iAdditionalWndGridY;
	int nMain = m_iMainWndGridX * m_iMainWndGridY;
	
	int nWindows = -1;
	int nOffset = -1;
	// page is in main
	if (nAdditional == 0)
	{		
		nWindows = nMain;
		nOffset = 0;
	}
	else
	{
		nWindows = nAdditional;
		nOffset = nMain;
	}
	assert(nWindows>0);
	assert(nOffset>=0);
	if (m_iOrderedModelPage < 0)
	{
		m_iOrderedModelPage = (int)m_ModelOrdering.size() / nWindows;
		if (((int)m_ModelOrdering.size() % nWindows) == 0)
			m_iOrderedModelPage--;
	}
	
	if (nWindows * m_iOrderedModelPage >= (int)m_ModelOrdering.size())
		m_iOrderedModelPage = 0;
	
	for (int wID=0; wID<nWindows; wID++)
	{
		int subwindowID = nOffset + wID;
		int orderID = nWindows * m_iOrderedModelPage + wID;
		assert(subwindowID < (int)m_WindowToModel.size());
		
		if (orderID >= (int)m_ModelOrdering.size())
			m_WindowToModel[subwindowID] = -1;
		else
			m_WindowToModel[subwindowID] = m_ModelOrdering[orderID];
	}
}

void DatabaseVisWindow::WriteAllCameras()
{
	FILE * file = fopen("./cameras.txt", "wb");
	assert(file!=NULL);
	for (int i=0; i<(int)m_DrawableModels.size(); i++)
		m_DrawableModels[i]->WriteCamera(file);
	for (int i=0; i<(int)m_EmptyWindowXForms.size(); i++)
		m_EmptyWindowXForms[i]->WriteCamera(file);
	fclose(file);
}

void DatabaseVisWindow::ReadAllCameras()
{
	FILE * file = fopen("./cameras.txt", "rb");
	assert(file!=NULL);
	for (int i=0; i<(int)m_DrawableModels.size(); i++)
		m_DrawableModels[i]->ReadCamera(file);
	for (int i=0; i<(int)m_EmptyWindowXForms.size(); i++)
		m_EmptyWindowXForms[i]->ReadCamera(file);
	fclose(file);	
}
