#!/usr/bin/python

##
# Learns a set of deformable templates from a collection (uses automatic segmentation to
# test several candidate initial templates)
#
import os, sys;
sys.path.append(os.path.realpath("./"+os.path.dirname(sys.argv[0])+"/scripts/largescale/"))
sys.path.append(os.path.realpath("./"+os.path.dirname(sys.argv[0])+"/scripts/experiment/"))
import mydb, expdb;

workDir = "./scratch"
if len(sys.argv)>=4:
	nsegStr = sys.argv[1]
	dbdir = sys.argv[2]
	outdir = sys.argv[3]
	i = 4;
	while i<len(sys.argv):
		if (sys.argv[i]=="-workDir"):
			i = i + 1;
			workDir = sys.argv[i];
		i = i + 1;

else:
	print("Usage:")
	print("   autoLearnTmplt nseg collection_dir outdir [optional]");
	print("Optional:");
	print("    -workDir dir - working directory (default=./scratch)");
	exit()

# prepare collection
if os.path.isdir(workDir):
	os.system("rm -r -f "+workDir);
os.system("mkdir "+workDir);

os.system("./scripts/largescale/createFromDir.py "+dbdir+" "+workDir+"/");
os.system("./scripts/largescale/createAllPointSets.py "+workDir+" 1000");

mList = mydb.ListModels(workDir);

# create initial templates
for mName in mList:
	cmd = "./CreateTemplate -readDB "+mydb.MyDBDir(workDir)+" -model "+mName+" ";
	cmd += "-initFromGeometry "+mName+" "+nsegStr+" -writeTemplate "+workDir+"/"+mName+".tmplt ";
	os.system(cmd);

# analyze collection with every template
for mName in mList:
	tmplt = workDir+"/"+mName+".tmplt";
	cmd = "./scripts/evolution/evolve.py "+workDir+" "+tmplt+" "+mName+" ";
	os.system(cmd);

# pick best-scored template
bestMName = "none";
bestEnergy = -1;
for mName in mList:
	labelsList = workDir+"/Templates/"+mName+"_GenFinal/LabeledModels.txt";
	if os.path.isfile(labelsList):
		mListScored = expdb.readModelList(labelsList);
		energy = 0;
		for mInst in mListScored:
			energy = energy + mInst.energy;
		energy = energy / len(mListScored);
		if energy < bestEnergy or bestMName=="none":
			bestEnergy = energy;
			bestMName = mName;
		print("Template: "+mName+" - "+str(energy));

print("Best template: "+bestMName);

# export results
cmd = "./exportAnalysis.py "+workDir+" ";
cmd += workDir+"/Templates/"+bestMName+"_GenFinal "+outdir;
os.system(cmd);

#os.system("cp "+workDir+"/Templates/"+mName+"_GenFinal/main.tmplt "+outFile);







