#!/usr/bin/python

##########
# Current evolution algorithm
##########
import sys, os;
sys.path.append(os.path.realpath("./"+os.path.dirname(sys.argv[0])+"/../largescale/"))
import string, decimal, glob, re, subprocess, shlex, datetime, mydb, expdb

# Manual, using sid's data, with a consistent segmentation
outTmplt = "none";
readTmplt = "none";
refDB = "none";
refMdl = "none";
createMode = "manual";
vis = False;
nsegs = 0;
if len(sys.argv) < 2:
	print("Usage: "+sys.argv[0]+" outTmplt [options]");
	print("Options:");
	print("   -readTmplt [none, filename] - read template (or set to none) ");
	print("   -refDB keyword  - read database as a reference ");
	print("   -refMdl model - read model as a reference ");
	print("   -nsegs N - number of segments (if createMode==ncuts) ");
	print("   -createMode [manual true sidi ncuts wordnet auto] - template creation mode... ");
	print("   -vis - visualize ");
	exit();
else:
	outTmplt = sys.argv[1];
	if os.path.isfile(outTmplt):
		readTmplt = outTmplt;
	i = 2;
	while i<len(sys.argv):
		if (sys.argv[i]=="-readTmplt"):
			i = i + 1;
			readTmplt = sys.argv[i];
		elif (sys.argv[i]=="-refDB"):
			i = i + 1;
			refDB = sys.argv[i];
		elif (sys.argv[i]=="-refMdl"):
			i = i + 1;
			refMdl = sys.argv[i];
		elif (sys.argv[i]=="-vis"):
			vis = True;
		elif (sys.argv[i]=="-createMode"):
			i = i + 1;
			createMode = sys.argv[i];
		elif (sys.argv[i]=="-nsegs"):
			i = i + 1;
			nsegs = int(sys.argv[i]);
		else:
			print("[WARNING] Unknown argument: "+sys.argv[i]);
		i = i + 1;

pntsName = "main_Even"+str(expdb.NPoints())+".pnts";

if createMode=="manual":  	# manually create a template
	# find or create pnts for a reference model
	pntsFile = mydb.MyDBDir(refDB)+"/Models/"+refMdl+"/Geometry/"+pntsName;
	if (refMdl!="none" and not os.path.isfile(pntsFile)):
		cmd="./FindPoints "+mydb.MyDBDir(refDB)+" -model "+refMdl+" -type IterFar eucl ";
		cmd+="-npoints "+str(expdb.NPoints())+" -pntsName Even"+str(expdb.NPoints())+" -writeGeometry; ";
		os.system(cmd);

	# visualize collection (if given)
	#if (refDB!="none"):
	#	cmd = "./CollectionViewer "+mydb.MyDBDir(refDB)+" -vis &"
	#	os.system(cmd);

	cmd = "./CreateTemplate -initFromScratch -writeTemplate "+outTmplt+" -vis "
	if (refDB!="none"):
		cmd += "-readDB "+mydb.MyDBDir(refDB)+" ";
	if readTmplt!="none":
		cmd += "-readTemplate "+readTmplt+" ";
	if (refMdl!="none"):
		cmd += "-model "+refMdl+" -refGeom "+pntsName+" ";
	os.system(cmd);
elif createMode=="true": 	# create a template from a ground truth segmentation
	cmd="./CreateTemplate -readDB "+mydb.MyDBDir(refDB)+" -model "+refMdl+" -verb ";
	cmd += "-train "+refMdl+" main_Even"+str(expdb.NPoints())+".pnts SceneGraphNodes ";
	cmd+="-writeTemplate "+outTmplt+" -splitLabelsIntoCC -useMeshThresh 0 ";
	if (vis):
		cmd += "-vis ";
	cmd += ";"
	os.system(cmd);
elif createMode=="sidi": 	# create a template from a segmentation by Sidi et al.
	dataset, dbName = expdb.ExtractDBAndKeyword(refDB);
	segFile = mydb.cosegDataDir+"/SidiResults/"+dbName+"/"+refMdl+".seg";
	if not os.path.isfile(segFile):
		print("[ERROR] Do not have pre-analyzed results for "+refDB+". in createTemplate.py");
		exit();
	cmd="./CreateTemplate -readDB "+mydb.MyDBDir(refDB)+" -model "+refMdl+" -verb ";
	cmd += "-train "+refMdl+" main_Even"+str(expdb.NPoints())+".pnts SidiResults "+" ";
	cmd+="-writeTemplate "+outTmplt+" -splitLabelsIntoCC -useMeshThresh 0 -trainLabelsFile "+segFile+" ";
	if (vis):
		cmd += "-vis ";
	cmd += ";"
	os.system(cmd);
elif createMode=="ncuts" or createMode=="ncuts_fuzzy" or createMode=="kmedian" or createMode=="kmedian_fuzzy":
	# create a template from a segmentation by Sidi et al.
	dataset, dbName = expdb.ExtractDBAndKeyword(refDB);
	ncutsExe = "/u/funk/gaps/bin/x86_64/ncuts";
#	if nsegs > 0 and os.path.isfile(ncutsExe): # re-execute ncuts
#		segDir = mydb.MyDBDir(refDB)+"../Segmentations";
#		if not os.path.isdir(segDir):
#			os.system("mkdir "+segDir);
#		segDir += "/ncuts/";
#		if not os.path.isdir(segDir):
#			os.system("mkdir "+segDir);
#		
#		inOffFile = mydb.MyDBDir(refDB)+"/Models/"+refMdl+"/Geometry/main.ascii.off";
#		segFile = segDir+refMdl+".ply";
#		os.system(ncutsExe+" "+inOffFile+" "+segFile+" -nsegments "+str(nsegs));
#	else:	# data might have been processed
#		mydir = "/Users/vk/Research/DBShapesSVN/bin/Segmentations/";
#		if os.path.isdir(mydir):
#			segDir = mydir;
#		else:
	segDir = "/n/fs/graphicslab/funk/models/CoSegB/";
	if not os.path.isdir(segDir):
		segDir = "../Segmentations/"
	if createMode[0:5]=="ncuts":
		segDir += "ncuts/";
	elif createMode[0:7]=="kmedian":
		segDir += "kmedian/";
	else:
		print("[ERROR] Missing create mode: "+createMode);
		exit();
	segFile = refDB;
	if segFile[0:7] == "CoSegB_":
		segFile=segFile[7:];
	if segFile[0:5]=="Chair":
		segFile = "Chair";
	elif segFile[0:4]=="Vase":
		segFile = "Vase";
	elif segFile=="FourLegged":
		segFile = "FourLeg";
	if segFile[-1]=="s":
		segFile=segFile[0:-1];
	if createMode[-5:]=="fuzzy":
		segFile = segDir+segFile+"_"+refMdl+"_fuzzy.ply";
	else:
		segFile = segDir+segFile+"_"+refMdl+".ply"
	if not os.path.isfile(segFile):
		print("[ERROR] Cannot find segmentation "+segFile+" (see createTemplate.py)");
		exit();

	cmd="./CreateTemplate -readDB "+mydb.MyDBDir(refDB)+" -model "+refMdl+" -verb ";
	cmd += "-train "+refMdl+" main_Even"+str(expdb.NPoints())+".pnts ncuts_"+str(nsegs)+" ";
	cmd+="-writeTemplate "+outTmplt+" -useMeshThresh 0 -trainLabelsFile "+segFile+" ";
	if (vis):
		cmd += "-vis "; 
	cmd += ";"
	os.system(cmd);
elif createMode=="wordnet":		 # use wordnet to create a template
	cmd = "./CreateTemplate -verb -initFromPartDir "+readTmplt+" -writeTemplate "+tmpltFile+" ";
	if (vis):
		cmd += "-vis ";
	os.system(cmd);
elif createMode=="auto":		# consistent segmentation for a set of models
	print("[ERROR] Creation mode: "+createMode+" not YET implemented");
else:
	print("[ERROR] Creation mode: "+createMode+" is not impelemented!");




