##########
# Some helper functions for experiments
##########
import sys, os
sys.path.append(os.path.realpath("./"+os.path.dirname(sys.argv[0])+"/../largescale/"))
import re, subprocess, shlex, sys, datetime, glob, string, time, mydb

mCurrMethod = "aligned_sep15";

classesPSB=["PSB_ant", "PSB_bench", "PSB_biplane", "PSB_butterfly", "PSB_commercial", "PSB_desk_chair",  \
			"PSB_dining_chair", "PSB_dog", "PSB_eyeglasses", "PSB_horse", "PSB_human", "PSB_motorcycle", \
			"PSB_race_car"];

classesSD=["SD_Chair", "SD_Animals", "SD_Commercial", "SD_CommercialShips", "SD_Dinosaurs", "SD_Fantasy", \
			"SD_Helicopters", "SD_Military", "SD_MilitaryShips", "SD_Reptiles", "SD_Table"];


# define templates to read (otherwise uses training models)
templateFiles=dict();
templateFiles[""]="";
templateFiles["SD_Chair"] = "PSB_dining_chair.tmplt";
templateFiles["SD_Fantasy"] = "PSB_human.tmplt";
templateFiles["chair1000"] = "PSB_dining_chair.tmplt"
templateFiles["plane1000"] = "PSB_commercial.tmplt"
templateFiles["bike1000"] = "PSB_motorcycle.tmplt"
templateFiles["car1000"] = "PSB_race_car.tmplt"

#### This part is on qsub and running
def NPoints():
	return 1000;

def FindEvenPnts(keyword, nPoints):
	mList = mydb.ListModels(keyword);
	# compute point clouds
	for mStr in mList:	
		gDir=mydb.GeometryDir(keyword, mStr);
		if not os.path.isfile(gDir+"/main_Even"+str(nPoints)+".pnts"):
			cmd="./FindPoints "+mydb.MyDBDir(keyword)+" -model "+mStr+" -type IterFar eucl ";
			cmd+="-npoints "+str(nPoints)+" -pntsName Even"+str(nPoints)+" -writeGeometry; ";
			mydb.ExecuteNoWait(cmd, "pts_"+mStr);

def PrepareWebTemplateDir(keyword, outDir, waitForMe):
	# prepare output
	tmpltWeb="none";
	if mydb.WebspaceExists() and outDir!="none":
		print("Preparing: "+keyword+" dir="+outDir);
		# copy template
		tmpltWeb1 = mydb.Webpage(keyword)+"/Templates/"
		tmpltWeb = tmpltWeb1 + outDir+"/";
		mydb.InitWeb(keyword);
		cmd="";
		if not os.path.isdir(tmpltWeb1):
			cmd += "mkdir "+tmpltWeb1+"; ";
		if not os.path.isdir(tmpltWeb):
			cmd += "mkdir "+tmpltWeb+";";
		if (cmd!=""):
			mydb.ExecuteNoWait(cmd, "webinit")
			if waitForMe:
				print("Waiting for template dirs");
				mydb.WaitForAllJobs();
	return tmpltWeb;

def PrepareDataset(dataset, keyword, nModels):
	# prepare: create database, render database
	if (dataset=="GW"):
		fullKeyword = keyword+str(nModels);
	else:
		fullKeyword = dataset+"_"+keyword;

	if not os.path.isdir(mydb.MyDBDir(fullKeyword)):
		if (dataset=="GW"):
			cmd="./scripts/largescale/createDB.py ./Processing/"+fullKeyword+" "+str(nModels)+" "+keyword;
		else:
			cmd="./scripts/largescale/createFrom"+dataset+".py "+fullKeyword+" "+str(nModels)+" *"+keyword;
		mydb.ExecuteNoWait(cmd, "createDB");

def BasicTemplateFile(keyword, useMeshThreshold):
	# get models to be processed
	if keyword in templateFiles:
		tmpltName = templateFiles[keyword];
	else:
		tmpltName = keyword+".tmplt";
	if (useMeshThreshold==-1):
		threshDir = "";
	else:
		threshDir = str(useMeshThreshold);
	dir = "./Processing/SharedTemplates/BasicTemplates/"+threshDir+"/";
	
	if (keyword==""):
		return dir;
	else:
		return dir+tmpltName;

def SuperTemplateFile(keyword):
	pfx = "./Processing/SharedTemplates/SuperTemplates/";
	if (keyword==""):
		return pfx;
	else:
		return pfx+keyword+".tmplt";

def CurrMethod():
	return mCurrMethod;

def AllBenchmarkSets():
	#return (classesPSB+classesSD);
	return ["PSB_dining_chair"];
	#return ["PSB_motorcycle"];
	#return ["SD_MilitaryShips"];

def ExtractDBAndKeyword(benchmarkSet):
	if benchmarkSet[:4]=="PSB_":
		dataset = "PSB";
		keyword = benchmarkSet[4:];
	elif benchmarkSet[:3]=="SD_":
		dataset = "SD";
		keyword = benchmarkSet[3:];
	elif benchmarkSet[:len("CoSegB_")]=="CoSegB_":
		dataset = "CoSegB";
		keyword = benchmarkSet[len("CoSegB_"):];
	elif benchmarkSet[:len("PrSegB_")]=="PrSegB_":
		dataset = "PrSegB";
		keyword = benchmarkSet[len("PrSegB_"):];
	elif benchmarkSet[:len("FC_")]=="FC_":
		dataset = "FC";
		keyword = benchmarkSet[len("FC_"):];
	else:
		dataset = "GW";
		keyword = benchmarkSet;
	return (dataset, keyword);

### MODEL LISTS
class ModelInst:
	def __init__(self, mname):
		self.mname = mname;
		self.energy = -1;
		self.timeStr = "Xs";
		self.seedid = -1;
		self.partFlags = "unknown";
		self.ntmplt = 0;
		self.deformDist = -1;

def readModelList(modelList):
	if not os.path.isfile(modelList):
		return [];
	f = open(modelList, "r");
	modelLines = f.readlines();
	f.close();
	mList = [];
	for s in modelLines:
		sSplit = s.strip().split(" ");
		mName = sSplit[0].strip();
		mEnergy = -1;
		mList.append(ModelInst(mName));
		if len(sSplit)>1:
			mList[-1].energy = float(sSplit[1].strip());
		if len(sSplit)>2:
			mList[-1].timeStr = sSplit[2].strip();
		if len(sSplit)>3:
			mList[-1].seedid = int(sSplit[3].strip());
		if len(sSplit)>4:
			mList[-1].partFlags = sSplit[4].strip();
		if len(sSplit)>5:
			mList[-1].ntmplt = int(sSplit[5].strip());
		if len(sSplit)>6:
			mList[-1].deformDist = float(sSplit[6].strip());
	return mList;

def writeModelList(mList, mListFile, writeAllData):
	if len(mList)==0:
		return;	# do not write empty
	f = open(mListFile, "w");
	for s in mList:
		f.write(s.mname+" ");
		if writeAllData:
			f.write(str(s.energy)+" ");
			f.write(s.timeStr.replace(" ", "_")+" ");
			f.write(str(s.seedid)+" ");
			f.write(str(s.partFlags)+" ");
			f.write(str(s.ntmplt)+" ");
			f.write(str(s.deformDist)+" ");
		f.write("\n");
	f.close();



