##########
# Some helper functions for exploring large-scale collections
##########

import os, re, subprocess, shlex, sys, datetime, glob, string, time

webpage="./"

ProcDir="./"

maxJobs=1000

#### This part is on qsub and running
def WaitForAllJobs():
	if QSub():
		while NumMyJobs()>0:
			print("        Waiting for jobs: "+str(NumMyJobs()))
			time.sleep(5)

def WaitForJobsInFile(filename):
	if (QSub()):
		if (filename=="none"):
			WaitForAllJobs();
		else:
			WaitForJobsInArray(ReadJobsFile(filename));

def WaitForJobsInArray(jobsIds):
	if QSub():		
		# list of jobs and their status
		jobID = [];
		jobCMD = [];
		jobStatus = [];
		jobInitRunTime = [];
		lastVisit = [];
		jobRestarts = [];
		# jobID map
		jMap = dict();
		
		# initialize job data
		for job in jobsIds:
			jobSplitStr = job.split(" ");
			jobID.append(jobSplitStr[0]);
			jobCMD.append(" ".join(jobSplitStr[1:]));
			jobStatus.append("Q");
			jobInitRunTime.append(-1);
			jMap[jobSplitStr[0]] = len(jobID)-1;
			lastVisit.append(0);
			jobRestarts.append(0);
		
		# wait for jobs to be done...
		counterSleeps = 0;
		sleepTime = 5;	# in seconds
		visitCounter = 0;
		while len(jobID):
			nRunning = 0;
			nTotal = 0;
			# check if idle jobs can be udpated
			# go over running jobs - set check id
			subPID = subprocess.Popen("qstat | grep $USER ", \
									  shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
			jobList, errors = subPID.communicate();
			
			jList = jobList.split("\n");
			visitCounter = visitCounter + 1;
			for j in jList:
				if (j==""):
					continue;
				nTotal = nTotal + 1;
				jobDescrWithEmpty = j.strip().split(" ");
				jobDescr = [];
				for val in jobDescrWithEmpty:
					if val != "":
						jobDescr.append(val);
				jid = jobDescr[0].split(".")[0].strip();
				jName = jobDescr[1].strip();
				jTimeUse = jobDescr[3].strip();
				jstat = jobDescr[4].strip();
				if jid in jMap:
					lid = jMap[jid];
					lastVisit[lid] = visitCounter;
					if jstat != "R" and jstat != "Q" and jstat != "E":
						print("[WARNING] Invalid jstat: "+jstat);
					if jstat != jobStatus[lid]:
						if (jobStatus[lid]+jstat)=="QR":
							jobStatus[lid] = jstat;
							jobInitRunTime[lid] = 0;
						elif (not jobStatus[lid]+jstat)=="RE":
							print("[WARNING] Unexpected state change: "+jobStatus[lid]+" -> "+jstat);
					if (jobStatus[lid]=="R"):
						nRunning = nRunning + 1;
					if jobStatus[lid]=="R" and (jTimeUse=="00:00:00" or jTimeUse=="0"): # weird problem on ionic.cs.princeton
						jobInitRunTime[lid] = jobInitRunTime[lid] + sleepTime;
						if jobInitRunTime[lid] > 120: # 30s and no results
							print("Invalid job: "+jName+". Deleting... R"+str(jobRestarts[lid]));
							os.system("qdel "+jid);
							del jMap[jid];
							if (jobRestarts[lid] < 3):
								#print("... Restarting invalid job...");
								jobRestarts[lid] = jobRestarts[lid] + 1;
								jobStatus[lid] = "Q";
								newJobStr = ExecuteNoWait(jobCMD[lid], "R"+str(jobRestarts[lid])+"_"+jName);
								#print("Attempting to execute: "+newJobStr);
								newJobSplitStr = newJobStr.split(" ");
								jobID[lid]=newJobSplitStr[0];
								jMap[newJobSplitStr[0]] = lid;
								time.sleep(.1);	# sleep for a bit

			# if some job was never visited - done. erase
			toRemove = [];
			for j in range(0, len(lastVisit)):
				if (visitCounter!=lastVisit[j]):
					toRemove.append(j);
			
			# erase all jobs that were visited
			if (len(toRemove)>0):
				toRemove = sorted(toRemove, reverse=True);
				for j in toRemove:
					lastVisit.pop(j);
					jobID.pop(j);
					jobCMD.pop(j);
					jobStatus.pop(j);
					jobInitRunTime.pop(j);
					jobRestarts.pop(j);
				jMap = dict();
				for j in range(0, len(jobID)):
					jMap[jobID[j]] = j;
		
			counterSleeps += sleepTime;
			if (len(jMap)>0):
				print("        Waiting for jobs: "+str(nRunning)+" / "+str(len(jMap))+" / "+str(nTotal));
				time.sleep(sleepTime)

def WriteJobsFile(jobsIDs, filename):
	if filename != "none":
		f = open(filename, "w");
		f.write("\n".join(jobsIDs));
		f.close();

def ReadJobsFile(filename):
	if (QSub()):
		return open(filename, "r").read().strip().split("\n");
	else:
		return [];

def NumMyJobs():
	subPID = subprocess.Popen("qstat | grep -c $USER", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	nJobs, errors = subPID.communicate();
	if nJobs=='\n' or nJobs=='':
		return 0;
	else:
		return int(nJobs)

def QSub():
	subPID = subprocess.Popen("type qsub", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	qsubExists, errors = subPID.communicate();

	if len(qsubExists)>=4 and qsubExists[:4]=="qsub":
		return 1	# can qsub jobs
	else:
		return 0	# cannot qsub jobs

def ExecuteNoWait(cmd, name):		# (potentially) parallel execution
	if (cmd==""):
		return "-1 NOCMD";
	if QSub():
		cmdToSchedule = cmd;
		currTime = datetime.datetime.now();
		tmpDir="./TempDirQSub_"+str(currTime.month)+"."+str(currTime.day)+"/";
		os.system("if [ ! -d "+tmpDir+" ]; then mkdir "+tmpDir+"; fi");
		timeStr = string.replace(str(currTime), " ", ".");
		timeStr = string.replace(timeStr, ":", ".");
		timeStr = string.replace(timeStr, "-", ".");
		scriptName = tmpDir+timeStr+"_"+name;
		f = open(scriptName+".sh", 'w'); f.write(cmd); f.close();
		os.system("chmod 777 "+scriptName+".sh");
		qsubParams = "";
		#qsubParams += "-l mem=4000mb,cput=1:00:00 ";
		#qsubParams += "-l mem=4000mb,cput=00:1:00 ";	# 1 hr
		qsubParams += "-N "+name;
		qsubParams += " -d "+os.getcwd()+" -o "+scriptName+".log -e "+scriptName+".err";
		cmd = "qsub "+qsubParams+" "+scriptName+".sh ";
		
		if int(NumMyJobs()) >= int(maxJobs):
			while (int(NumMyJobs()) >= .9 * int(maxJobs)):
				print("Too many jobs ("+str(NumMyJobs())+" / "+str(.9 * int(maxJobs))+")... waiting 5s");
				time.sleep(5);

		time.sleep(.01);
		subPID = subprocess.Popen("qsub "+qsubParams+" "+scriptName+".sh ", \
								  shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		idStr, errors = subPID.communicate();

		return idStr.split(".")[0] +" " +cmdToSchedule;
	else:
		os.system(cmd);
		return "-1 NOCMD";


#### This part is on directory structure of the database ####
###		DATA DIRECTORY ####
def MetaDir(keyword):
	return ProcDir + "/"+keyword+"/meta/"

def MyDBDir(keyword):
	return ProcDir + "/"+keyword+"/mydb/"

def ListModels(keyword):
	mListFile = MyDBDir(keyword)+"/../ModelList.txt";
	if os.path.isfile(mListFile):
		f = open(MyDBDir(keyword)+"/../ModelList.txt", "r");
		mList = f.readlines();
		f.close();
		mList = [NameOnly(s.strip()) for s in mList];
	else:
		mList = glob.glob(MyDBDir(keyword)+"/Models/*")
		mList = [NameOnly(s.strip()) for s in mList];
		mList = sorted(mList);
		f = open(mListFile, "w");
		f.write("\n".join(mList))
	return mList

def ListGroundTruthGW(keyword):
	gtDir = MyDBDir(keyword)+"/../gt/";
	mGTListFile = gtDir+"ModelList.txt";
	if os.path.isfile(mGTListFile):
	#if False:
		f = open(mGTListFile, "r");
		mList = f.readlines();
		f.close();
		mList = [NameOnly(s.strip()) for s in mList];
	elif os.path.isdir(gtDir):
		mList = glob.glob(gtDir+"*.bin")
		mList = [NameOnly(s.strip().replace(".bin", "")) for s in mList];
		mList = sorted(mList);
		toRemove = [];
		for mName in mList:
			if not open(gtDir+mName+".txt").readline().strip()=="VALID":
				toRemove.append(mName);
#				print("Removing "+mName)
		for mName in toRemove:  # remove invalid
			mList.remove(mName);
		f = open(mGTListFile, "w");
		f.write("\n".join(mList))
	else:
		mList = [];
	return mList


def NameOnly(filename):
	id = string.rfind(filename, "/")
	return filename[(id+1):]

def MetaFile(keyword, nameOnly):
	return MetaDir(keyword)+nameOnly+"/info.txt"

def ModelsDir(keyword, nameOnly):
	return MyDBDir(keyword)+"/Models/"+nameOnly+"/"

def GeometryDir(keyword, nameOnly):
	return ModelsDir(keyword, nameOnly)+"/Geometry/";

####	WEBPAGE ####
def Webpage(keyword):
	return webpage+keyword+"/"

def WebspaceExists():
	return True;

def InitWeb(keyword):
	InitWebWaiting(keyword, True);

def InitWebWaiting(keyword, waitForMe):
	if WebspaceExists():
		cmd = "";
		web = Webpage(keyword);
		webMeta = Webpage(keyword)+"/meta";
		mListFile = MyDBDir(keyword)+"/../ModelList.txt"
		if not os.path.isdir(web):
			cmd += "mkdir "+web+"; ";
		if not os.path.isdir(webMeta) and os.path.isdir(MetaDir(keyword)):
			cmd += "cp -r "+MetaDir(keyword)+" "+webMeta+"/; "
		if not os.path.isfile(web+"/ModelList.txt") and os.path.isfile(mListFile):
			cmd += "cp "+mListFile+" "+web+"/; "
		if (cmd!=""):
			ExecuteNoWait(cmd, "initweb");
			if waitForMe:
				WaitForAllJobs();


