//
// $Id: BindableNodes.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "BindableNodes.h"
// #include "R3Gems/R3Pilot.h"
#include "R3Io.h"

#define DEBUG(x)





//
// Background
//

void
Background::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Background::eventIn





//
// Fog
//

void
Fog::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Fog::sendEventIn





//
// NavigationInfo
//

void
NavigationInfo::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // NavigationInfo::sendEventIn





//
// Viewpoint
//

void
Viewpoint::add_Ring()
{
  DEBUG(cout << "Adding Viewpoint [" << description.get_string() 
	<< "] to stack" << endl);

  viewpoint_stack.Push(this);
  if (R3Vrml_get_y_up()) {
    R3Vector p = position.get_vector();
    position.set_vector(p.Z(), p.X(), p.Y());
    p = orientation.get_axis();
    orientation.set_axis(p.Z(), p.X(), p.Y());
  }  // x becomes y, y becomes z, z becomes x
}  // Viewpoint::add_Ring



int
Viewpoint::move_to(R3Session *session_ptr)
{
  DEBUG(cout << "Moving to Viewpoint [" << description.get_string() << "]" << endl);

  R3Vector towards;
  if (R3Vrml_get_y_up()) towards.Reset(-1, 0, 0);
  else towards.Reset(0, 0, -1);
  towards.Rotate(orientation.get_axis(), orientation.get_angle());

  R3Point origin = position.get_point();
  
  R3Camera current_camera = session_ptr->Camera();
  R3Camera new_camera(origin, towards, R3posy_vector,
		      current_camera.XFOV(), current_camera.YFOV(),
		      current_camera.Near(), current_camera.Far());
  //  session_ptr->SetCamera(new_camera);
  R3Vector travel_vector = current_camera.Origin() - origin;
  RNScalar distance = travel_vector.Length();
  // should get speed from current NavigationInfo node...
  RNScalar speed = distance/4;
  DEBUG(cout << "Distance between here and there is " << distance
	<< ", speed is " << speed << endl);
  if (distance > 0) {
    // SetAutopilotPath(*session_ptr, new_camera, speed, 0.5);
    return 1;
  }
  return 0;

}  // Viewpoint::move_to



void
Viewpoint::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Viewpoint::sendEventIn





//
// ViewpointStack
//

ViewpointStack viewpoint_stack;

int
ViewpointStack::home_viewpoint(R3Session *session_ptr)
{
  if (current_index == -1) return 0;
  Viewpoint *viewpoint_ptr = my_stack.Kth(0);
  return viewpoint_ptr->move_to(session_ptr);
}  // ViewpointStack::home_viewpoint



int
ViewpointStack::prev_viewpoint(R3Session *session_ptr)
{
  if (current_index == -1) return 0;
  if (current_index == 0) current_index = my_stack.NEntries() - 1;
  else current_index--;
  Viewpoint *viewpoint_ptr = my_stack.Kth(current_index);
  return viewpoint_ptr->move_to(session_ptr);

}  // ViewpointStack::prev_viewpoint



int
ViewpointStack::next_viewpoint(R3Session *session_ptr)
{
  if (current_index == -1) return 0;
  if (current_index == (my_stack.NEntries() - 1)) current_index = 0;
  else current_index++;
  Viewpoint *viewpoint_ptr = my_stack.Kth(current_index);
  return viewpoint_ptr->move_to(session_ptr);

}  // ViewpointStack::next_viewpoint


