//
// $Id: Field.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __FIELD_H
#define __FIELD_H

#include "R3Graphics/R3Graphics.h"
#include "Files.h"
#include <iostream>
#include <string>

static const int MAX_FIELDNAME_LENGTH = 128;

typedef enum { UndefinedType,
	       SFBoolType, SFColorType, MFColorType, SFFloatType,
	       MFFloatType, SFImageType, SFInt32Type, MFInt32Type,
	       SFNodeType, MFNodeType, SFRotationType, MFRotationType,
	       SFStringType, MFStringType, SFTimeType, MFTimeType,
	       SFVec2fType, MFVec2fType, SFVec3fType, MFVec3fType }
FieldVarType;

typedef enum { eventIn, eventOut,
	       exposedField, field } FieldType;

class Event;





class Field
{

public:

  Field();
  virtual ~Field();


  void add_route(Event *new_route);
  int sendEventOut();

  virtual void parse();

  virtual FieldVarType get_field_var_type() { return UndefinedType; }
  virtual char *get_field_var_type_string() { return (char*)"Field"; }

  virtual void sendEventIn(void *new_value) {
    cerr << "Error: called Field's set_value function for a "
	 << get_field_var_type_string() << endl;
  }

  virtual char *to_string(char *buffer) { 
    strcpy(buffer, "Error: called Field's to_string function");
    return NULL;
  }

  virtual Field *create_copy() {
    cerr << "Error: called Field's create_copy function for a "
	 << get_field_var_type_string() << endl;
    return NULL;
  }

  virtual void clear() {}

  //
  // Operators
  //

  void conversion_error(Field *other) {
    cerr << "Error: trying to convert a "
	 << other->get_field_var_type_string() << " to a "
	 << get_field_var_type_string() << endl;
  }  // conversion_error

  void operator_not_impl_error(char *operator_string) {
    cerr << "Error: operator " << operator_string << " not implemented for "
	 << get_field_var_type_string() << endl;
  }  // operator_not_impl_error

  void unary_operator_error(char *operator_string) {
    cerr << "Error: called " << operator_string << " on a "
	 << get_field_var_type_string() << endl;
  }  // unary_operator_error

  void binary_operator_error(char *operator_string, Field *other) {
    cerr << "Error: called " << operator_string << " on a "
	 << get_field_var_type_string() << " and a "
	 << other->get_field_var_type_string() << endl;
  }  // binary_operator_error

  //
  // unary
  //

  virtual Field *operator-() { 
    operator_not_impl_error((char*)"-"); 
    return NULL;
  }
    
  virtual Field *operator!() { 
    operator_not_impl_error((char*)"!"); 
    return NULL;
  }
    
  virtual Field *operator~() { 
    operator_not_impl_error((char*)"~"); 
    return NULL;
  }
    
  // prefix ++ and --

  virtual Field *operator++() { 
    operator_not_impl_error((char*)"prefix ++"); 
    return NULL;
  }
    
  virtual Field *operator--() {
    operator_not_impl_error((char*)"prefix --");
    return NULL;
  }

  // postfix ++ and -- (special C++ way of distinguishing them)

  virtual Field *operator++(int dummy) { 
    operator_not_impl_error((char*)"postfix ++");
    return NULL;
  }
    
  virtual Field *operator--(int dummy) { 
    operator_not_impl_error((char*)"postfix --");
    return NULL;
  }

  //
  // assignment
  //

  virtual Field *operator=(Field *other) { 
    operator_not_impl_error((char*)"="); 
    return NULL;
  }

  virtual Field *operator+=(Field *other) { 
    operator_not_impl_error((char*)"+="); 
    return NULL;
  }

  virtual Field *operator-=(Field *other) { 
    operator_not_impl_error((char*)"-="); 
    return NULL;
  }

  virtual Field *operator*=(Field *other) { 
    operator_not_impl_error((char*)"*="); 
    return NULL;
  }

  virtual Field *operator/=(Field *other) { 
    operator_not_impl_error((char*)"/="); 
    return NULL;
  }

  virtual Field *operator%=(Field *other) { 
    operator_not_impl_error((char*)"%="); 
    return NULL;
  }

  virtual Field *operator&=(Field *other) { 
    operator_not_impl_error((char*)"&="); 
    return NULL;
  }

  virtual Field *operator|=(Field *other) { 
    operator_not_impl_error((char*)"|="); 
    return NULL;
  }

  virtual Field *operator^=(Field *other) { 
    operator_not_impl_error((char*)"^="); 
    return NULL;
  }

  virtual Field *operator<<=(Field *other) { 
    operator_not_impl_error((char*)"<<="); 
    return NULL;
  }

  virtual Field *operator>>=(Field *other) { 
    operator_not_impl_error((char*)">>="); 
    return NULL;
  }

  //  virtual Field *operator>>>=(Field *other) {
  //  operator_not_impl_error(">>>="); 
  //  return NULL;
  //  }

  //
  // binary
  //

  virtual Field *operator==(Field *other) { 
    operator_not_impl_error((char*)"=="); 
    return NULL;
  }

  virtual Field *operator!=(Field *other) { 
    operator_not_impl_error((char*)"!="); 
    return NULL;
  }

  virtual Field *operator<(Field *other) { 
    operator_not_impl_error((char*)"<"); 
    return NULL;
  }

  virtual Field *operator<=(Field *other) { 
    operator_not_impl_error((char*)"<="); 
    return NULL;
  }

  virtual Field *operator>=(Field *other) { 
    operator_not_impl_error((char*)">="); 
    return NULL;
  }

  virtual Field *operator>(Field *other) { 
    operator_not_impl_error((char*)">"); 
    return NULL;
  }

  virtual Field *operator+(Field *other) { 
    operator_not_impl_error((char*)"+"); 
    return NULL;
  }

  virtual Field *operator-(Field *other) { 
    operator_not_impl_error((char*)"-"); 
    return NULL;
  }

  virtual Field *operator*(Field *other) { 
    operator_not_impl_error((char*)"*"); 
    return NULL;
  }

  virtual Field *operator/(Field *other) { 
    operator_not_impl_error((char*)"/"); 
    return NULL;
  }

  virtual Field *operator%(Field *other) { 
    operator_not_impl_error((char*)"%"); 
    return NULL;
  }

  virtual Field *operator&&(Field *other) { 
    operator_not_impl_error((char*)"&&"); 
    return NULL;
  }

  virtual Field *operator||(Field *other) { 
    operator_not_impl_error((char*)"||"); 
    return NULL;
  }

  virtual Field *operator&(Field *other) { 
    operator_not_impl_error((char*)"&"); 
    return NULL;
  }

  virtual Field *operator|(Field *other) { 
    operator_not_impl_error((char*)"|"); 
    return NULL;
  }

  virtual Field *operator^(Field *other) { 
    operator_not_impl_error((char*)"^"); 
    return NULL;
  }

  virtual Field *operator<<(Field *other) { 
    operator_not_impl_error((char*)"<<"); 
    return NULL;
  }

  virtual Field *operator>>(Field *other) { 
    operator_not_impl_error((char*)">>"); 
    return NULL;
  }

  //  virtual Field *operator>>>(Field *other) { 
  //  operator_not_impl_error(">>>");
  //  return NULL;
  //  }

  //
  // tertiary: a ? b : c
  //

private:

  RNArray<Event *> *outgoing_routes;


};  // Field class


#endif
