//
// $Id: Files.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __FILES_H
#define __FILES_H

#include <fstream>
#include <string>
#include "common.h"

static const int MAX_BUFFERSIZE = 128;
static const int MY_EOF = 0;


using namespace std;


class Datafile : public ifstream {

public:

  Datafile();

  ~Datafile() {
    ifstream::close();
  }  // destructor


  int open(char *filespec);
  void set_filespec(char *new_filespec) { my_filespec = new_filespec; }
  char *get_filespec() { return my_filespec; }
  void print_error_line(int warning = 0);

  int get_current_char() { return current_char; }
  void set_current_char(char new_char) { current_char = new_char; }

  void set_parse_source(char *new_string_source);
  char *get_parse_source() { return string_source; }

  void start_storage(char *new_dest) { string_dest = new_dest; }
  void stop_storage() { string_dest = NULL; }

  void reset_line_nr() { line_nr = 0; }
  int get_line_nr() { return line_nr; }
  void set_line_nr(int new_nr) { line_nr = new_nr; }

  void fill_buffer();
  void refill_buffer();
  int get_buffered_char();
  int get_source_char();
  void go_back_char();


protected:

  char *my_filespec;
  int line_nr;

  char buffer[MAX_BUFFERSIZE];
  int buffer_index;

  char *string_source;
  char *string_dest;

  int current_char;


};  // Datafile





static const char OPEN_PAREN = '(';
static const char CLOSE_PAREN = ')';
static const char OPEN_BRACE = '{';
static const char CLOSE_BRACE = '}';
static const char OPEN_BRACKET = '[';
static const char CLOSE_BRACKET = ']';
static const char DOUBLE_QUOTE = '"';
static const char SINGLE_QUOTE = '\'';
static const char BACKSLASH = '\\';
static const char SEMI_COLON = ';';

static const int MAX_TOKEN_LENGTH = 64;

//static const char *number_chars = "+-0123456789.eExX";
static char *default_whitespace_chars = (char*)", \t\n";

static char default_tokens[][MAX_TOKEN_LENGTH] = { "\"", "'", "#", "[", "]",
						   "\\", "{", "}", ".", "\0" };
static const int MAX_DEFAULT_TOKEN_LENGTH = 2;





class Tokenfile : public Datafile {

public:

  Tokenfile() : Datafile() {
    standard_tokens = default_tokens;
    max_standard_token_length = MAX_DEFAULT_TOKEN_LENGTH;
    whitespace_chars = default_whitespace_chars;
  }  // constructor

  ~Tokenfile() {}


  void get_token(char *token);
  void lookahead_token(char *token);
  int accept_token(const char *wanted);
  int accept_token(const char wanted);
  int get_standard_token(char *token);

  void read_until(char until_char);

  void get_number(char *number_string);

  int get_int();
  Float get_float();

  int skip_whitespace();
  int is_whitespace(int k);

  int is_id_char(int k);
  int is_id_firstchar(int k);
  int is_number_char(int k);

  void set_standard_tokens(char (*new_standard_tokens)[MAX_TOKEN_LENGTH]);
  void set_whitespace_chars(char *new_whitespace_chars) {
    whitespace_chars = new_whitespace_chars;
  }


private:

  char (*standard_tokens)[MAX_TOKEN_LENGTH];
  int max_standard_token_length;

  char *whitespace_chars;

  void skip_comments();


};  // Tokenfile





class VrmlFile;
extern VrmlFile *cur_vrml_file;

class VrmlFile : public Tokenfile
{

public:

  VrmlFile() {
    just_parsing = 0;
  }  // constructor

  ~VrmlFile() {}


  void skip_node();

  void set_just_parsing(int new_value) { just_parsing = new_value; }
  int get_just_parsing() { return just_parsing; }


private:

  int just_parsing;


};  // class VrmlFile


#endif



