//
// $Id: GeometryNodes.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "GeometryNodes.h"
#include "VrmlInfo.h"
#include "GeometricProperties.h"

#define DEBUG(x)





void
Box::add_Ring()
{
  R3Vector V = size.get_vector() / 2.0;
  R3Shape *shape = new R3Box(-V.X(), -V.Y(), -V.Z(), V.X(), V.Y(), V.Z());
  R3ShapeNode *node = new R3ShapeNode(shape);
  set_ring_node_ptr(node);
}  // Box::create



void
Box::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // eventIn





//
// Cone
//

void
Cone::add_Ring()
{
  R3Point bottom(0.0, -height.get_float()/2, 0.0);
  R3Point top(0.0, height.get_float()/2, 0.0);
  R3Shape *shape = new R3Cone(bottom, top, bottomRadius.get_float());
  R3ShapeNode *node = new R3ShapeNode(shape);

  set_ring_node_ptr(node);
}  // Cone::add_Ring



void
Cone::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Cone::eventIn





//
// Cylinder
//

void
Cylinder::add_Ring()
{
  R3Point bottom(0.0, -height.get_float()/2, 0.0);
  R3Point top(0.0, height.get_float()/2, 0.0);
  R3Shape *shape = new R3Cylinder(bottom, top, radius.get_float());
  R3ShapeNode *node = new R3ShapeNode(shape);

  set_ring_node_ptr(node);
}  // Cylinder::add_Ring



void
Cylinder::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Cylinder::eventIn





//
// ElevationGrid
//

void
ElevationGrid::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // ElevationGrid::eventIn





//
// Extrusion
//

void
Extrusion::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Extrusion::eventIn





//
// IndexedFaceSet
//

void
IndexedFaceSet::add_Ring()
{
  // what about the fields:
  // solid, convex

  R3FaceSet *shape = new R3FaceSet();
  // now construct the face set
  // assuming we have coord and coordIndex,
  // loop through coordIndex until empty
  //   loop until -1 encountered: create R3FaceVertex for each vertex,
  //     call InsertVertex(R3Vertex *vertex) for each to add to an R3Face
  //     call InsertFace(R3Face *face) for those faces
  //
  R3Face *face = NULL;

  for(int i=0; i < coordIndex.get_size(); i++) {

    int index = coordIndex[i]->get_integer();
    if (index != -1) {
      if (face == NULL) face = new R3Face();
      if (index >= ((Coordinate *) coord.get_node_ptr())->get_size()) {
	cur_vrml_file->print_error_line();
	cerr << "Coordinate index " << index << " out of range ("
	     << ((Coordinate *) coord.get_node_ptr())->get_size() 
	       << " max)" << endl;
	break;
      }

      R3Point point = 
	(*((Coordinate *) coord.get_node_ptr()))[index]->get_point();
      RNScalar x = point.X();
      RNScalar y = point.Y();
      RNScalar z = point.Z();
      R3FaceVertex *face_vertex = new R3FaceVertex(point);

      if (color.get_node_ptr() != NULL) {  // we have colours
	if (colorPerVertex.get_boolean()) {  // assign them per vertex
	  if (colorIndex.get_size() > 0) {  // we also have indices
	    // how do I set the colour of a face vertex?
	  }
	  else {  // we don't have indices, use coordIndex instead
	    // idem
	  }
	}
	else {  // assign colours per face
	  // how do I set the Face colour?
	}
      }  // if, we have colours

      if (normal.get_node_ptr() != NULL) {  // we have normals
	if (normalPerVertex.get_boolean()) {  // assign them per vertex
	  R3Vector V;
	  if (normalIndex.get_size() > 0) {  // we also have indices
	    V = (*((Normal *) normal.get_node_ptr()))[normalIndex[i]->get_integer()]->get_vector();
	  }
	  else {  // we don't have indices, use coordIndex instead
	    V = (*((Normal *) normal.get_node_ptr()))[index]->get_vector();
	  }
	  // have to do it this way because parse_MFVec3f returns Points...
	  face_vertex->SetNormal(V);
	}
	else {  // assign them per face
	  // how to set a face normal?
	}
      }  // if, we have normals
				   
      if (cur_vrml_info->read_textures()) {
	if (texCoord.get_node_ptr() != NULL) {  // we have texture coordinates
	  if (texCoordIndex.get_size() > 0) {  // we also have indices
	    if (i >= texCoordIndex.get_size()) {
	      cur_vrml_file->print_error_line();
	      cerr << "Not enough texture coordinate indices for "
		   << coordIndex.get_size() << " points" << endl;
	    }
	    else {
	      int index = texCoordIndex[i]->get_integer();
	      if (index >= ((TextureCoordinate *) texCoord.get_node_ptr())->get_size()) {
		cur_vrml_file->print_error_line();
		cerr << "Texture coordinate index " << index 
		     << " out of range (" 
		     << ((TextureCoordinate *) texCoord.get_node_ptr())->get_size() - 1
		       << " max)" << endl;
	      }
	      else
		face_vertex->SetTextureCoords((*((TextureCoordinate *) texCoord.get_node_ptr()))[texCoordIndex[i]->get_integer()]->get_point());
	    }
	  }
	  else {  // we don't have indices, use coordIndex instead
	    if (index >= ((TextureCoordinate *) texCoord.get_node_ptr())->get_size()) {
	      cur_vrml_file->print_error_line();
	      cerr << "Texture coordinate index " << index
		   << " out of range ("
		   << ((TextureCoordinate *) texCoord.get_node_ptr())->get_size() - 1
		     << " max)" << endl;
	    }
	    else
	      face_vertex->SetTextureCoords((*((TextureCoordinate *) texCoord.get_node_ptr()))[index]->get_point());
	  }
	} // if, we have texture coordinates
      }  // if we're reading textures

      face->InsertVertex(face_vertex);

      if (i == (coordIndex.get_size() - 1)) {  // missing last -1 marker
	if (!ccw.get_boolean()) face->Flip();
	shape->InsertFace(face);
      }
    }
    else {  // we got -1
      if (face != NULL) {  // at the end it is NULL, in which case we're done
	if (!ccw.get_boolean()) face->Flip();
	shape->InsertFace(face);
	face = NULL;
      }
    }  // else, got -1, the end marker of a face
  }  // for
  
  int nr_coordinates = ((Coordinate *) coord.get_node_ptr())->get_size();

  if (normal.get_node_ptr() == NULL) {  // if there were no normals, generate them
    DEBUG(if (nr_coordinates > 1000)
	  cout << endl 
	  << "Generating IndexedFaceSet normals, please wait..." << endl);
    shape->CreateVertexNormals(creaseAngle.get_float());
    DEBUG(if (nr_coordinates > 1000)
	  cout << "Done generating normals." << endl);
  }
  if (cur_vrml_info->read_textures()) {
    if (texCoord.get_node_ptr() == NULL) {  // if there were no texture coords, generate them
      DEBUG(if (nr_coordinates > 1000)
	    cout << "Generating IndexedFaceSet texture coordinates..." << endl);
      shape->CreateVertexTextureCoords(TRUE);
      DEBUG(if (nr_coordinates > 1000)
	    cout << "Done generating texture coordinates." << endl);
    }
  }  // if we're reading textures

  // create the Ring node and the Vrml node
  R3ShapeNode *node = new R3ShapeNode(shape);

  set_ring_node_ptr(node);

}  // IndexedFaceSet::add_Ring



void
IndexedFaceSet::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // IndexedFaceSet::eventIn





//
// IndexedLineSet
//

void
IndexedLineSet::add_Ring()
{
  // R3PolylineSet *shape = new R3PolylineSet();  ??? doesn't exist
  R3FaceSet *shape = new R3FaceSet();  // just to make it compile

  // now construct the polyline
  // assuming we have coord and coordIndex,
  // loop through coordIndex until empty
  //   loop until -1 encountered: create R3PolylineVertex for each vertex,
  //     call InsertVertex(R3Vertex *vertex) for each to add to an R3Polyline
  //     call InsertPolyline(R3Polyline *polyline) for those faces (???)
  //
  R3Polyline *polyline = NULL;
  for(int i=0; i<coordIndex.get_size(); i++) {
    int index = coordIndex[i]->get_integer();
    if (index != -1) {
      if (polyline == NULL) polyline = new R3Polyline();
      R3PolylineVertex *polyline_vertex =
	new R3PolylineVertex((*((Coordinate *) coord.get_node_ptr()))[index]->get_point());

      if (color.get_node_ptr() != NULL) {  // we have colours
	if (colorPerVertex.get_boolean()) {  // assign them per vertex
	  if (colorIndex.get_size() > 0) {  // we also have indices
	    // how do I set the colour of a polyline vertex?
	  }
	  else {  // we don't have indices, use coordIndex instead
	    // idem
	  }
	}
	else {  // assign colours per polyline
	  // how do I set the Polyline colour?
	}
      }  // if, we have colours

      polyline->InsertVertex(polyline_vertex);
    }
    else {
      if (polyline != NULL) {
	// shape->InsertPolyline(polyline);
	polyline = NULL;
      }
    }  // else, got -1, the end marker of a polyline
  }  // for

  // create the Ring node and the Vrml node
  R3ShapeNode *node = new R3ShapeNode(shape);
  IndexedLineSet *vrml_node = new IndexedLineSet;

  set_ring_node_ptr(node);

}  // IndexedLineSet::add_Ring



void
IndexedLineSet::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // IndexedLineSet::eventIn





//
// PointSet
//

void
PointSet::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // PointSet::eventIn





//
// Sphere
//

void
Sphere::add_Ring()
{
  R3Shape *shape = new R3Sphere(R3Point(0, 0, 0), radius.get_float());
  R3ShapeNode *node = new R3ShapeNode(shape);

  set_ring_node_ptr(node);
}  // Sphere::add_Ring



void
Sphere::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Sphere::eventIn





//
// Text
//

void
Text::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Text::eventIn



void
Text::add_Ring()
{
  // just add a box to keep us from seg faulting
  R3Vector V(0.1, 0.1, 0.1);
  R3Shape *shape = new R3Box(-V.X(), -V.Y(), -V.Z(), V.X(), V.Y(), V.Z());
  R3ShapeNode *node = new R3ShapeNode(shape);
  set_ring_node_ptr(node);
}  // Text::add_Ring


