//
// $Id: GroupingNodes.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "GroupingNodes.h"
#include "parse_nodes.h"
#include "Runtime.h"





void
GroupingNode::init_Ring()
{
  R3GroupNode *group = new R3GroupNode();
  //  group_stack.Head()->InsertNode(group);
  group_stack.Push(group);
  set_group_node_ptr(group);
  set_ring_node_ptr(group);
}  // GroupingNode::init_Ring



void
GroupingNode::add_Ring()
{
  group_stack.RemoveHead();
  group_stack.Head()->InsertNode(get_group_node_ptr());
}  // GroupingNode::create



void
GroupingNode::sendEventIn(int field_id, void *new_value)
{
  //  const int addChildren_index = 0;
  //  const int removeChildren_index = 1;
  //  const int children_index = 2;
  //  const int bboxCenter_index = 3;
  //  const int bboxSize_index = 4;
  
  VrmlNode::sendEventIn(field_id, new_value);

}  // GroupingNode::sendEventIn





//
// Anchor
//

void
Anchor::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Anchor::eventIn





//
// Billboard
//

void
Billboard::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Billboard::eventIn





//
// Collision
//

void
Collision::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Collision::eventIn





//
// DirectionalLight
//

void
DirectionalLight::add_Ring()
{
  R3Light *light = new R3DirectionalLight(direction.get_vector(), 
  					  color.get_colour(), 
  					  (RNScalar) intensity.get_float(),
  					  (RNBoolean) on.get_boolean());
  R3LightNode *node = new R3LightNode(light);
  group_stack.Head()->InsertNode(node);
			
  // Replace top of group_stack
  group_stack.RemoveHead();
  group_stack.Push(node);
  set_group_node_ptr(node);
}  // DirectionalLight::add_Ring



void
DirectionalLight::sendEventIn(int field_id, void *new_value)
{
  //  const int ambientIntensity_index = 5;
  //  const int color_index = 6;
  //  const int direction_index = 7;
  //  const int intensity_index = 8;
  //  const int on_index = 9;
  
  VrmlNode::sendEventIn(field_id, new_value);

  R3LightNode *light_node = (R3LightNode *) get_group_node_ptr();
  R3DirectionalLight *light = (R3DirectionalLight *) light_node->Light();

  switch (field_id) {
    case 6:
      light->SetColor(color.get_colour());
#ifdef __DEBUG__
      cout << "DirectionalLight:: set new colour to [" 
	   << color.get_colour().R() << ", " 
	   << color.get_colour().G() << ", " 
	   << color.get_colour().B() << "]"
	   << endl;
#endif
      break;
      
    case 7:
      light->SetDirection(direction.get_vector()); break;

    case 8:
      light->SetIntensity(intensity.get_float());
#ifdef __DEBUG__
      cout << "DirectionalLight: set new light intensity to " 
	   << intensity.get_float() << endl;
#endif
      break;
      
    case 9:
      light->SetActive(on.get_boolean()); break;

    default:
      event_error(field_id);
  }  // switch

}  // DirectionalLight::eventIn





//
// Group
//

void
Group::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Group::eventIn





//
// Inline
//

#include "R3Vrml.h"
#include "DEFTable.h"

extern DEFTable my_DEF_table;





void
Inline::parse()
{
  // first read the fields
  VrmlNode::parse();

  // then load the inlined file
  int nr_stack_entries_before = group_stack.NEntries();

  char *filespec = url[0]->get_string();
  if (filespec[4] == ':') filespec += 5;  // skip the "file:" part

  VrmlFile *prev_vrml_file = cur_vrml_file;  // keep ptr to current file

  my_DEF_table.store_state();

  R3GroupNode *group = new R3GroupNode();
  set_group_node_ptr(group);
  if (!R3LoadVrml2File((R3Model *) group, group, filespec)) {
    prev_vrml_file->print_error_line();
    cerr << "Couldn't read " << filespec << " in Inline node" << endl;
  }
  group_stack.Head()->InsertNode(group);

  cur_vrml_file = prev_vrml_file;  // restore file ptr

  my_DEF_table.restore_state();

  int nr_stack_entries_after = group_stack.NEntries();
  if (nr_stack_entries_before != nr_stack_entries_after) {
    cerr << "Error: stack is not in the same state after reading "
	 << filespec << endl;
    exit(1);
  }

}  // Inline::parse



void
Inline::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Inline::sendEventIn





//
// PointLight
//

void
PointLight::add_Ring()
{
  // what about the radius?  it seems I can get it, but not set it
  R3Light *light = 
    new R3PointLight(location.get_point(), color.get_colour(), 
		     (RNScalar) intensity.get_float(), 
		     (RNBoolean) on.get_boolean());
  R3LightNode *node = new R3LightNode(light);
  group_stack.Head()->InsertNode(node);
			
  // Replace top of group_stack
  group_stack.RemoveHead();
  group_stack.Push(node);
  set_group_node_ptr(node);
}  // PointLight::add_Ring



void
PointLight::sendEventIn(int field_id, void *new_value)
{
  //  const int ambientIntensity_index = 5;
  //  const int attenuation_index = 6;
  //  const int color_index = 7;
  //  const int intensity_index = 8;
  //  const int location_index = 9;
  //  const int on_index = 10;
  //  const int radius_index = 11;
  
  VrmlNode::sendEventIn(field_id, new_value);

}  // PointLight::eventIn





//
// SpotLight
//

void
SpotLight::add_Ring()
{
  // what about the radius?  it seems I can get it, but not set it
  RNScalar dropoffrate = 0.0; // how to compute from attenuation?
  R3Light *light = 
    new R3SpotLight(location.get_point(), direction.get_vector(), 
		    color.get_colour(), (RNScalar) dropoffrate,
		    (RNAngle) cutOffAngle.get_float(), 
		    (RNScalar) intensity.get_float(), 
		    (RNBoolean) on.get_boolean());
  R3LightNode *node = new R3LightNode(light);
  group_stack.Head()->InsertNode(node);
			
  // Replace top of group_stack
  group_stack.RemoveHead();
  group_stack.Push(node);

  set_group_node_ptr(node);

}  // SpotLight::add_Ring



void
SpotLight::sendEventIn(int field_id, void *new_value)
{
  //  const int ambientIntensity_index = 5;
  //  const int attenuation_index = 6;
  //  const int beamWidth_index = 7;
  //  const int color_index = 8;
  //  const int cutOffAngle_index = 9;
  //  const int direction_index = 10;
  //  const int intensity_index = 11;
  //  const int location_index = 12;
  //  const int on_index = 13;
  //  const int radius_index = 14;
  
  VrmlNode::sendEventIn(field_id, new_value);

  R3LightNode *light_node = (R3LightNode *) get_group_node_ptr();
  R3Light *light = (R3Light *) light_node->Light();

  switch(field_id) {
    case 5:
    case 6:
    case 7:
    case 8:
    case 9:
    case 10:
    case 11:
    case 12:
      break;

    case 13:
      light->SetActive(on.get_boolean()); break;
    default: event_error(field_id);
  }  // switch

}  // SpotLight::eventIn





//
// Switch
//

void
Switch::init_Ring()
{
  switch_node = new R3SwitchNode();
  group_stack.Head()->InsertNode(switch_node);
  group_stack.Push(switch_node);
}  // Switch::init_Ring



void
Switch::add_Ring()
{
  group_stack.RemoveHead();  // just Pop with switch_node didn't compile
  switch_node->SetIndex(whichChoice.get_integer());
  set_group_node_ptr(switch_node);
}  // Switch::add_Ring



void
Switch::sendEventIn(int field_id, void *new_value)
{
  //  const int choice_index = 5;
  //  const int whichChoice_index = 6;
  
  VrmlNode::sendEventIn(field_id, new_value);
  
  if (field_id == 6) {
    switch_node->SetIndex(whichChoice.get_integer());
  }
}  // Switch::eventIn





//
// LOD
//

void
LOD::init_Ring()
{
  lod_node = new R3LodNode();
  group_stack.Head()->InsertNode(lod_node);
  group_stack.Push(lod_node);
}  // LOD::init_Ring



void
LOD::add_Ring()
{
  group_stack.RemoveHead();

  // now set the center (?), and range values for lod_node
  // loop through range array, create interval for each consecutive pair
  // of values
  if (range.get_size() > 0) {
    for(int i=0; i<range.get_size() - 1; i++) {
      Float min = range[i]->get_float();
      Float max = range[i+1]->get_float();
      RNInterval interval(min, max);
      lod_node->InsertRange(interval);
    }  // for
  }
  else {
    // range field is empty, browser should choose correct LOD automatically
  }

  set_group_node_ptr(lod_node);

}  // LOD::add_Ring



void
LOD::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // LOD::eventIn





//
// Transform
//

void
Transform::init_Ring()
{
  affine = new R3Affine(R4identity_matrix);
  transform_node = new R3TransformationNode(affine);
  
  group_stack.Head()->InsertNode(transform_node);
  group_stack.Push(transform_node);

}  // Transform::init_Ring



void
Transform::construct_transform()
{
  affine = new R3Affine(R4identity_matrix);
  affine->Translate(translation.get_vector());
  affine->Translate(center.get_vector());
  rotation.get_axis().Normalize();
  affine->Rotate(rotation.get_axis(), rotation.get_angle());
  affine->Rotate(scaleOrientation.get_axis(), scaleOrientation.get_angle());
  affine->Scale(scale.get_vector());
  affine->Rotate(scaleOrientation.get_axis(), -scaleOrientation.get_angle());
  affine->Translate(-center.get_vector());
}  // Transform::construct_transform



void
Transform::add_Ring()
{
  R3Affine *old_affine = affine;
  construct_transform();

  transform_node->SetTransformation(affine);
  delete old_affine;

  group_stack.RemoveHead();

  set_group_node_ptr(transform_node);

}  // Transform::construct_transform



void
Transform::sendEventIn(int field_id, void *new_value)
{
  //  const int center_index = 5;
  //  const int rotation_index = 6;
  //  const int scale_index = 7;
  //  const int scaleOrientation_index = 8;
  //  const int translation_index = 9;
  
  VrmlNode::sendEventIn(field_id, new_value);

  // first delete the old one (don't name it affine... another great bug :-)
  R3Affine *old_affine = (R3Affine *)
    ((R3TransformationNode *) get_group_node_ptr())->Transformation();
  delete old_affine;

  // this switch no longer needed, we already have the value
  /*  switch(field_id) {
    case center_index: break;
    case rotation_index: break;
    case scale_index: break;
    case scaleOrientation_index: break;
    case translation_index: break;
    default: event_error(field_id);
  }  */

  // construct and set the new one
  construct_transform();
  ((R3TransformationNode *) get_group_node_ptr())->SetTransformation(affine);
  
}  // Transform::eventIn



