//
// $Id: Interpolators.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __INTERPOLATORS_H
#define __INTERPOLATORS_H

#include "VrmlNode.h"
#include "Fields.h"






class Interpolator : public VrmlNode
{

public:

  Interpolator() {
    add_field("set_fraction", &set_fraction, eventIn);
    add_field("key", &key, exposedField);
  }  // constructor

  ~Interpolator() {}


  USE_ActionType get_USE_action() { return ReturnVrmlNode; }


protected:

  SFFloat set_fraction;
  MFFloat key;


};  // Interpolator class





class ScalarInterpolator : public Interpolator
{

public:

  ScalarInterpolator() : Interpolator() { 
    set_node_name("ScalarInterpolator"); 
    set_node_type(ScalarInterpolatorType); 

    add_field("keyValue", &keyValue, exposedField);
    add_field("value_changed", &value_changed, eventOut);

  }  // constructor

  ~ScalarInterpolator() {}


  void sendEventIn(int field_id, void *new_value);


private:

  MFFloat keyValue;
  SFFloat value_changed;


};  // ScalarInterpolator class





class OrientationInterpolator : public Interpolator
{

public:

  OrientationInterpolator() { 
    set_node_name("OrientationInterpolator"); 
    set_node_type(OrientationInterpolatorType);

    add_field("keyValue", &keyValue, exposedField);
    add_field("value_changed", &value_changed, eventOut);

  }  // constructor

  ~OrientationInterpolator() {}


  void sendEventIn(int field_id, void *new_value);


private:

  MFRotation keyValue;
  SFRotation value_changed;


};  // OrientationInterpolator class





class SFVec3fInterpolator : public Interpolator
{

public:

  SFVec3fInterpolator() : Interpolator() {
    add_field("keyValue", &keyValue, exposedField);
    add_field("value_changed", &value_changed, eventOut);
  }  // constructor

  ~SFVec3fInterpolator() {}


  void sendEventIn(int field_id, void *new_value);


private:

  MFVec3f keyValue;
  SFVec3f value_changed;


};  // SFVec3fInterpolator class





class MFVec3fInterpolator : public Interpolator
{

public:

  MFVec3fInterpolator() {
    add_field("keyValue", &keyValue, exposedField);
    add_field("value_changed", &value_changed, eventOut);
  }  // constructor

  ~MFVec3fInterpolator() {}


  void sendEventIn(int field_id, void *new_value);


private:

  MFVec3f keyValue;
  MFVec3f value_changed;

  int nr_outgoing_vectors;

  void send_out_vectors(int start_index);


};  // MFVec3fInterpolator class





class PositionInterpolator : public SFVec3fInterpolator
{

public:

  PositionInterpolator() : SFVec3fInterpolator() { 
    set_node_name("PositionInterpolator"); 
    set_node_type(PositionInterpolatorType);
  }

  ~PositionInterpolator() {}


private:


};  // PositionInterpolator class





class ColorInterpolator : public SFVec3fInterpolator
{

public:

  ColorInterpolator() : SFVec3fInterpolator() { 
    set_node_name("ColorInterpolator"); 
    set_node_type(ColorInterpolatorType); 
  }  // constructor

  ~ColorInterpolator() {}


private:

};  // ColorInterpolator class





class CoordinateInterpolator : public MFVec3fInterpolator
{

public:

  CoordinateInterpolator() { 
    set_node_name("CoordinateInterpolator"); 
    set_node_type(CoordinateInterpolatorType); 
  }  // constructor

  ~CoordinateInterpolator() {}


private:

};  // CoordinateInterpolator class





class NormalInterpolator : public MFVec3fInterpolator
{

public:

  NormalInterpolator() { 
    set_node_name("NormalInterpolator");
    set_node_type(NormalInterpolatorType);
  }  // constructor

  ~NormalInterpolator() {}


private:

};  // NormalInterpolator class



#endif


