//
// $Id: MatAppShape.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __MATAPPSHAPE_H
#define __MATAPPSHAPE_H

#include "VrmlNode.h"
#include "Fields.h"





class Material : public VrmlNode
{

public:

  Material() { 
    set_node_name("Material");
    set_node_type(MaterialType);

    add_field("ambientIntensity", &ambientIntensity, exposedField);
    add_field("diffuseColor", &diffuseColor, exposedField);
    add_field("emissiveColor", &emissiveColor, exposedField);
    add_field("shininess", &shininess, exposedField);
    add_field("specularColor", &specularColor, exposedField);
    add_field("transparency", &transparency, exposedField);

    ambientIntensity.set_float(0.2);
    diffuseColor.set_colour(0.8, 0.8, 0.8);
    emissiveColor.set_colour(0.0, 0.0, 0.0);
    shininess.set_float(0.0);
    specularColor.set_colour(0.0, 0.0, 0.0);
    transparency.set_float(0.0);

    add_Ring();

  }  // constructor

  ~Material() {} 


  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  R3Brdf *get_brdf_ptr() { return brdf_ptr; }

  USE_ActionType get_USE_action() { return ReturnVrmlNode; }


private:

  SFFloat ambientIntensity;
  SFColor diffuseColor;
  SFColor emissiveColor;
  SFFloat shininess;
  SFColor specularColor;
  SFFloat transparency;

  R3Brdf *brdf_ptr;


};  // Material class





class Appearance : public VrmlNode
{

public:

  Appearance() { 
    set_node_name("Appearance"); 
    set_node_type(AppearanceType); 

    add_field("material", &material, exposedField);
    add_field("texture", &texture, exposedField);
    add_field("textureTransform", &textureTransform, exposedField);

    material.set_node_ptr(NULL);
    texture.set_node_ptr(NULL);
    textureTransform.set_node_ptr(NULL);

    has_texture = FALSE;

  }  // constructor

  ~Appearance() {}


  void init_Ring();
  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return ReturnVrmlNode; }


private:

  SFNode material;
  SFNode texture;
  SFNode textureTransform;

  RNBoolean has_texture;

};  // Appearance class





class Shape : public VrmlNode
{

public:

  Shape() { 
    set_node_name("Shape"); 
    set_node_type(ShapeType);

    add_field("appearance", &appearance, exposedField);
    add_field("geometry", &geometry, exposedField);

    appearance.set_node_ptr(NULL);
    geometry.set_node_ptr(NULL);

  }  // constructor

  ~Shape() {}


  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertNode; }


private:

  SFNode appearance;
  SFNode geometry;


};  // Shape class


#endif

