/* Include file for the R3 session class */



/* Initialization functions */

int R3InitSession();
void R3StopSession();



/* Class definition */

class R3Session : public R3View {
    public:
        // Constructor functions
	R3Session(void);
        R3Session(const R3Session& session);
        R3Session(const R3View& view);
        R3Session(const R3Camera& camera, const R2Viewport& viewport, R3Model *model);
	~R3Session(void);

	// Session property functions
	const int ID(void) const;
	const R3Frame *Frame(void) const;
	const R3Frame *PreviousFrame(void) const;
	const RNScalar SimulationTime(void) const;
	const RNScalar FrameRate(void) const;
	const RNFlags Flags(void) const;

	// Sequence functions/operators
	virtual void BeginSequence(void);
	virtual void EndSequence(void);

	// Frame functions/operators
	virtual void RefreshFrame(void);
	virtual void BeginFrame(void);
	virtual void SimulateFrame(void);
	virtual void DrawFrame(void);
	virtual void EndFrame(void);

	// Message functions/operators
	int SendMessage(R3TraversalState& state, R3Node *dst, 
	    R3MessageType type, const void *buffer = NULL, int length = 0);
	int SendPickMessages(int x, int y,
	    R3MessageType type = R3_PICK_MESSAGE, const void *buffer = NULL, int length = 0);

	// Do not use these ???
	const RNScalar LodThreshold(void) const;
	void SetLodThreshold(RNScalar t);

    private:
	int id;
	R3Frame *frame;
	RNQueue<R3Frame *> frames;
	RNScalar frame_rate;
	RNScalar simulation_time;
	RNScalar lod_threshold;
	RNScalar delta_lod_threshold;
	RNFlags flags;
};



/* Public variables */

extern R3Session R3null_session;
extern RNBoolean R3session_update_lod_threshold;
extern RNTime R3start_time;



/* Public session flags */

extern const RNFlags R3_SESSION_INSIDE_SEQUENCE_FLAG;
extern const RNFlags R3_SESSION_BEGINNING_SEQUENCE_FLAG;



/* Inline functions */

inline const int R3Session::
ID(void) const
{
    // Return ID
    return id;
}



inline const R3Frame *R3Session::
Frame(void) const
{
    // Return current frame
    return frame;
}



inline const R3Frame *R3Session::
PreviousFrame(void) const
{
    // Return previous frame
    return (frames.IsEmpty()) ? NULL : frames.Head();
}



inline const RNScalar R3Session::
SimulationTime(void) const
{
    // Return simulation time
    return simulation_time;
}



inline const RNScalar R3Session::
FrameRate(void) const
{
    // Return averaged frame rate
    return frame_rate;
}



inline const RNFlags R3Session::
Flags(void) const
{
    // Return session flags
    return flags;
}



inline int R3Session::
SendMessage(R3TraversalState& state, R3Node *dst, 
    R3MessageType type, const void *buffer, int length)
{
    // Send message to dst
    return dst->ReceiveMessage(*this, state, NULL, type, buffer, length);
}



inline const RNScalar R3Session::
LodThreshold(void) const
{
    // Return lod threshold
    return lod_threshold;
}



inline void R3Session::
SetLodThreshold(RNScalar t)
{
    // Set lod threshold
    lod_threshold = t;
}



