//
// $Id: Textures.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "R3Nodes/R3Nodes.h"
#include "R2Images/R2Io.h"
#include "R3Io.h"
#include "Textures.h"
#include "VrmlInfo.h"






void
Texture::add_Ring()
{
  if (!read_an_image) return;

  R2TextureWrap swrap, twrap;
  if (repeatS.get_boolean()) swrap = R2_REPEAT_TEXTURE_WRAP;
  else swrap = R2_NO_TEXTURE_WRAP;

  if (repeatT.get_boolean()) twrap = R2_REPEAT_TEXTURE_WRAP;
  else twrap = R2_NO_TEXTURE_WRAP;

  r2_texture = new R2Texture(image.get_image(), swrap, twrap);
  
}  // Texture::add_Ring





//
// TextureTransform
//

void
TextureTransform::add_Ring()
{
  R2Affine affine(R3identity_matrix);

  affine.Translate(translation.get_vector());
  affine.Translate(center.get_vector());
  affine.Rotate(rotation.get_float());
  affine.Scale(scale.get_vector());
  affine.Translate(-center.get_vector());

  R2TextureTransformation *texture_transform =
    new R2TextureTransformation(affine);

  if (ring_ptr) delete ring_ptr;
  ring_ptr = texture_transform;

}  // TextureTransform::add_Ring



void
TextureTransform::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // TextureTransform::eventIn





//
// ImageTexture
//

static RNArray<ImageTexture *> image_texture_ptrs;

void
ImageTexture::read_texture()
{
  read_an_image = 0;
      
  if (!cur_vrml_info->read_textures()) {  // maybe we're not doing textures
    return;
  }

  char *filespec;

  // just get the first string, and treat it as a filename
  if (url.get_size() == 0) return;
  filespec = url[0]->get_string();
  if (filespec) read_an_image = 1;
  else return;

  int prev_index = already_used(filespec);
  if (prev_index >= 0) {
    if (R3Vrml_get_info_output()) {
      cout << "The texture: " << filespec << endl
	   << "  has already been loaded. Adding a reference to it." << endl;
    }
    ImageTexture *prev_texture = image_texture_ptrs.Kth(prev_index);
    image.set_image(prev_texture->get_SFImage()->get_image());
    reusing_texture = 1;
  }
  else {
    if (filespec[4] == ':') filespec += 5;  // skip the "file:" part
    image.set_image(R2ReadImage(filespec));
  }  // else, read the texture file

  if (!image.get_image()) {
    cur_vrml_file->print_error_line();
    cerr << "Did not read an image" << endl;
    read_an_image = 0;
    return;
  }

  if (read_an_image && !reusing_texture) {
    cur_vrml_info->add_texture_size(image.get_size_in_bytes());
    if (R3Vrml_get_info_output()) {
      cout << "Read texture: " << filespec << endl
	   << "  dimensions: " << image.get_image()->Width() << " x "
	   << image.get_image()->Height() << " x "
	   << image.get_image()->Depth() << " (W x H x D) = "
	   << image.get_size_in_bytes() << " bytes" << endl;
    }
  }

}  // ImageTexture::read_texture



void
ImageTexture::create_dummy_texture()
{
  unsigned char *image_data = new unsigned char[12];
  for(int i=0; i<12; i++) image_data[i] = 128;
  R2ColorImage *r2_image = new R2ColorImage(2, 2, 3, image_data);
  image.set_image(r2_image);
}  // ImageTexture::create_dummy_texture



int
ImageTexture::already_used(char *filespec)
{
  int nr_prev_textures = image_texture_ptrs.NEntries();
  //  cur_vrml_file->print_error_line();
  //  cout << "ImageTexture::already_used, " << nr_prev_textures 
  //       << " prev entries" << endl;
  for(int i=0; i<nr_prev_textures; i++) {
    char *other_filespec = image_texture_ptrs.Kth(i)->get_filespec();
    if (strcmp(filespec, other_filespec) == 0)
      return i;
  }
  return -1;
}  // ImageTexture::already_used



void
ImageTexture::add_Ring()
{
  read_texture();
  Texture::add_Ring();
  if (!reusing_texture && read_an_image) image_texture_ptrs.Insert(this);
}  // ImageTexture::add_Ring



void
ImageTexture::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // ImageTexture::eventIn





//
// MovieTexture
//

void
MovieTexture::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // MovieTexture::eventIn





//
// PixelTexture
//

void
PixelTexture::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // PixelTexture::eventIn



