//
// $Id: Textures.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __TEXTURES_H
#define __TEXTURES_H

#include "VrmlNode.h"
#include "Fields.h"





class Texture : public VrmlNode
{

public:

  Texture() {
    add_field("repeatS", &repeatS, field);
    add_field("repeatT", &repeatT, field);

    repeatS.set_boolean(TRUE);
    repeatT.set_boolean(TRUE);
  }  // constructor

  ~Texture() {}


  R2Texture *get_ring_ptr() { return r2_texture; }
  void set_ring_ptr(R2Texture *new_texture) { r2_texture = new_texture; }
  SFImage *get_SFImage() { return &image; }

  void add_Ring();
  int have_read_image() { return read_an_image; }
  int get_size_in_bytes() { return image.get_size_in_bytes(); }


protected:

  SFBool repeatS;
  SFBool repeatT;

  SFImage image;
  R2Texture *r2_texture;

  int read_an_image;


};  // Texture class





class TextureTransform : public VrmlNode
{

public:

  TextureTransform() { 
    set_node_name("TextureTransform"); 
    set_node_type(TextureTransformType); 

    add_field("center", &center, exposedField);
    add_field("rotation", &rotation, exposedField);
    add_field("scale", &scale, exposedField);
    add_field("translation", &translation, exposedField);

    center.set_vector(0, 0);
    rotation.set_float(0);
    scale.set_vector(1, 1);
    translation.set_vector(0, 0);

    ring_ptr = NULL;
    add_Ring();

  }  // constructor

  ~TextureTransform() {}


  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  R2TextureTransformation *get_ring_ptr() { return ring_ptr; }

  USE_ActionType get_USE_action() { return ReturnVrmlNode; }


private:

  SFVec2f center;
  SFFloat rotation;
  SFVec2f scale;
  SFVec2f translation;
  
  R2TextureTransformation *ring_ptr;


};  // TextureTransform class





class ImageTexture : public Texture
{

public:

  ImageTexture() {
    set_node_name("ImageTexture"); 
    set_node_type(ImageTextureType); 

    add_field("url", &url, exposedField);

    reusing_texture = 0;

  }  // ImageTexture constructor
    
  ~ImageTexture() {}


  void sendEventIn(int field_id, void *new_value);
  void add_Ring();

  USE_ActionType get_USE_action() { return ReturnVrmlNode; }

  char *get_filespec() { return url[0]->get_string(); }


private:

  void read_texture();
  void create_dummy_texture();
  int already_used(char *filespec);
  int reusing_texture;

  MFString url;


};  // ImageTexture class





class MovieTexture : public VrmlNode
{

public:

  MovieTexture() { 
    set_node_name("MovieTexture"); 
    set_node_type(MovieTextureType);

    add_field("loop", &loop, exposedField);
    add_field("speed", &speed, exposedField);
    add_field("startTime", &startTime, exposedField);
    add_field("stopTime", &stopTime, exposedField);
    add_field("url", &url, exposedField);
    add_field("duration_changed", &duration_changed, eventOut);
    add_field("isActive", &isActive, eventOut);

    loop.set_boolean(FALSE);
    speed.set_float(1);
    startTime.set_time(0);
    stopTime.set_time(0);

  }  // constructor

  ~MovieTexture() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return Unknown; }


private:

  SFBool loop;
  SFFloat speed;
  SFTime startTime;
  SFTime stopTime;
  MFString url;
  SFFloat duration_changed;
  SFBool isActive;


};  // MovieTexture class





class PixelTexture : public Texture
{

public:

  PixelTexture() : Texture() { 
    set_node_name("PixelTexture");
    set_node_type(PixelTextureType);

    add_field("image", &image, exposedField);
  }  // constructor

  ~PixelTexture() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return ReturnVrmlNode; }


private:

  // SFImage image already in Texture class


};  // PixelTexture class


#endif

