//
// $Id: VrmlNode.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "VrmlNode.h"
#include "Runtime.h"






VrmlNode::VrmlNode()
{ 
  node_name = new char[MAX_TOKEN_LENGTH];
  node_ptr = NULL;
  fields = new VsSymbolTable;
}  // constructor



VrmlNode::~VrmlNode()
{
#ifdef __DEBUG_DESTRUCTORS__
  cout << "VrmlNode destructor" << endl;
#endif
  // don't do anything because the fields contain pointer to member variables
}  // destructor



VrmlNode::VrmlNode(VrmlNode& other)
{
  node_name = new char[MAX_TOKEN_LENGTH];
  strcpy(node_name, other.get_node_name());
  node_type = other.get_node_type();
  node_ptr = other.get_ring_node_ptr();

  fields = new VsSymbolTable(*other.get_var_table_ptr());

}  // VrmlNode copy constructor

	 

void
VrmlNode::parse()
{
  cur_vrml_file->accept_token(OPEN_BRACE);

  char token[MAX_TOKEN_LENGTH];
  int done = 0;

  while (!done) {
    cur_vrml_file->get_token(token);
    if (token[0] == CLOSE_BRACE) done = 1;
    else {
      VsVariableNode *var_ptr = (VsVariableNode *) fields->find_symbol(token);
      if (var_ptr) var_ptr->get_field_ptr()->parse();
      else {
	parse_error(UnknownField, token);
	cur_vrml_file->read_until(CLOSE_BRACE);
      }
    }
  }  // while
}  // VrmlNode::parse



void
VrmlNode::parse_error(ParseErrorType error_type, char *token)
{
  cerr << "Parse error: ";
  switch(error_type) {
    case UnknownField : 
      cur_vrml_file->print_error_line();
      cerr << "Unknown field [" << token << "]" << endl;
      break;
    default:
      cur_vrml_file->print_error_line();
      cerr << "(unknown error)" << endl;
  }  // switch
}  // VrmlNode::parse_error




void
VrmlNode::event_error(int field_id)
{
  cerr << "Event error: unknown field id "
       << field_id << " in node " << node_name << endl;
}  // VrmlNode::event_error



void
VrmlNode::no_eventIn_error(int field_id)
{
  cerr << "Event error: no eventIn/exposedField fields in node "
       << node_name << endl;
}  // VrmlNode::no_eventIn_error



void
VrmlNode::sendEventIn(int field_id, void *new_value)
{
  (*fields)[field_id]->get_field_ptr()->sendEventIn(new_value);
}  // eventIn

