//
// $Id: VrmlScript.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "VrmlScript.h"
#include "VsExprNodes.h"

static char VRMLSCRIPT_TOKENS[][MAX_TOKEN_LENGTH] = {
  "{", "}", "(", ")", "[", "]", ";", ",", "-", "!", "~", "=", "+=", "-=", "*=",
  "/=", "%=", "&=", "|=", "^=", "<<=", ">>=", ">>>=", "++", "--", "?",
  ":", "==", "!=", "<", "<=", ">=", ">", "+", "*", "/", "%", "&&", "||",
  "&", "|", "^", "<<", ">>", ">>>", "'", ".", "\0" };





VrmlScript::VrmlScript()
{
  source = NULL;
  scriptfile = NULL;
  source_info = NULL;
}  // default constructor



VrmlScript::VrmlScript(char *new_source)
{
  source = new_source;
  scriptfile = new VrmlFile;
  scriptfile->set_parse_source(source);
  scriptfile->set_standard_tokens(VRMLSCRIPT_TOKENS);
  scriptfile->set_whitespace_chars(vrmlscript_whitespace_chars);
  source_info = NULL;
}  // constructor



void
VrmlScript::set_source_info(char *filename, int line_nr)
{
  int nr_chars = strlen(filename) + 64;
  source_info = new char[nr_chars];
  sprintf(source_info, "%s at line %d, in script", filename, line_nr);
  scriptfile->set_filespec(source_info);

}  // VrmlScript::set_source_info

  

void
VrmlScript::parse()
{
  int done = 0;

  VrmlFile *prev_vrml_file = cur_vrml_file;
  cur_vrml_file = scriptfile;
  cur_symbol_table->store_state();

  while (!done) {
    int k = scriptfile->skip_whitespace();

    if (k == MY_EOF) done = 1;
    else {
      VsFunctionNode *function_ptr = new VsFunctionNode;
      cur_symbol_table->add(function_ptr);  // may be recursive
      function_ptr->parse(scriptfile);
      functions.add(function_ptr);
    }
  }  // while

  cur_symbol_table->restore_state();
  cur_vrml_file = prev_vrml_file;

}  // VrmlScript::parse



void
VrmlScript::execute(char *function_id, int argc, Field **argv)
{
  // find the function
  int nr_functions = functions.get_size();

  VsFunctionNode *function_ptr = 
    (VsFunctionNode *) functions.find_symbol(function_id);

  if (function_ptr) {
    cur_symbol_table->store_state();
    // also add all our functions to the symbol table
    for(int i=0; i<nr_functions; i++) {
      cur_symbol_table->add(functions[i]);
    }
    function_ptr->evaluate(argc, argv);
    cur_symbol_table->restore_state();
  }
  else
    cerr << "Warning: attempt to execute a non-existing VrmlScript function: "
	 << function_id << endl;

}  // VrmlScript::execute
