//
// $Id: VsStatements.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include <iostream.h>
#include "VsStatements.h"





//
// VsReturnStatement
//

void
VsReturnStatement::parse(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsReturnStatement::parse" << endl;
#endif

  tokenfile_ptr->accept_token("return");
  int k = tokenfile_ptr->skip_whitespace();

  if (k == SEMI_COLON)
    tokenfile_ptr->accept_token(SEMI_COLON);
  else
    return_value.parse_entry(tokenfile_ptr);

}  // VsReturnStatement::parse



Field *
VsReturnStatement::evaluate()
{
  Field *field_ptr = return_value.evaluate();

  if (!field_ptr) field_ptr = new Field;
  return field_ptr;

}  // VsReturnStatement::evaluate





//
// VsExprStatement
//

void
VsExprStatement::parse(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsExprStatement::parse" << endl;
#endif
  expr_ptr = new VsExpression;
  expr_ptr->parse_entry(tokenfile_ptr);

}  // VsExprStatement::parse



Field *
VsExprStatement::evaluate()
{
#ifdef __DEBUG_VRMLSCRIPT__
  cout << "Executing expression statement" << endl;
#endif

  expr_ptr->evaluate();
  // don't return anything, otherwise this'll be interpreted as a
  // return statement (maybe need other solution, return flag)
  return NULL;

}  // VsExprStatement::execute





//
// VsStatementBlock
//

void
VsStatementBlock::parse(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsStatementBlock::parse" << endl;
#endif
  char token[MAX_TOKEN_LENGTH];
  int expect_close_brace = 0;

  int k = tokenfile_ptr->skip_whitespace();
  if (k == OPEN_BRACE) {
    tokenfile_ptr->get_token(token);  // may or may not have braces
    expect_close_brace = 1;
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
    cout << "  found a {" << endl;
#endif
  }  // if, open brace encountered

  int done = 0;
  VsStatement *statement_ptr;

  while (!done) {
    tokenfile_ptr->lookahead_token(token);
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
    cout << "  got token [" << token << "]" << endl;
#endif
    if (!token[0]) {
      tokenfile_ptr->print_error_line();
      cerr << "Unexpected end of VrmlScript code" << endl;
      return;
    }
    else if (token[0] == CLOSE_BRACE) {
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
      cout << "  found a }" << endl;
#endif
      if (expect_close_brace) done = 1;
      else {
	tokenfile_ptr->print_error_line();
	cerr << "Unexpected }" << endl;
	return;
      }
    }
    else {
      if (strcmp(token, "if") == 0) 
	statement_ptr = new VsIfStatement;
      else if (strcmp(token, "for") == 0)
	statement_ptr = new VsForStatement;
      else if (strcmp(token, "while") == 0)
	statement_ptr = new VsWhileStatement;
      else if (strcmp(token, "debug") == 0)
	statement_ptr = new VsDebugStatement;
      else if (strcmp(token, "return") == 0)
	statement_ptr = new VsReturnStatement;
      else statement_ptr = new VsExprStatement;

      statement_ptr->parse(tokenfile_ptr);
      statements.Insert(statement_ptr);

      tokenfile_ptr->lookahead_token(token);
      if (token[0] == SEMI_COLON) tokenfile_ptr->get_token(token);
      if (!expect_close_brace) done = 1;  // only one statement in the block
    }
  }  // while
  if (expect_close_brace) 
    tokenfile_ptr->get_token(token);  // get the close brace

}  // VsStatementBlock::parse



Field *
VsStatementBlock::evaluate()
{
  int nr_statements = statements.NEntries();

#ifdef __DEBUG_VRMLSCRIPT__
  cout << "Executing statement block with " << nr_statements
       << " statement(s)" << endl;
#endif

  Field *return_value;
  for(int i=0; i<nr_statements; i++) {
    return_value = statements.Kth(i)->evaluate();
    if (return_value) return return_value;
  }  // for
  return NULL;

}  // VsStatementBlock::execute





//
// VsIfStatement
//

void
VsIfStatement::parse(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsIfStatement::parse" << endl;
#endif
  tokenfile_ptr->accept_token("if");
  condition.parse_entry(tokenfile_ptr);
  then_statements.parse(tokenfile_ptr);

  char token[MAX_TOKEN_LENGTH];
  tokenfile_ptr->lookahead_token(token);
  if (strcmp(token, "else") == 0) {
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
    cout << "  found an else" << endl;
#endif
    tokenfile_ptr->get_token(token);
    else_statements.parse(tokenfile_ptr);
  }

}  // VsIfStatement::parse



Field *
VsIfStatement::evaluate()
{
  // evaluate condition
  SFBool result;
  result = condition.evaluate();  // returns a pointer to an SFBool

#ifdef __DEBUG_VRMLSCRIPT__
  cout << "Executing if statement: (" << result.get_boolean() << ")" << endl;
#endif

  // depending on result, run then_statements or else_statements
  if (result.get_boolean()) return then_statements.evaluate();
  else return else_statements.evaluate();

}  // VsIfStatement::execute





//
// VsWhileStatement
//

void
VsWhileStatement::parse(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsWhileStatement::parse" << endl;
#endif

  tokenfile_ptr->accept_token("while");

  tokenfile_ptr->accept_token(OPEN_PAREN);
  condition.parse_entry(tokenfile_ptr);
  tokenfile_ptr->accept_token(CLOSE_PAREN);

  loop_statements.parse(tokenfile_ptr);

}  // VsWhileStatement::parse



Field *
VsWhileStatement::evaluate()
{
  while(((SFBool *) condition.evaluate())->get_boolean()) {
    Field *return_value = loop_statements.evaluate();
    if (return_value) return return_value;
  }  // while
  return NULL;

}  // VsWhileStatement::execute


  


//
// VsForStatement
//

void
VsForStatement::parse(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsForStatement::parse" << endl;
#endif

  tokenfile_ptr->accept_token("for");
  tokenfile_ptr->accept_token(OPEN_PAREN);

  init_statement.parse(tokenfile_ptr);
  tokenfile_ptr->accept_token(SEMI_COLON);
  condition.parse_entry(tokenfile_ptr);
  tokenfile_ptr->accept_token(SEMI_COLON);
  incr_statement.parse(tokenfile_ptr);

  tokenfile_ptr->accept_token(CLOSE_PAREN);

  loop_statements.parse(tokenfile_ptr);

}  // VsForStatement::parse



Field *
VsForStatement::evaluate()
{
  
  for(init_statement.evaluate(); 
      ((SFBool *) condition.evaluate())->get_boolean();
      incr_statement.evaluate()) {
    Field *return_value = loop_statements.evaluate();
    if (return_value) return return_value;
  }  // for
  return NULL;

}  // VsForStatement::execute


  


//
// VsDebugStatement
//

void
VsDebugStatement::parse(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsDebugStatement::parse" << endl;
#endif
  tokenfile_ptr->accept_token("debug");

  expression.parse_entry(tokenfile_ptr);

}  // VsDebugStatement::parse



Field *
VsDebugStatement::evaluate()
{
  Field *result = expression.evaluate();
  char value[MAX_TOKEN_LENGTH];
  result->to_string(value);

  cout << "VrmlScript Debug: " << value << endl;
  return NULL;

}  // VsDebugStatement::execute


