//
// $Id: VsStatements.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __VRMLSCRIPT_STATEMENTS_H
#define __VRMLSCRIPT_STATEMENTS_H

#include "R3Graphics/R3Graphics.h"
#include "Files.h"
#include "VsExpression.h"
#include "Field.h"





class VsStatement
{

public:

  VsStatement() {}
  ~VsStatement() {}

  virtual void parse(Tokenfile *tokenfile_ptr) {}
  virtual Field *evaluate() { 
    cerr << "Error: called VsStatement's evaluate function" << endl;
    return NULL;
  }


private:


};  // VsStatement class





class VsReturnStatement : public VsStatement
{

public:

  VsReturnStatement() {}
  ~VsReturnStatement() {}


  void parse(Tokenfile *tokenfile_ptr);
  Field *evaluate();


private:

  VsExpression return_value;


};  // VsReturnStatement class





class VsStatementBlock
{

public:

  VsStatementBlock() {}
  ~VsStatementBlock() {
    int nr_statements = statements.NEntries();
    for(int i=0; i<nr_statements; i++) delete statements.Kth(i);
  }  // destructor


  void parse(Tokenfile *tokenfile_ptr);
  Field *evaluate();


private:

  RNArray<VsStatement *> statements;


};  // VsStatementBlock class





class VsIfStatement : public VsStatement
{

public:

  VsIfStatement() {}
  ~VsIfStatement() {}


  void parse(Tokenfile *tokenfile_ptr);
  Field *evaluate();


private:

  VsExpression condition;
  VsStatementBlock then_statements;
  VsStatementBlock else_statements;


};  // VsIfStatement class





class VsExprStatement : public VsStatement
{

public:

  VsExprStatement() {}
  ~VsExprStatement() {
    if (expr_ptr) delete expr_ptr;
  }

  void parse(Tokenfile *tokenfile_ptr);
  Field *evaluate();


private:

  VsExpression *expr_ptr;


};  // VsExprStatement class





class VsWhileStatement : public VsStatement
{

public:

  VsWhileStatement() {}
  ~VsWhileStatement() {}


  void parse(Tokenfile *tokenfile_ptr);
  Field *evaluate();


private:

  VsExpression condition;

  VsStatementBlock loop_statements;


};  // VsWhileStatement class





class VsForStatement : public VsStatement
{

public:

  VsForStatement() {}
  ~VsForStatement() {}


  void parse(Tokenfile *tokenfile_ptr);
  Field *evaluate();


private:

  VsExprStatement init_statement;
  VsExpression condition;
  VsExprStatement incr_statement;

  VsStatementBlock loop_statements;


};  // VsForStatement class





class VsDebugStatement : public VsStatement
{

public:

  VsDebugStatement() {}
  ~VsDebugStatement() {}


  void parse(Tokenfile *tokenfile_ptr);
  Field *evaluate();


private:

  VsExpression expression;


};  // VsDebugStatement class


#endif

