//
// $Id: parse_nodes.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//
// node parsing functions
//

#include "R3Nodes/R3Nodes.h"
#include "vrml_includes.h"
#include "parse_nodes.h"
#include "Files.h"

#include "Runtime.h"
#include "DEFTable.h"
#include "PROTONodes.h"
#include "UserNodeList.h"

static PROTOList my_protos;

//
// ROUTE, DEF, USE
//
void parse_route();
VrmlNode *parse_DEF();
VrmlNode *parse_USE();





//
// parse_node
//

VrmlNode *parse_grouping(char *node_type)
{
  if (strcmp(node_type, "Anchor") == 0) return new Anchor;
  else if (strcmp(node_type, "Billboard") == 0) return new Billboard;
  else if (strcmp(node_type, "Collision") == 0) return new Collision;
  else if (strcmp(node_type, "Group") == 0) return new Group;
  else if (strcmp(node_type, "Transform") == 0) return new Transform;
  else
    return NULL;

}  // parse_grouping



VrmlNode *parse_lights(char *node_type)
{
  if (strcmp(node_type, "DirectionalLight") == 0) return new DirectionalLight;
  else if (strcmp(node_type, "PointLight") == 0) return new PointLight;
  else if (strcmp(node_type, "SpotLight") == 0) return new SpotLight;
  else
    return NULL;
}  // parse_lights



VrmlNode *parse_sensors(char *node_type)
{
  if (strcmp(node_type, "CylinderSensor") == 0) return new CylinderSensor;
  else if (strcmp(node_type, "PlaneSensor") == 0) return new PlaneSensor;
  else if (strcmp(node_type, "ProximitySensor") == 0) 
    return new ProximitySensor;
  else if (strcmp(node_type, "SphereSensor") == 0) return new SphereSensor;
  else if (strcmp(node_type, "TimeSensor") == 0) {
    VrmlNode *vrml_node_ptr = new TimeSensor;
    // add to our global list of TimeSensors
    timesensors.add_timesensor((TimeSensor *) vrml_node_ptr);
    return vrml_node_ptr;
  }
  else if (strcmp(node_type, "TouchSensor") == 0) {
    VrmlNode *vrml_node_ptr = new TouchSensor;
    touchsensors.add_touchsensor((TouchSensor *) vrml_node_ptr);
    return vrml_node_ptr;
  }
  else if (strcmp(node_type, "VisibilitySensor") == 0) 
    return new VisibilitySensor;
  else
    return NULL;
}  // parse_sensors



VrmlNode *parse_special(char *node_type)
{
  if (strcmp(node_type, "Inline") == 0) return new Inline;
  else if (strcmp(node_type, "LOD") == 0) return new LOD;
  else if (strcmp(node_type, "Switch") == 0) return new Switch;
  else
    return NULL;
}  // parse_special



VrmlNode *parse_geometry(char *node_type)
{
  if (strcmp(node_type, "Box") == 0) return new Box;
  else if (strcmp(node_type, "Cone") == 0) return new Cone;
  else if (strcmp(node_type, "Cylinder") == 0) return new Cylinder;
  else if (strcmp(node_type, "ElevationGrid") == 0) return new ElevationGrid;
  else if (strcmp(node_type, "IndexedFaceSet") == 0) return new IndexedFaceSet;
  else if (strcmp(node_type, "IndexedLineSet") == 0) return new IndexedLineSet;
  else if (strcmp(node_type, "PointSet") == 0) return new PointSet;
  else if (strcmp(node_type, "Sphere") == 0) return new Sphere;
  else if (strcmp(node_type, "Text") == 0) return new Text;
  else
    return NULL;
}  // parse_geometry



VrmlNode *parse_common(char *node_type)
{
  if (strcmp(node_type, "AudioClip") == 0) return new AudioClip;
  else if (strcmp(node_type, "Script") == 0) return new Script;
  else if (strcmp(node_type, "Shape") == 0) return new Shape;
  else if (strcmp(node_type, "Sound") == 0) return new Sound;
  else if (strcmp(node_type, "WorldInfo") == 0) return new WorldInfo;
  else
    return NULL;
}  // parse_common



VrmlNode *parse_properties(char *node_type)
{
  if (strcmp(node_type, "Color") == 0) return new Color;
  else if (strcmp(node_type, "Coordinate") == 0) return new Coordinate;
  else if (strcmp(node_type, "Normal") == 0) return new Normal;
  else if (strcmp(node_type, "TextureCoordinate") == 0)
    return new TextureCoordinate;
  else
    return NULL;
}  // parse_properties



VrmlNode *parse_appearance(char *node_type)
{
  if (strcmp(node_type, "Appearance") == 0) return new Appearance;
  else if (strcmp(node_type, "FontStyle") == 0) return new FontStyle;
  else if (strcmp(node_type, "ImageTexture") == 0) return new ImageTexture;
  else if (strcmp(node_type, "Material") == 0) return new Material;
  else if (strcmp(node_type, "MovieTexture") == 0) return new MovieTexture;
  else if (strcmp(node_type, "PixelTexture") == 0) return new PixelTexture;
  else if (strcmp(node_type, "TextureTransform") == 0) 
    return new TextureTransform;
  else
    return NULL;
}  // parse_appearance



VrmlNode *parse_bindable(char *node_type)
{
  if (strcmp(node_type, "Background") == 0) return new Background;
  else if (strcmp(node_type, "Fog") == 0) return new Fog;
  else if (strcmp(node_type, "NavigationInfo") == 0) return new NavigationInfo;
  else if (strcmp(node_type, "Viewpoint") == 0) return new Viewpoint;
  else
    return NULL;
}  // parse_bindable



VrmlNode *parse_interpolators(char *node_type)
{
  if (strcmp(node_type, "ColorInterpolator") == 0) 
    return new ColorInterpolator;
  else if (strcmp(node_type, "CoordinateInterpolator") == 0) 
    return new CoordinateInterpolator;
  else if (strcmp(node_type, "NormalInterpolator") == 0) 
    return new NormalInterpolator;
  else if (strcmp(node_type, "OrientationInterpolator") == 0) 
    return new OrientationInterpolator;
  else if (strcmp(node_type, "PositionInterpolator") == 0) 
    return new PositionInterpolator;
  else if (strcmp(node_type, "ScalarInterpolator") == 0) 
    return new ScalarInterpolator;
  else
    return NULL;
}  // parse_interpolators



VrmlNode *parse_node(char *node_type)
{
  VrmlNode *vrml_node_ptr = NULL;

  //
  // empty node
  //
  if (strcmp(node_type, "NULL") == 0) return NULL;

  // yes, there's a better way to do this ...

  //
  // DEF or USE
  //
  if (strcmp(node_type, "DEF") == 0) return parse_DEF();
  else if (strcmp(node_type, "USE") == 0) return parse_USE();

  //
  // IS (this is a bit superfluous, because testing for IS is also done
  //     in SFNode.C, but the parse function there, and this parse_node
  //     have to be integrated anyway)
  //
  else if (strcmp(node_type, "IS") == 0) {
    SFNode *SFNode_ptr = (SFNode *) PROTOVarTable.handle_IS();
    SFNode_ptr->parse_my_string();
    return SFNode_ptr->get_node_ptr();
  }

  //
  // nodes that are skipped
  //
  else if (strcmp(node_type, "Extrusion") == 0) {
    VrmlNode *vrml_node_ptr = new Box;
    vrml_node_ptr->add_Ring();
    cur_vrml_file->skip_node();
    return vrml_node_ptr;
  }
  else if ((strcmp(node_type, "CoKeyframeAnimation") == 0) ||
	   (strcmp(node_type, "CoHermiteOrientationInterpolator") == 0) ||
	   (strcmp(node_type, "CoHermitePositionInterpolator") == 0)) {
    cur_vrml_file->skip_node();
    VrmlNode *vrml_node_ptr = new Box;
    vrml_node_ptr->add_Ring();
    return vrml_node_ptr;
  }

  if (!vrml_node_ptr) vrml_node_ptr = parse_grouping(node_type);
  if (!vrml_node_ptr) vrml_node_ptr = parse_lights(node_type);
  if (!vrml_node_ptr) vrml_node_ptr = parse_sensors(node_type);
  if (!vrml_node_ptr) vrml_node_ptr = parse_geometry(node_type);
  if (!vrml_node_ptr) vrml_node_ptr = parse_special(node_type);
  if (!vrml_node_ptr) vrml_node_ptr = parse_common(node_type);
  if (!vrml_node_ptr) vrml_node_ptr = parse_properties(node_type);
  if (!vrml_node_ptr) vrml_node_ptr = parse_appearance(node_type);
  if (!vrml_node_ptr) vrml_node_ptr = parse_interpolators(node_type);
  if (!vrml_node_ptr) vrml_node_ptr = parse_bindable(node_type);

  //
  // ROUTEs and PROTOs
  //
  if (!vrml_node_ptr) {
    if (strcmp(node_type, "PROTO") == 0) {
      vrml_node_ptr = new PROTO;
      my_protos.add_PROTO((PROTO *) vrml_node_ptr);
    }
    else if (strcmp(node_type, "EXTERNPROTO") == 0) {
      vrml_node_ptr = new EXTERNPROTO;
      my_protos.add_PROTO((PROTO *) vrml_node_ptr);
    }
    else if (strcmp(node_type, "ROUTE") == 0) parse_route();
    else {
      // check if the node_type occurs in our list of PROTOs
      PROTO *proto_ptr = my_protos.find_PROTO(node_type);
      if (proto_ptr) {
#ifdef __DEBUG__
	cout << "Found corresponding PROTO for " << node_type << endl;
#endif
	PROTO *new_proto_ptr = new PROTO(*proto_ptr);
	new_proto_ptr->instantiate();
	return new_proto_ptr;
      }
      else {
	// check if the node_type occurs in the user defined list of nodes
	UserNode *usernode_ptr = user_nodelist.find_user_node(node_type);
	if (usernode_ptr) {
	return usernode_ptr->call_parse_and_init();
	}
	else {
	  cur_vrml_file->print_error_line();
	  cerr << "Unknown node type [" << node_type << "]" << endl;
	  exit(1);
	}
      }  // else, check if node in user_nodelist
    }  // else, check if node in my_protos
  }  // if, no VRML node match found

  if (vrml_node_ptr != NULL) {
    if (cur_vrml_file->get_just_parsing()) {
      vrml_node_ptr->parse();
      delete vrml_node_ptr;
      return NULL;
    }
    else {
      vrml_node_ptr->init_Ring();
      vrml_node_ptr->parse();
      vrml_node_ptr->add_Ring();
    }
  }
  return vrml_node_ptr;

}  // parse_node



void parse_route()
{
#ifdef __DEBUG__
  cout << "Parsing ROUTE statement" << endl;
#endif

  char source_node_id[MAX_TOKEN_LENGTH];
  char dest_node_id[MAX_TOKEN_LENGTH];
  char source_fieldname[MAX_TOKEN_LENGTH];
  char dest_fieldname[MAX_TOKEN_LENGTH];

  cur_vrml_file->get_token(source_node_id);  // get the eventOut
  cur_vrml_file->accept_token(".");
  cur_vrml_file->get_token(source_fieldname);

  cur_vrml_file->accept_token("TO");

  cur_vrml_file->get_token(dest_node_id);  // get the eventIn
  cur_vrml_file->accept_token(".");
  cur_vrml_file->get_token(dest_fieldname);

  VrmlNode *source = my_DEF_table.find_DEF(source_node_id);
  VrmlNode *dest = my_DEF_table.find_DEF(dest_node_id);

  if (!source || !dest) return;

  int source_field_id = source->get_field_var_index(source_fieldname);
  if (source_field_id == -1) {
    // try again without "_changed" at the end
    char *changed_ptr = strstr(source_fieldname, "_changed");
    if (changed_ptr) {
      *changed_ptr = '\0';
      source_field_id = source->get_field_var_index(source_fieldname);
    }
  }

  int dest_field_id = dest->get_field_var_index(dest_fieldname);
  if (dest_field_id == -1) {
    // try again without "set_" at the beginning
    if (strncmp(dest_fieldname, "set_", 4) == 0) {
      char *actual_dest_fieldname = dest_fieldname + 4;
      dest_field_id = dest->get_field_var_index(actual_dest_fieldname);
    }
  }

  if (source_field_id == -1) {
    cur_vrml_file->print_error_line();
    cerr << "Couldn't find field " << source_fieldname
	 << " in " << source->get_node_name() << " node" << endl;
    return;
  }
  if (dest_field_id == -1) {
    cur_vrml_file->print_error_line();
    cerr << "Couldn't find field " << dest_fieldname
	 << " in " << dest->get_node_name() << " node" << endl;
    return;
  }

#ifdef __DEBUG_ROUTING__
  cout << "Adding route from a " << source->get_node_name() << " to a "
       << dest->get_node_name() << ", fields: "
       << (*(source->get_var_table_ptr()))[source_field_id]->get_field_var_name() << " to "
       << (*(dest->get_var_table_ptr()))[dest_field_id]->get_field_var_name() << endl;
  cout << "Pointers: " << source << " and " << dest << endl;
#endif

  Event *new_route = new Event(source, source_field_id, dest, dest_field_id);
  // now add the route to the field itself instead of to a routing table
  source->get_field_ptr(source_field_id)->add_route(new_route);

}  // parse_route



VrmlNode *parse_DEF()
{
  VrmlNode *vrml_node_ptr;

  char node_id[MAX_TOKEN_LENGTH];
  cur_vrml_file->get_token(node_id);

#ifdef __DEBUG__
  cout << "Parsing DEF statement, node_id: " << node_id << endl;
#endif

  if (my_DEF_table.find_DEF(node_id)) {
    cur_vrml_file->print_error_line(1);
    cerr << "id [" << node_id << "] DEF'd more than once" << endl;
  }

  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->get_token(token);
  vrml_node_ptr = parse_node(token);

  // now store this node under this id
  // the constructor stores the name in Ring too, if there's a
  // corresponding node
  DEF *new_def = new DEF(node_id, vrml_node_ptr);
  my_DEF_table.add(new_def);

  VrmlNodeType node_type = vrml_node_ptr->get_node_type();
  // add a few more types here... Anchor etc.
  if ((node_type == TransformType) || (node_type == GroupType) ||
      (node_type == InlineType) || (node_type == AnchorType) ||
      (node_type == BillboardType) || (node_type == LODType) ||
      (node_type == SwitchType)) {
    R3Node *ring_ptr = vrml_node_ptr->get_ring_node_ptr();
    if (ring_ptr) {
      char *name = (char *) ring_ptr->Name();
      if (name && (strncmp(name, "D_", 2) == 0)) {
	GroupingNode *vrml_group_ptr = (GroupingNode *) vrml_node_ptr;
	R3GroupNode *ring_group_ptr = vrml_group_ptr->get_group_node_ptr();
	//	cout << "GroupingNode: found D_ identifier [" << name
	//	     << "]" << endl;
	ring_group_ptr->SetFlags(ring_group_ptr->Flags() | 
				 R3_NO_OPTIMIZE_NODE_FLAG);
      }
    }  // if it has a corresponding ring node
  }  // if it is a grouping node

  return vrml_node_ptr;
}  // parse_DEF



VrmlNode *parse_USE()
{
  VrmlNode *vrml_node_ptr;

  char node_id[MAX_TOKEN_LENGTH];
  cur_vrml_file->get_token(node_id);

#ifdef __DEBUG__
  cout << "Parsing USE statement, node_id: " << node_id << endl;
#endif

  vrml_node_ptr = my_DEF_table.find_DEF(node_id);
  if (!vrml_node_ptr) {
    cur_vrml_file->print_error_line();
    cerr << "Couldn't find DEF for this id: " << node_id << endl;
    return NULL;
  }
  
  // depending on the node type, get the appropriate thing to do
  // (see vrml_types.h)
  USE_ActionType action = vrml_node_ptr->get_USE_action();
  // do it once more if we get if from a PROTO...
  if (action == ReturnFirstVrmlNode) {
    vrml_node_ptr = ((PROTO *) vrml_node_ptr)->get_first_node_ptr();
    action = vrml_node_ptr->get_USE_action();
  }

  switch (action) {
    case HeadInsertNode:
      group_stack.Head()->InsertNode(vrml_node_ptr->get_ring_node_ptr());
      break;
    case HeadInsertGroupNode: {
      R3GroupNode *group_node_ptr =
	((GroupingNode *) vrml_node_ptr)->get_group_node_ptr();
      group_stack.Head()->InsertNode(group_node_ptr);
      break;
    }
    case CreateNodeReplaceHead: {
      R3LightNode *lightnode = 
	(R3LightNode *)((GroupingNode *) vrml_node_ptr)->get_group_node_ptr();
      R3LightNode *newnode = 
	new R3LightNode((R3Light *) lightnode->Light());
      group_stack.Head()->InsertNode(newnode);
      // Replace top of group_stack (for Lights)
      group_stack.RemoveHead();
      group_stack.Push(newnode);
      break;
    }
    case ReturnVrmlNode:
      break;
    case Unknown:
      fprintf(stderr, "Error: don't know how to handle USE of %s\n",
	      vrml_node_ptr->get_node_name());
      break;
    default:
      fprintf(stderr, "Error: unknown action type in USE\n");
  }  // switch

  return vrml_node_ptr;

}  // parse_USE
